MAIN_ROBOT = """
*** Settings ***
Documentation    Documentação da Suite de Tasks Robot Framework

Library     rpapy.core.activities
Library     ../resources/custom_keywords.py
Resource    ../resources/keywords.robot

Suite Setup     Toast Process Start Notifier

  
*** Variables ***
${VAR}=             Everybody
${LAPIS}=           btn_lapis
${NOME_IMAGEM}      robotframework


*** Tasks ***
Tarefa principal
    Keyword principal


*** Keywords ***
Keyword principal
    Primeira Keyword        ${VAR}
    Abrir Mspaint
    Reduzir zoom
    Definir tamanho da imagem    2700   1100
    Selecionar ferramenta "${LAPIS}"
    Desenhar            ${NOME_IMAGEM}
    ${DADOS_CORES}      Get Dados Cores     ${NOME_IMAGEM}
    Pintar desenho      ${DADOS_CORES}
"""

KEYWORDS_ROBOT = """
*** Settings ***
Documentation    Documentacao das Keywords da Suite de Tasks
Library     rpapy.core.activities


*** Variables ***
${VAR}=         variables
${LIMPAR}=      {HOME}{DEL 20}
${BALDE}=       btn_balde



*** Keywords ***
Primeira Keyword
    [Arguments]         ${arg}
    Log To Console      Hello ${arg}!


Abrir Mspaint
    Open Executable     mspaint.exe     Sem Título - Paint      work_dir=C:/Windows/System32/


Editar Cor RGB
    [Arguments]     ${r}    ${g}    ${b}    ${matiz}=0   ${sat}=0    ${lum}=

    Click Vision    btn_editar_cores    backend=uia

    Wait Element Vision    titulo_editar_cores

    Write Text Vision    campo_red    backend=uia
    ...    text=${LIMPAR}${r}      move_x=30

    Write Text Vision    campo_green    backend=uia
    ...    text=${LIMPAR}${g}      move_x=30

    Write Text Vision    campo_blue    backend=uia
    ...    text=${LIMPAR}${b}      move_x=30

    Write Text Vision    campo_matiz      backend=uia
    ...    text=${LIMPAR}${matiz}    move_x=30

    Write Text Vision    campo_sat      backend=uia
    ...    text=${LIMPAR}${sat}    move_x=30

    Write Text Vision    campo_lum    backend=uia
    ...    text=${LIMPAR}${lum}    move_x=30

    Click Vision    btn_ok_cor_definida    backend=uia


Reduzir zoom
    Wait Element Vision    icone_paint
    Click Vision    btn_reduzir_zoom    backend=uia     after=0.5


Aumentar zoom
    Wait Element Vision    icone_paint
    Click Vision    btn_aumentar_zoom    backend=uia     after=0.5


Definir tamanho da imagem
    [Arguments]    ${largura}    ${altura}

    Wait Element Vision    icone_paint    

    Click Vision    btn_menu_arquivo    backend=uia
    
    Click Vision    btn_item_menu_propriedades    backend=uia

    Write Text Vision    campo_largura_imagem    backend=uia
    ...    text=${LIMPAR}${largura}    move_x=40
    
    Write Text Vision    campo_altura_imagem    backend=uia
    ...    text=${LIMPAR}${altura}    move_x=40
    
    Click Vision    btn_ok_propriedades_imagem    backend=uia    after=0.5


Selecionar ferramenta "${nome_ferramenta}"

    Wait Element Vision    icone_paint

    Click Vision    ${nome_ferramenta}    backend=uia


Pintar desenho
    [Arguments]     ${dados_imagem}

    Selecionar ferramenta "${BALDE}"

    FOR     ${key}     ${value}   IN      &{dados_imagem}
        
        ${cor}      Set Variable      ${value}[param_cor]
        Editar Cor RGB  @{cor}

        Pintar area     ${value}[coordenadas]
   
    END


Pintar area
    [Arguments]     ${coordenadas}
    
    FOR    ${coord}   IN    @{coordenadas}
        Click Coord     ${coord}[0]     ${coord}[1]
    END
"""

CUSTOM_KEYWORDS = """
import os
from pathlib import Path

import mouse
import pyautogui


def desenhar(nome_imagem: str):
        
    # The directory containing this file
    HERE = os.path.abspath(os.path.dirname(__file__))

    # The text of the README file
    with open(os.path.join(HERE, F"desenhos/{nome_imagem}.txt")) as fid:
        DESENHO = fid.read()


    flag_levantado = True
    for line in DESENHO.splitlines():
        
        partes_line = line.split()
        if len(partes_line) < 2:
            continue        

        if 'Z5' in partes_line[1]:
            flag_levantado = True
        
        elif 'Z-' in partes_line[1]:
            flag_levantado = False
        
        if 'X' in partes_line[1]:
            x, y = [int(float(coord.replace('X', '').replace('Y', ''))) for coord in partes_line[1:3]]
            
            if flag_levantado:
                pyautogui.moveTo(x+100, y+100, .2)
                
            else:
                start_x, start_y = mouse.get_position()
                mouse.drag(start_x, start_y, x+100, y+100, absolute=True, duration=float(1*10**-250))

def get_dados_cores(nome_imagem: str):        
    return {
        'robotframework':{
            'preenchimento': {
                'param_cor': (0, 0, 0),
                'coordenadas': [(331, 309),(341, 213),(540, 229),(616, 248),(683, 233),(690, 266),
                                (766, 249),(608, 332),(687, 358),(615, 450),(687, 427),(538, 523),
                                (613, 528),(687, 550),(764, 529)]
            },
            'iniciais_azul': {
                'param_cor': (0, 192, 182, 118, 240, 90),
                'coordenadas': [(538, 244),(519, 344),(538, 471)]
            },
            'amarelo': {
                'param_cor': (255, 201, 14, 31, 240, 127),
                'coordenadas': [(537, 537),(596, 547),(686, 519),(743, 544),(832, 540)]
            },
            'vermelho': {
                'param_cor': (255, 0, 0, 0, 240, 120),
                'coordenadas': [(930, 384),(1082, 308)]
            }
        },
        'borboleta':{
            'contorno': {
                'param_cor': (0, 0, 0),                         # Cor preto
                'coordenadas': [(419, 459)]
            },
            'asas': {
                'param_cor': (255, 0, 128, 220, 240, 120),      # Cor rosa
                'coordenadas': [(203, 247),(242, 233),(303, 252),(224, 277),(273, 280),(326, 277),(358, 297),
                                (350, 309),(302, 316),(228, 319),(336, 321),(351, 345),(395, 345),(442, 387),(246, 348),(233, 376),
                                (251, 400),(282, 359),(309, 347),(299, 382),(326, 376),(283, 402),(360, 385),(396, 409),(428, 429),
                                (394, 430),(328, 430),(333, 439),(411, 450),(272, 442),(252, 457),(259, 484),(291, 475),(337, 449),
                                (304, 489),(432, 464),(265, 507),(262, 528),(285, 520),(311, 526),(359, 491),(359, 522),(380, 538),
                                (420, 496),(252, 546),(297, 552),(318, 555),(348, 563)] + [(666, 232),(714, 249),(609, 248),
                                (550, 299),(582, 274),(633, 283),(673, 276),(626, 309),(559, 306),(574, 316),(675, 322),(521, 336),
                                (475, 386),(559, 347),(587, 354),(651, 353),(631, 360),(542, 388),(583, 372),(610, 377),(678, 372),
                                (637, 395),(660, 397),(528, 413),(481, 424),(513, 434),(578, 431),(499, 450),(581, 438),(627, 441),
                                (574, 449),(607, 469),(650, 453),(657, 456),(656, 483),(647, 500),(617, 490),(475, 461),(489, 488),
                                (518, 514),(546, 514),(557, 502),(589, 521),(623, 514),(646, 525),(557, 563),(592, 557),(618, 558),(642, 551)]
            },
            'corpo': {
                'param_cor': (251, 251, 0, 40, 240, 118),       # Cor amarelo
                'coordenadas': [(455, 371),(458, 414),(456, 464)]
            }
        }
    }.get(nome_imagem)
"""

ROBOTFRAMEWORK = """
%
(Header)
(Generated by gcodetools from Inkscape.)
(Using default header. To add your own header create file "header" in the output dir.)
M3
(Header end.)
G21 (All units in mm)



(Start cutting path id: path25585)
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X132.091120 Y433.699631

G01 Z-1.000000 F100.0(Penetrate)
G03 X129.189570 Y432.353844 Z-1.000000 I7.454922 J-19.873803 F400.000000
G03 X125.783180 Y430.173721 Z-1.000000 I16.558055 J-29.622821
G03 X124.068844 Y428.753673 Z-1.000000 I10.426539 J-14.332143
G03 X117.813876 Y422.730292 Z-1.000000 I249.631983 J-265.489254
G03 X111.835978 Y416.758756 Z-1.000000 I462.655328 J-469.126184
G03 X104.539661 Y409.308753 Z-1.000000 I816.455776 J-806.910963
G03 X97.306672 Y401.786687 Z-1.000000 I809.493080 J-785.622423
G03 X91.768576 Y395.893622 Z-1.000000 I428.753800 J-408.476596
G03 X86.170442 Y389.686396 Z-1.000000 I224.092832 J-207.731446
G03 X85.308660 Y388.469417 Z-1.000000 I5.398381 J-4.736383
G03 X83.172079 Y383.615795 Z-1.000000 I23.749692 J-13.351771
G03 X81.661317 Y377.359218 Z-1.000000 I38.132791 J-12.518531
G03 X81.555615 Y376.235686 Z-1.000000 I8.872502 J-1.401468
G03 X81.370540 Y368.314064 Z-1.000000 I521.889482 J-16.155991
G03 X81.272891 Y360.541037 Z-1.000000 I1210.994098 J-19.100279
G03 X81.177207 Y348.002688 Z-1.000000 I3641.693252 J-34.060519
G03 X81.078803 Y322.223081 Z-1.000000 I6159.631990 J-36.402000
G03 X81.067259 Y288.635092 Z-1.000000 I12162.464040 J-20.974221
G03 X81.148549 Y255.044793 Z-1.000000 I12144.590471 J12.595245
G03 X81.300248 Y229.364063 Z-1.000000 I6102.740633 J23.208776
G03 X81.422027 Y216.873638 Z-1.000000 I3592.728788 J28.782315
G03 X81.535586 Y209.173398 Z-1.000000 I1181.522281 J13.573615
G03 X81.738226 Y201.322244 Z-1.000000 I503.422837 J9.065232
G03 X81.844949 Y200.297238 Z-1.000000 I7.348022 J0.247013
G03 X82.448553 Y197.266964 Z-1.000000 I66.949717 J11.760535
G03 X83.163569 Y194.876083 Z-1.000000 I17.107973 J3.813950
G03 X84.227182 Y192.625928 Z-1.000000 I14.615379 J5.532007
G03 X86.246863 Y189.571961 Z-1.000000 I25.205860 J14.474578
G03 X90.857044 Y184.124226 Z-1.000000 I51.786993 J39.150507
G03 X105.934590 Y168.432261 Z-1.000000 I973.357838 J920.156456
G03 X118.325887 Y155.840505 Z-1.000000 I1655.007102 J1616.266749
G03 X122.900501 Y151.375700 Z-1.000000 I138.793639 J137.631209
G03 X127.306399 Y147.642331 Z-1.000000 I36.775482 J38.933746
G03 X129.133140 Y146.519711 Z-1.000000 I6.624503 J8.731914
G03 X131.528263 Y145.388415 Z-1.000000 I49.103500 J100.858419
G03 X134.163646 Y144.488499 Z-1.000000 I5.784125 J12.629881
G03 X137.098538 Y144.007437 Z-1.000000 I5.096105 J21.897407
G03 X145.010522 Y143.485081 Z-1.000000 I12.558366 J130.036706
G03 X160.787806 Y143.188323 Z-1.000000 I16.737574 J470.310424
G03 X224.356010 Y143.130891 Z-1.000000 I57.039205 J27953.059560
G03 X287.794450 Y143.223418 Z-1.000000 I-6.052514 J25897.027378
G03 X303.834231 Y143.526837 Z-1.000000 I-1.326165 J494.216879
G03 X311.890262 Y144.042556 Z-1.000000 I-4.927992 J140.159457
G03 X314.916686 Y144.516119 Z-1.000000 I-2.270093 J24.414948
G03 X317.641777 Y145.406663 Z-1.000000 I-3.160355 J14.285476
G03 X320.001240 Y146.493761 Z-1.000000 I-29.417458 J66.952417
G03 X321.204681 Y147.130163 Z-1.000000 I-8.012190 J16.607155
G03 X322.597627 Y148.034876 Z-1.000000 I-7.885592 J13.665786
G03 X323.932763 Y149.084828 Z-1.000000 I-13.246427 J18.218238
G03 X326.260638 Y151.156421 Z-1.000000 I-43.657408 J51.402094
G03 X330.645304 Y155.400665 Z-1.000000 I-110.615143 J118.661888
G03 X343.308770 Y168.220751 Z-1.000000 I-1770.140381 J1761.179363
G03 X356.254608 Y181.510431 Z-1.000000 I-1976.113799 J1937.936268
G03 X360.328596 Y185.902343 Z-1.000000 I-102.125491 J98.818293
G03 X362.304261 Y188.245932 Z-1.000000 I-43.451007 J38.634101
G03 X363.333524 Y189.646993 Z-1.000000 I-16.787395 J13.411133
G03 X364.218527 Y191.108271 Z-1.000000 I-14.588023 J9.833674
G03 X364.975200 Y192.588558 Z-1.000000 I-31.190819 J16.877235
G01 X367.130720 Y197.061068 Z-1.000000
G01 X367.286860 Y287.908968 Z-1.000000
G03 X367.319711 Y334.302011 Z-1.000000 I-22951.642374 J39.448999
G03 X367.264629 Y357.209555 Z-1.000000 I-5449.000229 J-1.648653
G03 X367.188941 Y368.687916 Z-1.000000 I-2749.595525 J-12.391301
G03 X367.113640 Y374.619614 Z-1.000000 I-739.117956 J-6.416576
G03 X366.951949 Y380.521908 Z-1.000000 I-276.373460 J-4.617778
G03 X366.866670 Y381.136568 Z-1.000000 I-3.112098 J-0.118530
G03 X366.232947 Y383.477533 Z-1.000000 I-44.043967 J-10.666897
G03 X365.482078 Y385.468510 Z-1.000000 I-14.674636 J-4.397249
G03 X364.457733 Y387.344617 Z-1.000000 I-13.965233 J-6.407250
G03 X362.504851 Y390.077129 Z-1.000000 I-27.671929 J-17.712579
G03 X358.207186 Y395.026058 Z-1.000000 I-59.084122 J-46.968310
G03 X344.397130 Y409.402291 Z-1.000000 I-1064.142394 J-1008.412268
G03 X336.269284 Y417.663312 Z-1.000000 I-804.710592 J-783.608989
G03 X330.205431 Y423.690561 Z-1.000000 I-505.781582 J-502.789281
G03 X323.835864 Y429.806086 Z-1.000000 I-273.215102 J-278.189770
G03 X322.700830 Y430.710611 Z-1.000000 I-6.147553 J-6.549789
G03 X319.862741 Y432.365941 Z-1.000000 I-14.341186 J-21.327548
G03 X316.630620 Y433.748801 Z-1.000000 I-13.170082 J-26.313469
G01 X313.303000 Y434.942801 Z-1.000000
G01 X224.205480 Y434.887001 Z-1.000000
G03 X162.575523 Y434.811851 Z-1.000000 I32.504808 J-51929.110443
G03 X146.257055 Y434.691887 Z-1.000000 I2.670640 J-1473.227675
G03 X138.043889 Y434.505163 Z-1.000000 I5.380156 J-417.373513
G03 X135.363035 Y434.349036 Z-1.000000 I1.708213 J-52.426032
G03 X133.034169 Y433.984083 Z-1.000000 I1.377223 J-16.401511
G03 X132.091120 Y433.699591 Z-1.000000 I1.684664 J-7.289683
G01 X132.091120 Y433.699631 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X306.991180 Y403.982191

G01 Z-1.000000 F100.0(Penetrate)
G02 X307.516907 Y403.595748 Z-1.000000 I-1.346914 J-2.383203 F400.000000
G02 X312.066504 Y399.238099 Z-1.000000 I-204.834891 J-218.411406
G02 X316.387386 Y394.967365 Z-1.000000 I-409.291334 J-418.418521
G02 X322.451250 Y388.859367 Z-1.000000 I-711.698693 J-712.620246
G02 X330.969324 Y380.107409 Z-1.000000 I-771.373009 J-759.280382
G02 X334.164971 Y376.646685 Z-1.000000 I-82.313678 J-79.214594
G02 X336.814684 Y373.236123 Z-1.000000 I-25.087957 J-22.225761
G02 X337.475250 Y371.900598 Z-1.000000 I-4.523978 J-3.068734
G02 X337.655550 Y371.114235 Z-1.000000 I-3.940547 J-1.317354
G02 X337.931253 Y368.039653 Z-1.000000 I-39.548788 J-5.096060
G02 X338.053074 Y364.726116 Z-1.000000 I-118.543662 J-6.017236
G02 X338.190535 Y356.442415 Z-1.000000 I-669.292518 J-15.249295
G02 X338.289719 Y340.096106 Z-1.000000 I-1884.338050 J-19.607030
G02 X338.333570 Y289.411668 Z-1.000000 I-29291.562075 J-50.684565
G02 X338.286273 Y245.504266 Z-1.000000 I-20380.198965 J0.000077
G02 X338.191360 Y226.558882 Z-1.000000 I-3317.573412 J7.147441
G02 X337.874007 Y208.422714 Z-1.000000 I-941.637976 J7.406227
G02 X337.686870 Y207.189078 Z-1.000000 I-5.065879 J0.137461
G02 X337.141980 Y206.011703 Z-1.000000 I-3.932410 J1.105147
G02 X334.504581 Y202.587088 Z-1.000000 I-28.318974 J19.081414
G02 X331.336221 Y199.113491 Z-1.000000 I-88.932225 J77.935691
G02 X323.014040 Y190.445948 Z-1.000000 I-762.138223 J723.441387
G02 X316.987091 Y184.330876 Z-1.000000 I-419.018271 J406.952363
G02 X312.510271 Y179.931373 Z-1.000000 I-257.482884 J257.530338
G02 X307.776447 Y175.507504 Z-1.000000 I-136.367463 J141.177280
G02 X306.952820 Y174.931921 Z-1.000000 I-2.700066 J2.986567
G02 X306.313917 Y174.684025 Z-1.000000 I-1.395235 J2.648667
G02 X304.237209 Y174.324706 Z-1.000000 I-3.169989 J12.140278
G02 X301.816171 Y174.178465 Z-1.000000 I-3.926827 J44.895832
G02 X294.355707 Y174.005056 Z-1.000000 I-12.248201 J366.375481
G02 X279.542106 Y173.897992 Z-1.000000 I-16.583162 J1269.605927
G02 X224.879360 Y173.859901 Z-1.000000 I-54.662864 J39222.169012
G02 X175.286278 Y173.905470 Z-1.000000 I-0.000486 J26986.271011
G02 X158.589916 Y174.004445 Z-1.000000 I3.750957 J2041.090107
G02 X150.130982 Y174.136811 Z-1.000000 I7.528239 J751.446973
G02 X146.698769 Y174.253464 Z-1.000000 I2.876705 J135.190284
G02 X143.501114 Y174.523210 Z-1.000000 I1.997550 J42.767448
G02 X142.912760 Y174.681061 Z-1.000000 I0.263964 J2.159272
G02 X142.076109 Y175.201839 Z-1.000000 I1.257280 J2.952316
G02 X137.665330 Y179.186518 Z-1.000000 I81.654046 J94.819253
G02 X133.447843 Y183.269920 Z-1.000000 I186.846193 J197.201286
G02 X126.677400 Y190.063898 Z-1.000000 I476.273758 J481.394373
G02 X117.334479 Y199.662561 Z-1.000000 I1019.860627 J1002.034312
G02 X114.397961 Y202.877694 Z-1.000000 I57.891584 J55.823430
G02 X113.004945 Y204.606030 Z-1.000000 I24.210518 J20.938909
G02 X112.351235 Y205.590272 Z-1.000000 I8.233851 J6.177936
G02 X111.848378 Y206.629105 Z-1.000000 I6.390872 J3.734685
G02 X111.527250 Y207.581178 Z-1.000000 I9.861364 J3.856370
G02 X111.347871 Y208.482548 Z-1.000000 I5.434374 J1.550015
G02 X111.086295 Y211.720917 Z-1.000000 I46.870151 J5.415640
G02 X110.971427 Y215.181877 Z-1.000000 I133.199009 J6.153189
G02 X110.851166 Y223.593449 Z-1.000000 I713.283394 J14.404514
G02 X110.781906 Y240.183873 Z-1.000000 I1963.033608 J16.490452
G02 X110.828730 Y290.733378 Z-1.000000 I28873.196409 J-1.470622
G01 X110.973210 Y370.929978 Z-1.000000
G01 X112.169700 Y373.166238 Z-1.000000
G02 X112.732296 Y373.976870 Z-1.000000 I3.777493 J-2.021117
G02 X116.927203 Y378.613506 Z-1.000000 I176.329231 J-155.314512
G02 X121.064087 Y383.007482 Z-1.000000 I337.843297 J-313.931713
G02 X126.468559 Y388.623312 Z-1.000000 I639.721704 J-610.236557
G02 X131.919507 Y394.188945 Z-1.000000 I641.143116 J-622.479571
G02 X136.305398 Y398.572709 Z-1.000000 I350.457785 J-346.242006
G02 X140.915116 Y403.012816 Z-1.000000 I184.216058 J-186.640112
G02 X141.949350 Y403.838751 Z-1.000000 I5.631542 J-5.991305
G02 X142.384576 Y404.017363 Z-1.000000 I0.649548 J-0.963193
G02 X144.926655 Y404.334063 Z-1.000000 I3.838362 J-20.448983
G02 X147.794647 Y404.460220 Z-1.000000 I4.864763 J-77.930575
G02 X155.837691 Y404.606676 Z-1.000000 I13.959469 J-545.696267
G02 X171.746295 Y404.703798 Z-1.000000 I18.212913 J-1680.302881
G02 X224.501180 Y404.733531 Z-1.000000 I44.783107 J-32657.849531
G02 X272.255179 Y404.674965 Z-1.000000 I-5.900698 J-24280.388270
G02 X290.042269 Y404.574303 Z-1.000000 I-5.697551 J-2578.319756
G02 X299.042564 Y404.453264 Z-1.000000 I-9.447692 J-1037.197617
G02 X302.917956 Y404.347585 Z-1.000000 I-3.638410 J-204.537763
G02 X306.598238 Y404.105670 Z-1.000000 I-2.344024 J-63.775134
G02 X306.991180 Y403.982191 Z-1.000000 I-0.093287 J-0.983818
G01 X306.991180 Y403.982191 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X151.014960 Y350.995578

G01 Z-1.000000 F100.0(Penetrate)
G03 X147.273539 Y348.272009 Z-1.000000 I7.507412 J-14.244702 F400.000000
G03 X144.274120 Y344.414058 Z-1.000000 I12.543204 J-12.846836
G03 X143.517586 Y342.907209 Z-1.000000 I12.367314 J-7.152514
G03 X143.097759 Y341.567287 Z-1.000000 I6.295768 J-2.708336
G03 X142.915075 Y340.137068 Z-1.000000 I9.558381 J-1.947679
G03 X142.842060 Y337.386138 Z-1.000000 I51.785587 J-2.750930
G03 X142.915175 Y334.637814 Z-1.000000 I51.689759 J0.000000
G03 X143.098105 Y333.203530 Z-1.000000 I9.809524 J0.522307
G03 X143.517612 Y331.858780 Z-1.000000 I6.786907 J1.379429
G03 X144.273020 Y330.345678 Z-1.000000 I13.234699 J5.662232
G03 X145.798787 Y328.015530 Z-1.000000 I20.857475 J11.992741
G03 X147.268074 Y326.320644 Z-1.000000 I11.780817 J8.728424
G03 X148.983380 Y324.889316 Z-1.000000 I9.295493 J9.396259
G03 X150.988670 Y323.702048 Z-1.000000 I8.776650 J12.536621
G03 X151.727949 Y323.406905 Z-1.000000 I2.245345 J4.550730
G03 X153.506171 Y323.036551 Z-1.000000 I2.812249 J9.048609
G03 X155.587402 Y322.879651 Z-1.000000 I3.340488 J30.428518
G03 X162.127004 Y322.694573 Z-1.000000 I10.468036 J254.252475
G03 X175.138734 Y322.585886 Z-1.000000 I14.175220 J918.103395
G03 X225.065220 Y322.558328 Z-1.000000 I44.230767 J34906.377509
G03 X260.253305 Y322.611855 Z-1.000000 I-2.230797 J13032.587776
G03 X278.986534 Y322.712165 Z-1.000000 I-10.829348 J3771.713713
G03 X288.333361 Y322.807283 Z-1.000000 I-15.667386 J1998.856831
G03 X293.338957 Y322.891916 Z-1.000000 I-7.132440 J569.917858
G03 X298.350469 Y323.055971 Z-1.000000 I-4.672819 J219.370535
G03 X298.866090 Y323.137068 Z-1.000000 I-0.103197 J2.335874
G03 X299.723522 Y323.469599 Z-1.000000 I-1.193945 J4.350298
G03 X301.136881 Y324.281289 Z-1.000000 I-7.529851 J14.747743
G03 X302.463888 Y325.208454 Z-1.000000 I-11.124968 J17.335851
G03 X303.695140 Y326.236028 Z-1.000000 I-10.468377 J13.794785
G03 X307.267413 Y331.087800 Z-1.000000 I-10.538665 J11.500415
G03 X308.639779 Y336.402350 Z-1.000000 I-13.311020 J6.271746
G03 X307.969179 Y341.850279 Z-1.000000 I-14.686712 J0.957413
G03 X305.057000 Y347.126928 Z-1.000000 I-14.866894 J-4.763091
G03 X302.918621 Y349.217080 Z-1.000000 I-11.671747 J-9.802116
G03 X300.442610 Y350.826028 Z-1.000000 I-9.296633 J-11.596957
G01 X297.779770 Y352.201338 Z-1.000000
G01 X225.494400 Y352.176638 Z-1.000000
G03 X175.155355 Y352.123861 Z-1.000000 I12.171539 J-35616.228439
G03 X162.079620 Y352.008719 Z-1.000000 I1.627589 J-927.341403
G03 X155.510692 Y351.819894 Z-1.000000 I4.044569 J-255.060077
G03 X153.451245 Y351.662478 Z-1.000000 I1.239619 J-29.768063
G03 X151.706610 Y351.287152 Z-1.000000 I0.980434 J-8.799846
G03 X151.014960 Y350.995588 Z-1.000000 I1.343214 J-4.152532
G01 X151.014960 Y350.995578 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X150.702180 Y282.119948

G01 Z-1.000000 F100.0(Penetrate)
G03 X147.883567 Y280.864960 Z-1.000000 I4.665316 J-14.270666 F400.000000
G03 X145.550684 Y279.153858 Z-1.000000 I6.508907 J-11.319968
G03 X143.636379 Y276.975070 Z-1.000000 I9.274082 J-10.078662
G03 X141.962810 Y274.078228 Z-1.000000 I15.050674 J-10.626950
G03 X141.452795 Y272.606263 Z-1.000000 I7.123089 J-3.292388
G03 X141.044534 Y270.314323 Z-1.000000 I19.587511 J-4.671437
G03 X140.866610 Y267.989872 Z-1.000000 I27.535566 J-3.276733
G03 X140.864649 Y265.228244 Z-1.000000 I41.032729 J-1.409954
G03 X141.402960 Y259.525357 Z-1.000000 I46.813843 J1.542039
G03 X142.602950 Y254.644658 Z-1.000000 I28.978270 J4.536856
G03 X143.825394 Y251.601714 Z-1.000000 I28.179738 J9.553646
G03 X145.819574 Y247.696232 Z-1.000000 I56.815397 J26.548667
G03 X148.079932 Y243.940268 Z-1.000000 I55.266149 J30.701318
G03 X150.052680 Y241.234218 Z-1.000000 I26.246573 J17.062014
G03 X154.166725 Y236.886921 Z-1.000000 I31.223325 J25.427749
G03 X158.813290 Y233.264585 Z-1.000000 I25.980748 J28.535540
G03 X163.953947 Y230.370438 Z-1.000000 I22.361095 J33.705782
G03 X170.024750 Y227.984348 Z-1.000000 I21.285517 J45.239755
G03 X174.650115 Y226.989237 Z-1.000000 I7.409884 J23.194693
G03 X181.143569 Y226.625847 Z-1.000000 I6.338350 J55.062804
G03 X187.632677 Y227.019961 Z-1.000000 I-0.157293 J56.208572
G03 X192.428010 Y228.047048 Z-1.000000 I-3.139844 J26.367454
G03 X202.113692 Y232.346160 Z-1.000000 I-13.886362 J44.345499
G03 X209.631384 Y238.169158 Z-1.000000 I-19.036746 J32.341399
G03 X215.457083 Y245.692736 Z-1.000000 I-26.824869 J26.788465
G03 X219.866560 Y255.522888 Z-1.000000 I-41.734839 J24.624900
G03 X221.203770 Y260.959888 Z-1.000000 I-38.607063 J12.378209
G03 X221.702416 Y265.885123 Z-1.000000 I-34.301458 J5.960644
G03 X221.405684 Y270.806947 Z-1.000000 I-27.430008 J0.816126
G03 X220.418640 Y274.078298 Z-1.000000 I-11.864974 J-1.795360
G03 X218.937416 Y276.492661 Z-1.000000 I-11.247848 J-5.239059
G03 X216.617226 Y279.018429 Z-1.000000 I-17.251441 J-13.518752
G03 X213.920586 Y281.137470 Z-1.000000 I-14.015844 J-15.060860
G03 X211.714400 Y282.178578 Z-1.000000 I-4.753105 J-7.214101
G03 X204.701980 Y282.721085 Z-1.000000 I-4.707668 J-15.258739
G03 X199.044905 Y280.547362 Z-1.000000 I1.986166 J-13.617039
G03 X194.831773 Y276.155247 Z-1.000000 I7.838871 J-11.736210
G03 X192.443030 Y269.306808 Z-1.000000 I14.740276 J-8.982241
G02 X191.756539 Y265.314974 Z-1.000000 I-43.354604 J5.400911
G02 X191.106920 Y263.237318 Z-1.000000 I-12.323945 J2.712931
G02 X189.603976 Y260.826799 Z-1.000000 I-7.852991 J3.222495
G02 X187.126664 Y258.628008 Z-1.000000 I-10.413511 J9.237639
G02 X184.199649 Y257.087643 Z-1.000000 I-7.733985 J11.145031
G02 X181.544250 Y256.601398 Z-1.000000 I-2.655399 J7.007488
G02 X178.510244 Y257.109321 Z-1.000000 I0.000000 J9.315566
G02 X175.301789 Y258.655069 Z-1.000000 I4.959099 J14.396139
G02 X172.562929 Y260.914179 Z-1.000000 I7.728321 J12.159325
G02 X171.093440 Y263.243808 Z-1.000000 I5.168293 J4.888347
G02 X170.660140 Y264.741047 Z-1.000000 I8.443313 J3.254808
G02 X170.012560 Y268.702748 Z-1.000000 I58.881824 J11.658602
G03 X169.322107 Y272.365785 Z-1.000000 I-32.038229 J-4.142355
G03 X168.459096 Y274.752942 Z-1.000000 I-11.993135 J-2.986213
G03 X167.112326 Y276.881696 Z-1.000000 I-9.944006 J-4.800738
G03 X164.953420 Y279.181988 Z-1.000000 I-16.323716 J-13.157143
G03 X158.361070 Y282.551825 Z-1.000000 I-9.402304 J-10.260388
G03 X150.702180 Y282.119968 Z-1.000000 I-3.016554 J-14.632379
G01 X150.702180 Y282.119948 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X251.438150 Y281.751328

G01 Z-1.000000 F100.0(Penetrate)
G03 X247.869319 Y279.701532 Z-1.000000 I5.354524 J-13.454265 F400.000000
G03 X245.183868 Y276.894222 Z-1.000000 I8.362284 J-10.687395
G03 X243.384565 Y273.439946 Z-1.000000 I11.413381 J-8.140890
G03 X242.483280 Y269.196768 Z-1.000000 I15.582188 J-5.527091
G03 X242.412045 Y266.501873 Z-1.000000 I26.203397 J-2.041036
G03 X242.633706 Y264.350249 Z-1.000000 I13.914756 J0.346276
G03 X243.213033 Y262.286889 Z-1.000000 I11.316989 J2.064450
G03 X244.191814 Y260.327879 Z-1.000000 I10.624555 J4.084329
G03 X246.937508 Y257.087337 Z-1.000000 I12.608182 J7.899367
G03 X253.065990 Y252.305688 Z-1.000000 I41.527541 J46.906323
G02 X259.439460 Y247.925590 Z-1.000000 I-1351.272952 J-1973.061274
G02 X271.601330 Y239.523298 Z-1.000000 I-4958.284590 J-7189.855230
G03 X276.808957 Y235.964374 Z-1.000000 I303.313251 J438.236264
G03 X280.749590 Y233.343987 Z-1.000000 I195.774394 J290.139394
G03 X284.863986 Y230.720593 Z-1.000000 I107.443893 J163.971402
G03 X285.675920 Y230.300778 Z-1.000000 I3.092560 J4.986033
G03 X287.742616 Y229.741301 Z-1.000000 I3.095192 J7.336671
G03 X291.158654 Y229.519278 Z-1.000000 I3.423521 J26.283641
G03 X294.574715 Y229.739216 Z-1.000000 I0.007488 J26.522636
G03 X296.646270 Y230.297778 Z-1.000000 I-1.028145 J7.933814
G03 X300.777921 Y232.926986 Z-1.000000 I-5.640981 J13.425414
G03 X303.801645 Y236.645317 Z-1.000000 I-10.593918 J11.703511
G03 X305.555925 Y241.096572 Z-1.000000 I-13.373048 J7.841755
G03 X305.764500 Y245.847678 Z-1.000000 I-13.206049 J2.959881
G03 X305.110848 Y248.667413 Z-1.000000 I-15.115680 J-2.018384
G03 X303.887790 Y251.574328 Z-1.000000 I-19.845984 J-6.639266
G03 X302.222483 Y254.250712 Z-1.000000 I-16.934003 J-8.680432
G03 X300.593970 Y255.934478 Z-1.000000 I-7.294134 J-5.425358
G03 X298.744926 Y257.327951 Z-1.000000 I-26.661462 J-33.454434
G03 X292.824556 Y261.486284 Z-1.000000 I-326.398747 J-458.412125
G03 X286.999437 Y265.490455 Z-1.000000 I-547.901445 J-790.827215
G03 X280.037706 Y270.210864 Z-1.000000 I-877.163011 J-1286.159640
G03 X273.036568 Y274.901632 Z-1.000000 I-831.071483 J-1232.833854
G03 X267.779609 Y278.369451 Z-1.000000 I-402.592491 J-604.582026
G03 X262.261828 Y281.902745 Z-1.000000 I-193.872948 J-296.687326
G03 X261.596450 Y282.223278 Z-1.000000 I-1.765524 J-2.814081
G03 X259.438837 Y282.715820 Z-1.000000 I-3.246885 J-9.251132
G03 X256.650673 Y282.806372 Z-1.000000 I-1.966983 J-17.594915
G03 X253.880345 Y282.471258 Z-1.000000 I0.875558 J-18.856547
G03 X251.438150 Y281.751608 Z-1.000000 I2.670449 J-13.566128
G01 X251.438150 Y281.751328 Z-1.000000
G00 Z5.000000

(End cutting path id: path25585) ROBOT_FRAMEWORK



(Start cutting path id: path3843)   CODIGO100CERA
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X1020.703367 Y242.918913

G01 Z-1.000000 F100.0(Penetrate)
G03 X1021.215184 Y241.176878 Z-1.000000 I4.694442 J0.433044 F400.000000
G03 X1022.000983 Y240.274615 Z-1.000000 I2.260850 J1.175704
G03 X1023.155611 Y239.775573 Z-1.000000 I1.751755 J2.467771
G03 X1024.737784 Y239.652920 Z-1.000000 I1.294396 J6.431175
G01 X1081.604044 Y242.150438 Z-1.000000
G03 X1083.938558 Y242.492735 Z-1.000000 I-0.510520 J11.613839
G03 X1084.536832 Y242.811551 Z-1.000000 I-0.345639 J1.369364
G03 X1084.836217 Y243.407610 Z-1.000000 I-0.611132 J0.680172
G03 X1084.870012 Y244.071600 Z-1.000000 I-8.024042 J0.741248
G01 X1085.254249 Y284.031664 Z-1.000000
G03 X1084.453652 Y288.530727 Z-1.000000 I-12.370482 J0.119467
G03 X1082.860555 Y290.955914 Z-1.000000 I-6.235917 J-2.360507
G03 X1080.302221 Y292.524087 Z-1.000000 I-4.779807 J-4.926897
G03 X1077.185389 Y293.061118 Z-1.000000 I-3.233718 J-9.454641
G01 X1044.333592 Y293.445355 Z-1.000000
G02 X1043.084087 Y293.946194 Z-1.000000 I0.021832 J1.863532
G02 X1042.130896 Y295.321436 Z-1.000000 I2.687657 J2.880790
G02 X1041.639622 Y297.114618 Z-1.000000 I6.907217 J2.856244
G02 X1041.497947 Y299.405054 Z-1.000000 I14.048374 J2.018559
G02 X1041.720656 Y301.697176 Z-1.000000 I14.831228 J-0.284162
G02 X1042.234803 Y303.486619 Z-1.000000 I8.728534 J-1.539320
G02 X1043.156991 Y304.924678 Z-1.000000 I4.298975 J-1.742098
G02 X1044.141522 Y305.356539 Z-1.000000 I1.000149 J-0.941910
G01 X1079.106600 Y305.740777 Z-1.000000
G01 X1079.298669 Y331.676401 Z-1.000000
G03 X1069.153341 Y333.227312 Z-1.000000 I-27.179269 J-143.835467
G03 X1055.626119 Y334.626514 Z-1.000000 I-58.345618 J-497.985630
G03 X1042.099248 Y335.750861 Z-1.000000 I-92.217691 J-1027.527102
G02 X1028.388014 Y336.863506 Z-1.000000 I113.541593 J1484.222987
G03 X1025.636700 Y329.841324 Z-1.000000 I72.292887 J-32.374667
G03 X1023.084592 Y321.014349 Z-1.000000 I98.185742 J-33.170485
G03 X1021.264772 Y311.994899 Z-1.000000 I117.668563 J-28.434852
G03 X1019.811150 Y300.628389 Z-1.000000 I166.935604 J-27.124995
G03 X1018.884475 Y277.914232 Z-1.000000 I215.509471 J-20.168171
G03 X1020.703317 Y242.918913 Z-1.000000 I436.325414 J5.132591
G01 X1020.703367 Y242.918913 Z-1.000000
G00 Z5.000000

(End cutting path id: path3843)


(Start cutting path id: path3841)
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X1009.752726 Y329.178883

G01 Z-1.000000 F100.0(Penetrate)
G01 X949.044138 Y329.563120 Z-1.000000 F400.000000
G03 X948.018437 Y329.297017 Z-1.000000 I0.461224 J-3.887645
G03 X947.440170 Y328.940583 Z-1.000000 I0.795967 J-1.938649
G03 X947.002773 Y328.406243 Z-1.000000 I1.340846 J-1.543773
G03 X946.732167 Y327.780439 Z-1.000000 I2.746199 J-1.558903
G03 X946.459999 Y326.355335 Z-1.000000 I6.241934 J-1.930631
G02 X946.162408 Y324.952371 Z-1.000000 I-5.449680 J0.422922
G01 X946.162408 Y309.006745 Z-1.000000
G03 X946.539312 Y307.182729 Z-1.000000 I5.894140 J0.266984
G03 X947.043380 Y306.349917 Z-1.000000 I2.514528 J0.952991
G03 X947.836402 Y305.689406 Z-1.000000 I2.581337 J2.292888
G02 X948.659926 Y305.164421 Z-1.000000 I-10.966190 J-18.110644
G02 X954.549711 Y305.995591 Z-1.000000 I17.600923 J-103.439078
G03 X960.188583 Y306.658173 Z-1.000000 I-61.887826 J551.018997
G03 X965.818466 Y307.471586 Z-1.000000 I-16.185099 J131.911872
G03 X973.202729 Y309.006745 Z-1.000000 I-15.446156 J92.824614
G03 X974.130462 Y307.979512 Z-1.000000 I7.878877 J6.183159
G02 X974.281920 Y307.740534 Z-1.000000 I-0.415055 J-0.430535
G02 X974.225293 Y307.534998 Z-1.000000 I-0.180184 J-0.060926
G02 X973.442754 Y306.893464 Z-1.000000 I-16.147373 J18.898430
G02 X966.059406 Y305.642155 Z-1.000000 I-22.784467 J112.031367
G03 X956.872251 Y304.351797 Z-1.000000 I141.536977 J-1041.072921
G03 X952.529149 Y303.687745 Z-1.000000 I40.568178 J-279.862842
G03 X949.120440 Y303.068726 Z-1.000000 I13.653628 J-84.880153
G03 X945.739968 Y302.251122 Z-1.000000 I9.202678 J-45.446772
G03 X944.145100 Y301.610275 Z-1.000000 I2.178513 J-7.726618
G03 X943.446838 Y276.575302 Z-1.000000 I302.161133 J-20.954935
G03 X945.009634 Y251.372035 Z-1.000000 I260.633398 J3.511180
G03 X945.675703 Y247.882402 Z-1.000000 I22.594334 J2.504215
G03 X946.416980 Y246.003353 Z-1.000000 I8.961430 J2.449508
G03 X947.598256 Y244.412958 Z-1.000000 I5.707792 J3.005602
G03 X948.757904 Y243.567428 Z-1.000000 I3.419860 J3.472353
G03 X951.631163 Y242.774148 Z-1.000000 I3.271838 J6.250483
G02 X954.615394 Y242.342582 Z-1.000000 I-0.979205 J-17.304696
G01 X1001.491658 Y239.460826 Z-1.000000
G03 X1003.440968 Y239.645760 Z-1.000000 I0.423166 J5.905474
G03 X1004.945498 Y240.525887 Z-1.000000 I-0.846538 J3.173132
G03 X1005.856372 Y242.015864 Z-1.000000 I-2.811812 J2.742371
G03 X1006.486635 Y245.416406 Z-1.000000 I-11.640694 J3.916185
G01 X1007.639372 Y283.455358 Z-1.000000
G03 X1006.829560 Y286.737376 Z-1.000000 I-14.492664 J-1.835031
G03 X1006.003806 Y288.057529 Z-1.000000 I-3.651071 J-1.365411
G03 X1004.707667 Y288.798692 Z-1.000000 I-1.884008 J-1.790816
G03 X1003.220726 Y288.834557 Z-1.000000 I-0.830142 J-3.575461
G01 X966.334460 Y288.066107 Z-1.000000
G02 X965.963007 Y289.748827 Z-1.000000 I50.309233 J11.987897
G02 X965.877594 Y291.025163 Z-1.000000 I4.699086 J0.955490
G02 X965.992186 Y291.614201 Z-1.000000 I2.397726 J-0.160790
G02 X966.233821 Y292.036958 Z-1.000000 I1.153341 J-0.378781
G02 X966.613912 Y292.316010 Z-1.000000 I0.824952 J-0.725268
G02 X967.295053 Y292.484762 Z-1.000000 I0.820262 J-1.851824
G01 X1009.944746 Y294.598043 Z-1.000000
G03 X1012.078848 Y294.730094 Z-1.000000 I-0.015290 J17.557850
G03 X1013.771725 Y295.294000 Z-1.000000 I-0.552962 J4.483033
G03 X1014.469275 Y295.824959 Z-1.000000 I-1.677887 J2.928017
G03 X1015.072744 Y296.623046 Z-1.000000 I-2.479457 J2.502025
G03 X1015.452401 Y297.555824 Z-1.000000 I-3.731724 J2.062529
G03 X1015.708232 Y299.016698 Z-1.000000 I-7.553724 J2.075660
G01 X1014.747639 Y318.036198 Z-1.000000
G03 X1013.928552 Y323.203782 Z-1.000000 I-88.045377 J-11.306899
G03 X1013.011325 Y326.462067 Z-1.000000 I-18.980240 J-3.584809
G03 X1012.304522 Y327.939018 Z-1.000000 I-9.102375 J-3.448396
G03 X1011.632193 Y328.758833 Z-1.000000 I-2.936951 J-1.722991
G03 X1010.770595 Y329.210115 Z-1.000000 I-1.332847 J-1.496574
G03 X1009.752603 Y329.178883 Z-1.000000 I-0.452139 J-1.868864
G01 X1009.752726 Y329.178883 Z-1.000000
G00 Z5.000000

(End cutting path id: path3841)


(Start cutting path id: path3839)
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X879.065184 Y335.874312

G01 Z-1.000000 F100.0(Penetrate)
G03 X873.382501 Y313.760295 Z-1.000000 I141.264993 J-48.088301 F400.000000
G03 X870.500075 Y289.490472 Z-1.000000 I212.199762 J-37.508160
G03 X870.154653 Y264.946767 Z-1.000000 I258.534700 J-15.912845
G03 X871.865298 Y239.830924 Z-1.000000 I362.836377 J12.096680
G03 X905.129803 Y240.323745 Z-1.000000 I2.283114 J968.788009
G03 X936.120673 Y242.140353 Z-1.000000 I-18.687472 J584.059933
G01 X936.120673 Y283.980990 Z-1.000000
G03 X935.634631 Y287.381104 Z-1.000000 I-32.263730 J-2.877260
G03 X934.560980 Y289.852119 Z-1.000000 I-6.198393 J-1.224433
G03 X933.636952 Y290.817296 Z-1.000000 I-3.914653 J-2.822853
G03 X932.238286 Y291.630790 Z-1.000000 I-3.554554 J-4.502322
G03 X930.661022 Y292.077760 Z-1.000000 I-2.828027 J-6.973122
G03 X928.145484 Y292.292347 Z-1.000000 I-2.564962 J-15.216518
G01 X893.464926 Y292.403584 Z-1.000000
G02 X893.201224 Y292.580738 Z-1.000000 I0.000696 J0.285879
G02 X892.774690 Y294.041452 Z-1.000000 I6.648898 J2.734137
G02 X892.580699 Y295.604396 Z-1.000000 I14.205618 J2.556700
G02 X892.543879 Y297.856150 Z-1.000000 I21.385304 J1.475857
G02 X892.747461 Y300.077184 Z-1.000000 I20.198693 J-0.731584
G02 X893.249050 Y302.204878 Z-1.000000 I12.750049 J-1.882762
G02 X894.141788 Y304.130739 Z-1.000000 I8.593767 J-2.813816
G02 X895.366757 Y305.444837 Z-1.000000 I3.877279 J-2.386312
G01 X930.686804 Y305.716576 Z-1.000000
G01 X930.686804 Y332.342488 Z-1.000000
G03 X904.588509 Y334.387181 Z-1.000000 I-87.112694 J-944.318648
G03 X879.065184 Y335.874510 Z-1.000000 I-133.174312 J-2065.599088
G01 X879.065184 Y335.874312 Z-1.000000
G00 Z5.000000

(End cutting path id: path3839)


(Start cutting path id: path12)
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X981.770055 Y125.304992

G01 Z-1.000000 F100.0(Penetrate)
G02 X977.585089 Y125.930057 Z-1.000000 I0.387788 J16.918552 F400.000000
G02 X973.872770 Y127.336923 Z-1.000000 I6.059064 J21.589454
G02 X970.355340 Y129.332232 Z-1.000000 I13.541316 J27.969308
G02 X966.952092 Y131.666597 Z-1.000000 I56.356391 J85.809404
G03 X963.585355 Y134.069667 Z-1.000000 I-853.967046 J-1192.860818
G03 X960.216809 Y136.239583 Z-1.000000 I-25.135743 J-35.320756
G03 X956.671210 Y137.971666 Z-1.000000 I-13.844909 J-23.845741
G03 X952.875709 Y139.001451 Z-1.000000 I-6.246062 J-15.511784
G03 X948.494824 Y139.096171 Z-1.000000 I-2.565956 J-17.320465
G03 X943.986508 Y138.296219 Z-1.000000 I3.309235 J-31.753810
G03 X939.530448 Y136.955330 Z-1.000000 I12.689562 J-50.244753
G02 X935.031980 Y135.380352 Z-1.000000 I-1828.543224 J5215.490389
G02 X930.511445 Y133.913620 Z-1.000000 I-34.701247 J99.251225
G02 X926.210899 Y132.960317 Z-1.000000 I-8.668297 J28.927519
G02 X921.926438 Y132.791312 Z-1.000000 I-2.938566 J20.103431
G02 X917.722035 Y133.742588 Z-1.000000 I0.917706 J13.822882
G02 X914.282162 Y135.797663 Z-1.000000 I4.655479 J11.698982
G02 X911.510521 Y138.681549 Z-1.000000 I12.049068 J14.353925
G02 X909.236388 Y142.109945 Z-1.000000 I22.319224 J17.273289
G02 X907.140069 Y146.025116 Z-1.000000 I91.534056 J51.529331
G03 X905.115557 Y149.990441 Z-1.000000 I-1236.999363 J-629.055307
G03 X902.959547 Y153.734717 Z-1.000000 I-39.420877 J-20.206265
G03 X900.397172 Y157.115588 Z-1.000000 I-23.104361 J-14.849428
G03 X897.317822 Y159.771780 Z-1.000000 I-11.451580 J-10.162879
G03 X893.333121 Y161.675300 Z-1.000000 I-9.111491 J-13.950969
G03 X888.856207 Y162.749655 Z-1.000000 I-9.180216 J-28.389624
G03 X884.195405 Y163.262076 Z-1.000000 I-7.577461 J-47.469235
G03 X879.371821 Y163.547028 Z-1.000000 I-85.924507 J-1413.535134
G02 X874.566518 Y163.923967 Z-1.000000 I6.578523 J114.682393
G02 X870.163710 Y164.688177 Z-1.000000 I3.052709 J30.652372
G02 X866.082694 Y166.181669 Z-1.000000 I4.953648 J19.858507
G02 X862.530924 Y168.697382 Z-1.000000 I6.064023 J12.326510
G02 X860.108533 Y171.966322 Z-1.000000 I9.067398 J9.251242
G02 X858.625987 Y175.742268 Z-1.000000 I17.019142 J8.861229
G02 X857.815594 Y179.832132 Z-1.000000 I27.705123 J7.614895
G02 X857.360221 Y184.305439 Z-1.000000 I103.566489 J12.802683
G03 X856.996250 Y188.798922 Z-1.000000 I-2441.022991 J-195.460819
G03 X856.431990 Y193.147661 Z-1.000000 I-46.229562 J-3.787420
G03 X855.359999 Y197.321686 Z-1.000000 I-28.283909 J-5.039327
G03 X853.539659 Y201.029703 Z-1.000000 I-15.103012 J-5.113545
G03 X850.860206 Y203.790367 Z-1.000000 I-8.747575 J-5.809598
G03 X847.851567 Y205.435658 Z-1.000000 I-8.584070 J-12.123642
G03 X844.453611 Y206.573428 Z-1.000000 I-10.916105 J-26.958112
G02 X840.741930 Y207.693457 Z-1.000000 I15.857143 J59.259323
G02 X837.244770 Y209.161243 Z-1.000000 I8.327698 J24.741727
G02 X833.662746 Y211.659917 Z-1.000000 I7.138999 J14.051131
G02 X830.934850 Y214.976597 Z-1.000000 I10.899423 J11.744693
G02 X828.066015 Y221.191854 Z-1.000000 I24.028593 J14.860831
G02 X826.611978 Y227.633711 Z-1.000000 I35.367230 J11.368015
G02 X826.390718 Y232.237853 Z-1.000000 I26.914882 J3.600826
G02 X826.947776 Y236.850241 Z-1.000000 I27.921349 J-1.032347
G02 X827.872116 Y240.595475 Z-1.000000 I49.025764 J-10.113087
G02 X828.980222 Y244.312080 Z-1.000000 I201.464185 J-58.043092
G03 X829.832711 Y247.791570 Z-1.000000 I-29.158887 J8.988218
G03 X830.143471 Y251.277012 Z-1.000000 I-18.608742 J3.415721
G03 X829.595007 Y255.352985 Z-1.000000 I-15.853932 J-0.058424
G03 X827.938562 Y259.105493 Z-1.000000 I-13.087394 J-3.535231
G03 X825.572961 Y262.145833 Z-1.000000 I-17.366643 J-11.072009
G03 X822.704491 Y264.878508 Z-1.000000 I-26.362843 J-24.801053
G02 X819.550911 Y267.648332 Z-1.000000 I95.271845 J111.651801
G02 X816.600804 Y270.573448 Z-1.000000 I30.198312 J33.406530
G02 X813.951129 Y274.141916 Z-1.000000 I15.636527 J14.378467
G02 X812.193631 Y278.119606 Z-1.000000 I15.919876 J9.411131
G02 X811.195885 Y283.908018 Z-1.000000 I22.075940 J6.785415
G02 X811.929697 Y290.021073 Z-1.000000 I29.179253 J-0.402110
G02 X813.295058 Y293.999233 Z-1.000000 I19.473760 J-4.460284
G02 X815.485026 Y297.652199 Z-1.000000 I18.597599 J-8.666410
G02 X817.822332 Y300.454025 Z-1.000000 I28.495135 J-21.395062
G02 X820.375950 Y303.115785 Z-1.000000 I74.432710 J-68.852808
G03 X822.647884 Y305.680051 Z-1.000000 I-21.009676 J20.903128
G03 X824.439820 Y308.482919 Z-1.000000 I-13.305783 J10.480930
G03 X825.641886 Y312.084969 Z-1.000000 I-11.448146 J5.822043
G03 X825.834908 Y316.497011 Z-1.000000 I-16.777142 J2.944227
G03 X825.158328 Y320.326496 Z-1.000000 I-21.894541 J-1.893743
G03 X823.911034 Y324.102935 Z-1.000000 I-34.587367 J-9.329426
G02 X822.388166 Y328.080482 Z-1.000000 I222.896269 J87.619737
G02 X821.058607 Y332.101044 Z-1.000000 I53.078557 J19.782657
G02 X820.093158 Y336.906724 Z-1.000000 I28.149917 J8.155064
G02 X819.969544 Y341.754908 Z-1.000000 I27.089901 J3.116378
G02 X821.035065 Y348.365016 Z-1.000000 I34.758148 J-2.211930
G02 X822.691165 Y351.811725 Z-1.000000 I9.701491 J-2.540223
G02 X825.180138 Y354.323455 Z-1.000000 I10.085572 J-7.505120
G02 X828.330517 Y356.262100 Z-1.000000 I11.516180 J-15.185198
G02 X832.049955 Y357.938169 Z-1.000000 I28.823135 J-58.997678
G02 X835.845657 Y359.494543 Z-1.000000 I247.033528 J-597.061435
G03 X839.412136 Y361.175136 Z-1.000000 I-13.646570 J33.584743
G03 X842.610164 Y363.297492 Z-1.000000 I-10.011677 J18.556506
G03 X845.034302 Y366.000333 Z-1.000000 I-7.023118 J8.737438
G03 X846.613603 Y369.840506 Z-1.000000 I-9.867021 J6.302724
G03 X846.929638 Y373.369276 Z-1.000000 I-14.162634 J3.046937
G03 X846.548345 Y377.054107 Z-1.000000 I-26.527164 J-0.882793
G02 X845.934597 Y381.110848 Z-1.000000 I88.640114 J15.485229
G02 X845.669132 Y385.140797 Z-1.000000 I32.550502 J4.167921
G02 X846.235522 Y389.919897 Z-1.000000 I19.745304 J0.083017
G02 X847.903833 Y394.361444 Z-1.000000 I19.379987 J-4.745317
G02 X852.018755 Y400.491269 Z-1.000000 I29.134461 J-15.111745
G02 X855.977499 Y404.048312 Z-1.000000 I15.084980 J-12.807122
G02 X859.409701 Y405.789991 Z-1.000000 I8.507745 J-12.512959
G02 X863.254603 Y406.683076 Z-1.000000 I5.828201 J-16.368476
G02 X866.949868 Y406.922295 Z-1.000000 I4.249917 J-36.988981
G02 X870.697879 Y406.920576 Z-1.000000 I1.818047 J-122.027135
G03 X874.197312 Y407.074665 Z-1.000000 I0.463291 J29.292418
G03 X877.509556 Y407.772270 Z-1.000000 I-1.693518 J16.252960
G03 X880.710088 Y409.433587 Z-1.000000 I-3.313816 J10.297651
G03 X883.242293 Y411.927753 Z-1.000000 I-8.156402 J10.813294
G03 X885.122655 Y414.814460 Z-1.000000 I-16.195994 J12.605623
G03 X886.616170 Y417.996893 Z-1.000000 I-28.403936 J15.271632
G03 X887.993529 Y421.426722 Z-1.000000 I-718.184369 J290.401815
G02 X889.434085 Y424.820021 Z-1.000000 I77.871577 J-31.056350
G02 X891.216901 Y428.133063 Z-1.000000 I24.443957 J-11.017591
G02 X893.466560 Y431.062439 Z-1.000000 I17.174862 J-10.861176
G02 X896.686957 Y433.796172 Z-1.000000 I12.847020 J-11.870358
G02 X902.077741 Y436.365056 Z-1.000000 I11.937051 J-18.109104
G02 X907.338106 Y437.429263 Z-1.000000 I7.584753 J-23.958298
G02 X912.829495 Y437.484648 Z-1.000000 I3.080435 J-33.161513
G02 X918.490129 Y436.879220 Z-1.000000 I-6.045033 J-83.285417
G02 X924.150235 Y436.012483 Z-1.000000 I-35.495073 J-250.710374
G03 X929.753592 Y435.282997 Z-1.000000 I13.589272 J82.497350
G03 X935.332047 Y435.075763 Z-1.000000 I4.557565 J47.497580
G03 X940.739060 Y435.777550 Z-1.000000 I-0.542754 J25.362138
G03 X945.504363 Y437.442658 Z-1.000000 I-5.964233 J24.720160
G03 X949.863425 Y439.785745 Z-1.000000 I-18.124687 J38.945355
G03 X954.034646 Y442.550783 Z-1.000000 I-37.934338 J61.754951
G02 X958.224845 Y445.483325 Z-1.000000 I160.070740 J-224.259201
G02 X962.508695 Y448.233698 Z-1.000000 I47.076336 J-68.612508
G02 X967.152195 Y450.580818 Z-1.000000 I18.558946 J-30.949822
G02 X972.038430 Y452.134056 Z-1.000000 I11.237600 J-26.889380
G02 X977.974352 Y452.788753 Z-1.000000 I5.935921 J-26.582184
G02 X982.823335 Y452.280362 Z-1.000000 I0.000044 J-23.378344
G02 X987.395739 Y450.959133 Z-1.000000 I-6.826522 J-32.197190
G02 X991.819149 Y449.051155 Z-1.000000 I-15.581851 J-42.206137
G02 X996.212105 Y446.766113 Z-1.000000 I-58.384765 J-117.608992
G02 X1000.564189 Y444.383854 Z-1.000000 I-509.136600 J-935.294283
G03 X1004.931457 Y442.153512 Z-1.000000 I38.151552 J69.314169
G03 X1009.428324 Y440.293116 Z-1.000000 I21.254203 J45.009731
G03 X1014.061801 Y439.065147 Z-1.000000 I9.582882 J26.803240
G03 X1018.723854 Y438.798598 Z-1.000000 I3.458568 J19.587857
G03 X1023.239864 Y439.372506 Z-1.000000 I-2.092459 J34.520200
G03 X1027.742934 Y440.421803 Z-1.000000 I-12.051847 J61.907757
G02 X1032.337784 Y441.583232 Z-1.000000 I34.528897 J-126.933617
G02 X1036.954415 Y442.446994 Z-1.000000 I11.983775 J-51.281480
G02 X1041.812023 Y442.666849 Z-1.000000 I3.580948 J-25.346237
G02 X1046.521355 Y441.909203 Z-1.000000 I-1.087241 J-21.772829
G02 X1052.119045 Y439.592880 Z-1.000000 I-6.553432 J-23.759139
G02 X1056.063200 Y436.688947 Z-1.000000 I-10.161693 J-17.932195
G02 X1058.547854 Y433.823035 Z-1.000000 I-12.649707 J-13.476900
G02 X1060.479006 Y430.474950 Z-1.000000 I-17.944642 J-12.581329
G02 X1061.875964 Y427.172911 Z-1.000000 I-46.917288 J-21.795293
G02 X1063.147225 Y423.797362 Z-1.000000 I-312.774477 J-119.720754
G03 X1064.504819 Y420.706068 Z-1.000000 I26.839723 J9.943370
G03 X1066.272970 Y417.933887 Z-1.000000 I15.439505 J7.897643
G03 X1068.835865 Y415.486067 Z-1.000000 I9.624842 J7.511720
G03 X1072.708155 Y413.880692 Z-1.000000 I5.463137 J7.704697
G03 X1076.372019 Y413.777986 Z-1.000000 I2.193165 J12.834994
G03 X1080.224951 Y414.487649 Z-1.000000 I-3.211094 J28.247897
G02 X1084.455468 Y415.366289 Z-1.000000 I11.592282 J-45.191083
G02 X1088.611853 Y415.623612 Z-1.000000 I3.405601 J-21.312162
G02 X1093.458217 Y414.607232 Z-1.000000 I-0.503423 J-14.462999
G02 X1095.736042 Y413.481914 Z-1.000000 I-5.861141 J-14.731898
G02 X1098.420581 Y411.583773 Z-1.000000 I-11.058718 J-18.487771
G02 X1100.826039 Y409.324276 Z-1.000000 I-16.912111 J-20.414787
G02 X1103.752270 Y405.857070 Z-1.000000 I-30.966677 J-29.103507
G02 X1105.844144 Y402.042736 Z-1.000000 I-11.928910 J-9.022888
G02 X1106.813153 Y398.248231 Z-1.000000 I-17.404821 J-6.465683
G02 X1107.083188 Y394.235684 Z-1.000000 I-25.810795 J-3.752360
G02 X1106.957008 Y390.175017 Z-1.000000 I-96.329577 J0.961002
G03 X1106.768491 Y386.107041 Z-1.000000 I348.046039 J-18.167468
G03 X1106.861908 Y382.301867 Z-1.000000 I29.270876 J-1.185134
G03 X1107.564578 Y378.671970 Z-1.000000 I18.143446 J1.629222
G03 X1109.205924 Y375.293221 Z-1.000000 I11.333069 J3.417392
G03 X1111.509938 Y372.901476 Z-1.000000 I8.797248 J6.168936
G03 X1114.548179 Y371.003923 Z-1.000000 I11.958470 J15.766048
G03 X1117.912691 Y369.545397 Z-1.000000 I14.114100 J27.948367
G03 X1121.516322 Y368.273603 Z-1.000000 I58.660482 J160.473371
G02 X1125.122750 Y367.003957 Z-1.000000 I-59.174317 J-173.841390
G02 X1128.196858 Y365.660221 Z-1.000000 I-9.083761 J-24.969448
G02 X1130.906548 Y363.875291 Z-1.000000 I-6.936927 J-13.480145
G02 X1132.676289 Y361.721738 Z-1.000000 I-4.618541 J-5.599355
G02 X1135.242090 Y355.147442 Z-1.000000 I-25.401855 J-13.701612
G02 X1135.884963 Y350.200972 Z-1.000000 I-20.140187 J-5.132553
G02 X1135.358195 Y345.143299 Z-1.000000 I-22.600663 J-0.202356
G02 X1134.112238 Y340.657967 Z-1.000000 I-41.774734 J9.188662
G02 X1132.509345 Y336.221872 Z-1.000000 I-104.746478 J35.340387
G03 X1131.110256 Y331.966695 Z-1.000000 I41.474124 J-15.994157
G03 X1130.341226 Y327.675603 Z-1.000000 I23.410750 J-6.410024
G03 X1130.631160 Y323.001126 Z-1.000000 I15.677369 J-1.373846
G03 X1132.234899 Y318.261952 Z-1.000000 I21.315290 J4.572167
G03 X1134.727149 Y313.785285 Z-1.000000 I34.764895 J16.422239
G03 X1137.731222 Y309.545439 Z-1.000000 I52.233629 J33.825107
G02 X1140.974143 Y305.333563 Z-1.000000 I-1026.907320 J-794.018518
G02 X1144.098590 Y301.053554 Z-1.000000 I-89.430972 J-68.565897
G02 X1146.689563 Y296.701507 Z-1.000000 I-29.704370 J-20.631659
G02 X1148.483559 Y292.080419 Z-1.000000 I-22.947009 J-11.567253
G02 X1149.190831 Y286.944665 Z-1.000000 I-18.292644 J-5.135754
G02 X1148.543169 Y282.112038 Z-1.000000 I-18.353252 J-0.000034
G02 X1146.891378 Y277.722341 Z-1.000000 I-24.244904 J6.617444
G02 X1144.508599 Y273.571640 Z-1.000000 I-31.760935 J15.473613
G02 X1141.636405 Y269.476636 Z-1.000000 I-91.460699 J61.094829
G02 X1138.655435 Y265.441971 Z-1.000000 I-1051.116703 J773.487746
G03 X1135.891043 Y261.381713 Z-1.000000 I48.729571 J-36.148309
G03 X1133.597483 Y257.110687 Z-1.000000 I31.941159 J-19.903877
G03 X1132.120425 Y252.611735 Z-1.000000 I19.418894 J-8.867392
G03 X1131.683470 Y248.368444 Z-1.000000 I20.211856 J-4.225469
G03 X1131.955895 Y244.633471 Z-1.000000 I26.047864 J0.022481
G03 X1132.681008 Y240.905537 Z-1.000000 I39.696919 J5.786880
G02 X1133.618890 Y236.997223 Z-1.000000 I-494.992685 J-120.850530
G02 X1134.433103 Y233.071186 Z-1.000000 I-72.053807 J-16.990548
G02 X1134.963918 Y228.414683 Z-1.000000 I-36.120362 J-6.476021
G02 X1134.874430 Y223.756906 Z-1.000000 I-33.834682 J-1.679695
G02 X1133.845489 Y217.597540 Z-1.000000 I-40.172412 J3.545296
G02 X1132.065984 Y213.611715 Z-1.000000 I-12.039095 J2.984805
G02 X1129.240524 Y210.390179 Z-1.000000 I-14.420071 J9.797368
G02 X1125.688647 Y207.727708 Z-1.000000 I-17.519976 J19.672147
G02 X1121.458145 Y205.243889 Z-1.000000 I-46.471142 J74.306114
G02 X1117.131633 Y202.886853 Z-1.000000 I-291.619239 J530.139259
G03 X1112.986427 Y200.410022 Z-1.000000 I26.164425 J-48.495692
G03 X1109.223072 Y197.519233 Z-1.000000 I17.995248 J-27.322035
G03 X1106.313441 Y194.139930 Z-1.000000 I11.237925 J-12.618297
G03 X1104.560861 Y190.484567 Z-1.000000 I12.949035 J-8.456292
G03 X1103.588840 Y186.502386 Z-1.000000 I25.583198 J-8.354397
G03 X1103.101927 Y182.354061 Z-1.000000 I44.483240 J-7.323995
G02 X1102.774111 Y178.054689 Z-1.000000 I-384.042482 J27.120041
G02 X1102.292347 Y173.792602 Z-1.000000 I-69.280751 J5.672862
G02 X1101.350744 Y169.734360 Z-1.000000 I-24.405793 J3.524330
G02 X1099.681044 Y166.035728 Z-1.000000 I-17.353927 J5.608010
G02 X1096.800230 Y162.478917 Z-1.000000 I-13.598784 J8.069190
G02 X1093.642645 Y160.577480 Z-1.000000 I-6.203248 J6.728792
G02 X1089.555870 Y159.657784 Z-1.000000 I-5.925374 J16.790213
G02 X1085.205210 Y159.515340 Z-1.000000 I-3.180782 J30.638207
G02 X1080.192032 Y159.784077 Z-1.000000 I6.124768 J161.148804
G03 X1075.177828 Y160.098486 Z-1.000000 I-26.694739 J-385.587385
G03 X1070.308004 Y160.111973 Z-1.000000 I-2.561180 J-45.587614
G03 X1065.584089 Y159.471068 Z-1.000000 I1.431939 J-28.284142
G03 X1061.503072 Y157.895649 Z-1.000000 I3.196365 J-14.353473
G03 X1058.846861 Y155.781929 Z-1.000000 I5.460327 J-9.587560
G03 X1056.805762 Y153.092293 Z-1.000000 I13.086845 J-12.050579
G03 X1055.159111 Y150.039886 Z-1.000000 I24.838959 J-15.369972
G03 X1053.614855 Y146.712027 Z-1.000000 I1968.235177 J-915.359923
G02 X1051.992651 Y143.437939 Z-1.000000 I-62.822658 J29.087662
G02 X1050.056820 Y140.419971 Z-1.000000 I-18.906285 J9.997319
G02 X1047.638420 Y137.911230 Z-1.000000 I-12.281204 J9.418923
G02 X1044.258123 Y135.881245 Z-1.000000 I-8.401026 J10.159851
G02 X1040.317746 Y134.891040 Z-1.000000 I-5.372505 J13.043921
G02 X1036.240038 Y134.866376 Z-1.000000 I-2.167394 J21.239353
G02 X1032.117462 Y135.509824 Z-1.000000 I2.863675 J31.876023
G02 X1027.750518 Y136.566271 Z-1.000000 I28.352772 J126.753258
G03 X1023.391157 Y137.681865 Z-1.000000 I-137.430948 J-527.957883
G03 X1018.999763 Y138.568095 Z-1.000000 I-11.562645 J-45.971412
G03 X1014.599843 Y138.925217 Z-1.000000 I-4.715049 J-30.808585
G03 X1010.197971 Y138.459017 Z-1.000000 I-0.193967 J-19.182937
G03 X1002.235530 Y133.846693 Z-1.000000 I131.704069 J-236.545175
G03 X997.995021 Y131.221731 Z-1.000000 I104.900239 J-174.199322
G03 X993.885761 Y128.561600 Z-1.000000 I165.048596 J-259.464113
G02 X991.724891 Y127.215307 Z-1.000000 I-27.695506 J42.045438
G02 X989.533500 Y126.121293 Z-1.000000 I-9.026529 J15.339036
G02 X987.584322 Y125.542231 Z-1.000000 I-3.711514 J8.923230
G02 X985.513687 Y125.319675 Z-1.000000 I-2.490729 J13.429688
G02 X981.770055 Y125.304992 Z-1.000000 I-2.145299 J69.718034
G01 X981.770055 Y125.304992 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X979.375839 Y150.920280

G01 Z-1.000000 F100.0(Penetrate)
G03 X986.150725 Y151.089605 Z-1.000000 I0.001631 J135.555452 F400.000000
G03 X992.908685 Y151.597308 Z-1.000000 I-6.773623 J135.393484
G03 X999.633345 Y152.442124 Z-1.000000 I-13.529587 J134.880689
G03 X1006.307395 Y153.621997 Z-1.000000 I-20.254204 J134.035595
G03 X1012.913287 Y155.134472 Z-1.000000 I-26.941596 J132.852478
G03 X1019.435314 Y156.975195 Z-1.000000 I-33.546911 J131.337744
G03 X1025.858084 Y159.139001 Z-1.000000 I-40.052491 J129.500953
G03 X1032.164659 Y161.621119 Z-1.000000 I-46.476682 J127.341041
G03 X1038.338605 Y164.415569 Z-1.000000 I-52.793719 J124.857832
G03 X1044.365183 Y167.515081 Z-1.000000 I-58.966444 J122.060854
G03 X1050.229510 Y170.911626 Z-1.000000 I-64.983847 J118.959112
G03 X1055.916775 Y174.597022 Z-1.000000 I-70.854551 J115.572969
G03 X1061.413163 Y178.562619 Z-1.000000 I-76.541269 J111.879414
G03 X1066.704465 Y182.797983 Z-1.000000 I-82.037206 J107.913225
G03 X1071.776904 Y187.291841 Z-1.000000 I-87.325616 J103.678408
G03 X1076.618426 Y192.033620 Z-1.000000 I-92.391387 J99.177392
G03 X1081.217267 Y197.011631 Z-1.000000 I-97.246302 J94.452434
G03 X1085.561560 Y202.213264 Z-1.000000 I-101.831383 J89.462335
G03 X1089.640161 Y207.625470 Z-1.000000 I-106.195673 J84.271228
G03 X1093.443171 Y213.234750 Z-1.000000 I-110.251930 J78.843043
G03 X1096.961201 Y219.027777 Z-1.000000 I-114.073879 J73.240321
G03 X1100.185295 Y224.989436 Z-1.000000 I-117.580273 J67.440620
G03 X1103.107424 Y231.104160 Z-1.000000 I-120.802530 J61.485181
G03 X1105.720300 Y237.357310 Z-1.000000 I-123.735063 J55.375114
G03 X1108.017531 Y243.733063 Z-1.000000 I-126.328784 J49.118923
G03 X1109.993235 Y250.215690 Z-1.000000 I-128.641868 J42.748438
G03 X1111.642237 Y256.789163 Z-1.000000 I-130.609624 J36.257921
G03 X1112.960643 Y263.436825 Z-1.000000 I-132.276344 J29.688425
G03 X1113.945084 Y270.141962 Z-1.000000 I-133.571806 J23.035706
G03 X1114.593181 Y276.887925 Z-1.000000 I-134.581060 J16.333580
G03 X1114.847970 Y281.679674 Z-1.000000 I-135.166723 J9.589782
G03 X1114.932973 Y286.477420 Z-1.000000 I-135.472523 J4.799841
G03 X1114.763637 Y293.252306 Z-1.000000 I-135.551525 J0.001483
G03 X1114.255936 Y300.010267 Z-1.000000 I-135.409197 J-6.774755
G03 X1113.411087 Y306.734302 Z-1.000000 I-134.888755 J-13.533143
G03 X1112.231249 Y313.407701 Z-1.000000 I-134.027862 J-20.254754
G03 X1110.718806 Y320.014219 Z-1.000000 I-132.846080 J-26.936334
G03 X1108.878047 Y326.536881 Z-1.000000 I-131.344221 J-33.545555
G03 X1106.714209 Y332.959008 Z-1.000000 I-129.512863 J-40.061785
G03 X1104.232122 Y339.264964 Z-1.000000 I-127.333125 J-46.478109
G03 X1101.437706 Y345.439552 Z-1.000000 I-124.863407 J-52.789470
G03 X1098.338168 Y351.466749 Z-1.000000 I-122.056417 J-58.957983
G03 X1094.941635 Y357.331076 Z-1.000000 I-118.966398 J-64.987827
G03 X1091.256252 Y363.018341 Z-1.000000 I-115.565890 J-70.849717
G03 X1087.290639 Y368.514085 Z-1.000000 I-111.864540 J-76.540456
G03 X1083.055291 Y373.804771 Z-1.000000 I-107.913024 J-82.046918
G03 X1078.561422 Y378.877870 Z-1.000000 I-103.691429 J-87.325342
G03 X1073.819629 Y383.720017 Z-1.000000 I-99.189044 J-92.390509
G03 X1068.841602 Y388.318204 Z-1.000000 I-94.446516 J-97.254692
G03 X1063.639985 Y392.661865 Z-1.000000 I-89.451638 J-101.833694
G03 X1058.227162 Y396.740469 Z-1.000000 I-84.257679 J-106.189527
G03 X1052.617238 Y400.543452 Z-1.000000 I-78.848595 J-110.273359
G03 X1046.824809 Y404.061495 Z-1.000000 I-73.241090 J-114.063214
G03 X1040.863813 Y407.285625 Z-1.000000 I-67.449461 J-117.582652
G03 X1034.749085 Y410.207758 Z-1.000000 I-61.485692 J-120.803534
G03 X1028.495964 Y412.820630 Z-1.000000 I-55.376709 J-123.738503
G03 X1022.120197 Y415.117841 Z-1.000000 I-49.125028 J-126.347134
G03 X1015.637584 Y417.093540 Z-1.000000 I-42.748345 J-128.641589
G03 X1009.064096 Y418.742556 Z-1.000000 I-36.260073 J-130.617375
G03 X1002.416433 Y420.060973 Z-1.000000 I-29.687346 J-132.269780
G03 X995.711317 Y421.046053 Z-1.000000 I-23.049450 J-133.577674
G03 X988.965334 Y421.694772 Z-1.000000 I-16.343909 J-134.559263
G03 X984.173579 Y421.948890 Z-1.000000 I-9.576924 J-135.281907
G03 X979.375839 Y422.033279 Z-1.000000 I-4.782034 J-135.447208
G03 X972.600937 Y421.863946 Z-1.000000 I-0.001483 J-135.554750
G03 X965.842977 Y421.356266 Z-1.000000 I6.775169 J-135.420392
G03 X959.118948 Y420.511406 Z-1.000000 I13.532195 J-134.879308
G03 X952.445519 Y419.331555 Z-1.000000 I20.257161 J-134.040645
G03 X945.839019 Y417.819757 Z-1.000000 I26.926133 J-132.857189
G03 X939.316339 Y415.979662 Z-1.000000 I33.533118 J-131.347293
G03 X932.894202 Y413.815192 Z-1.000000 I40.073134 J-129.509540
G03 X926.588257 Y411.332452 Z-1.000000 I46.487942 J-127.325000
G03 X920.413685 Y408.538657 Z-1.000000 I52.775952 J-124.860357
G03 X914.386465 Y405.439759 Z-1.000000 I58.953444 J-122.072605
G03 X908.522109 Y402.042578 Z-1.000000 I64.992914 J-118.953751
G03 X902.834873 Y398.356532 Z-1.000000 I70.860281 J-115.561481
G03 X897.339147 Y394.391541 Z-1.000000 I76.545366 J-111.887869
G03 X892.048446 Y390.156857 Z-1.000000 I82.032887 J-107.912083
G03 X886.975341 Y385.662382 Z-1.000000 I87.325718 J-103.678606
G03 X882.133217 Y380.919934 Z-1.000000 I92.396567 J-99.181709
G03 X877.535028 Y375.941895 Z-1.000000 I97.243112 J-94.435642
G03 X873.191366 Y370.740265 Z-1.000000 I101.845781 J-89.461536
G03 X869.112131 Y365.328104 Z-1.000000 I106.177873 J-84.271402
G03 X865.308489 Y359.718804 Z-1.000000 I110.253747 J-78.856837
G03 X861.791110 Y353.926373 Z-1.000000 I114.076359 J-73.235559
G03 X858.567614 Y347.965404 Z-1.000000 I117.586725 J-67.439105
G03 X855.644837 Y341.850698 Z-1.000000 I120.800010 J-61.497250
G03 X853.031328 Y335.597530 Z-1.000000 I123.725394 J-55.383712
G03 X850.734719 Y329.221125 Z-1.000000 I126.344322 J-49.107617
G03 X848.759679 Y322.737890 Z-1.000000 I128.646551 J-42.733079
G03 X847.110669 Y316.164404 Z-1.000000 I130.615705 J-36.259535
G03 X845.792261 Y309.516729 Z-1.000000 I132.267786 J-29.686739
G03 X844.807191 Y302.811614 Z-1.000000 I133.581744 J-23.049855
G03 X844.158464 Y296.065629 Z-1.000000 I134.568350 J-16.344926
G03 X843.904355 Y291.274517 Z-1.000000 I135.228793 J-9.574499
G03 X843.819950 Y286.477420 Z-1.000000 I135.467774 J-4.782852
G03 X843.989237 Y279.701892 Z-1.000000 I135.554673 J-0.003040
G03 X844.496971 Y272.943288 Z-1.000000 I135.391915 J6.772815
G03 X845.341820 Y266.219253 Z-1.000000 I134.879264 J13.531968
G03 X846.521665 Y259.545854 Z-1.000000 I134.032924 J20.255778
G03 X848.033473 Y252.939337 Z-1.000000 I132.851968 J26.925073
G03 X849.873582 Y246.416674 Z-1.000000 I131.344238 J33.532630
G03 X852.037419 Y239.994530 Z-1.000000 I129.498616 J40.056853
G03 X854.519506 Y233.688590 Z-1.000000 I127.337755 J46.480083
G03 X857.313954 Y227.514628 Z-1.000000 I124.850929 J52.790413
G03 X860.413463 Y221.488066 Z-1.000000 I122.058916 J58.965551
G03 X863.810644 Y215.623721 Z-1.000000 I118.962475 J64.998106
G03 X867.496687 Y209.936474 Z-1.000000 I115.562421 J70.860689
G03 X871.461650 Y204.440103 Z-1.000000 I111.884079 J76.532438
G03 X875.696362 Y199.148784 Z-1.000000 I107.917358 J82.027473
G03 X880.190880 Y194.076309 Z-1.000000 I103.671708 J87.331921
G03 X884.933295 Y189.234823 Z-1.000000 I99.178938 J92.406024
G03 X889.911310 Y184.635994 Z-1.000000 I94.444334 J97.237888
G03 X895.112949 Y180.291714 Z-1.000000 I89.468848 J101.839595
G03 X900.525150 Y176.213100 Z-1.000000 I84.258256 J106.178021
G03 X906.134415 Y172.410103 Z-1.000000 I78.848253 J110.259709
G03 X911.926829 Y168.892085 Z-1.000000 I73.245868 J114.071662
G03 X917.887833 Y165.667954 Z-1.000000 I67.449162 J117.582199
G03 X924.002560 Y162.745817 Z-1.000000 I61.488507 J120.809212
G03 X930.255689 Y160.132941 Z-1.000000 I55.372751 J123.729051
G03 X936.632105 Y157.835687 Z-1.000000 I49.121014 J126.345601
G03 X943.115354 Y155.860017 Z-1.000000 I42.744400 J128.642256
G03 X949.688840 Y154.211007 Z-1.000000 I36.259535 J130.615705
G03 X956.336515 Y152.892599 Z-1.000000 I29.686739 J132.267786
G03 X963.041654 Y151.908166 Z-1.000000 I23.037587 J133.585724
G03 X969.787605 Y151.260075 Z-1.000000 I16.332263 J134.568630
G03 X974.578722 Y151.005319 Z-1.000000 I9.591062 J135.196475
G03 X979.375839 Y150.920280 Z-1.000000 I4.800992 J135.480817
G01 X979.375839 Y150.920280 Z-1.000000
G00 Z5.000000

(End cutting path id: path12)


(Start cutting path id: path3734)
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X979.375221 Y158.395592

G01 Z-1.000000 F100.0(Penetrate)
G02 X974.842623 Y158.475957 Z-1.000000 I0.003929 J128.082075 F400.000000
G02 X970.315709 Y158.716672 Z-1.000000 I4.536526 J128.001710
G02 X963.941718 Y159.329017 Z-1.000000 I9.057544 J127.761413
G02 X957.606298 Y160.259158 Z-1.000000 I15.431535 J127.149068
G02 X951.325174 Y161.504872 Z-1.000000 I21.768432 J126.218672
G02 X945.114160 Y163.062956 Z-1.000000 I28.049555 J124.972958
G02 X938.988396 Y164.929699 Z-1.000000 I34.261600 J123.414589
G02 X932.963594 Y167.100291 Z-1.000000 I40.387363 J121.547846
G02 X927.055268 Y169.569086 Z-1.000000 I46.411899 J119.377357
G02 X921.277714 Y172.330086 Z-1.000000 I52.320225 J116.908562
G02 X915.645401 Y175.376427 Z-1.000000 I58.098613 J114.147138
G02 X910.172380 Y178.700456 Z-1.000000 I63.730925 J111.100797
G02 X904.872410 Y182.293759 Z-1.000000 I69.203293 J107.777187
G02 X899.758649 Y186.147452 Z-1.000000 I74.503263 J104.183884
G02 X894.843822 Y190.252201 Z-1.000000 I79.619977 J100.327848
G02 X890.140291 Y194.597460 Z-1.000000 I84.534804 J96.223099
G02 X885.659377 Y199.171975 Z-1.000000 I89.229914 J91.886018
G02 X881.412696 Y203.964728 Z-1.000000 I93.710829 J87.311503
G02 X877.411485 Y208.964309 Z-1.000000 I97.968356 J82.505873
G02 X873.665154 Y214.157622 Z-1.000000 I101.969567 J77.506292
G02 X870.182364 Y219.531256 Z-1.000000 I105.706572 J72.326610
G02 X866.972501 Y225.072246 Z-1.000000 I109.189363 J66.952975
G02 X864.043900 Y230.766503 Z-1.000000 I112.400914 J61.408894
G02 X861.403544 Y236.600009 Z-1.000000 I115.329515 J55.714637
G02 X859.058318 Y242.558254 Z-1.000000 I117.972156 J49.875727
G02 X857.013798 Y248.626261 Z-1.000000 I120.317383 J43.917482
G02 X855.275152 Y254.789288 Z-1.000000 I122.362841 J37.846440
G02 X853.846708 Y261.031511 Z-1.000000 I124.101487 J31.683413
G02 X852.731926 Y267.336917 Z-1.000000 I125.529251 J25.444546
G02 X851.933664 Y273.690158 Z-1.000000 I126.644033 J19.139140
G02 X851.453917 Y280.076087 Z-1.000000 I127.442368 J12.785175
G02 X851.293957 Y286.478013 Z-1.000000 I127.922115 J6.399246
G02 X851.373714 Y291.010600 Z-1.000000 I128.082074 J0.013223
G02 X851.613820 Y295.537526 Z-1.000000 I128.002317 J-4.519364
G02 X852.226777 Y301.911524 Z-1.000000 I127.760555 J-9.069642
G02 X853.157533 Y308.246922 Z-1.000000 I127.147599 J-15.443640
G02 X854.403239 Y314.528045 Z-1.000000 I126.218697 J-21.768291
G02 X855.961316 Y320.739060 Z-1.000000 I124.972990 J-28.049414
G02 X857.827450 Y326.864807 Z-1.000000 I123.417650 J-34.250570
G02 X859.997424 Y332.889626 Z-1.000000 I121.551515 J-40.376317
G02 X862.466824 Y338.797975 Z-1.000000 I119.373427 J-46.422007
G02 X865.228441 Y344.575518 Z-1.000000 I116.904028 J-52.330356
G02 X868.274188 Y350.207811 Z-1.000000 I114.151731 J-58.089588
G02 X871.597610 Y355.680852 Z-1.000000 I111.105984 J-63.721881
G02 X875.191522 Y360.980845 Z-1.000000 I107.771974 J-69.211411
G02 X879.045839 Y366.094588 Z-1.000000 I104.178062 J-74.511404
G02 X883.149988 Y371.009372 Z-1.000000 I100.333109 J-79.613348
G02 X887.494631 Y375.712906 Z-1.000000 I96.228960 J-84.528132
G02 X892.069748 Y380.193861 Z-1.000000 I91.880713 J-89.235377
G02 X896.863120 Y384.440536 Z-1.000000 I87.305595 J-93.716332
G02 X901.862091 Y388.441718 Z-1.000000 I82.511162 J-97.963902
G02 X907.054780 Y392.188070 Z-1.000000 I77.512191 J-101.965083
G02 X912.428419 Y395.670852 Z-1.000000 I72.326455 J-105.706678
G02 X917.969412 Y398.880706 Z-1.000000 I66.952816 J-109.189460
G02 X923.664286 Y401.808729 Z-1.000000 I61.395041 J-112.408482
G02 X929.498385 Y404.448464 Z-1.000000 I55.700166 J-115.336505
G02 X935.456615 Y406.794307 Z-1.000000 I49.886505 J-117.967599
G02 X941.524640 Y408.839433 Z-1.000000 I43.928276 J-120.313443
G02 X947.687667 Y410.578080 Z-1.000000 I37.846440 J-122.362841
G02 X953.929890 Y412.006524 Z-1.000000 I31.683413 J-124.101487
G02 X960.235321 Y413.121300 Z-1.000000 I25.444353 J-125.529290
G02 X966.588587 Y413.919553 Z-1.000000 I19.138922 J-126.644066
G02 X972.973906 Y414.399260 Z-1.000000 I12.784978 J-127.442388
G02 X979.375221 Y414.559240 Z-1.000000 I6.399658 J-127.922094
G02 X983.908426 Y414.479515 Z-1.000000 I0.014736 J-128.082074
G02 X988.435969 Y414.239397 Z-1.000000 I-4.518469 J-128.002349
G02 X994.809967 Y413.626447 Z-1.000000 I-9.069496 J-127.760566
G02 X1001.145365 Y412.695699 Z-1.000000 I-15.443494 J-127.147616
G02 X1007.426484 Y411.449988 Z-1.000000 I-21.768387 J-126.218680
G02 X1013.637493 Y409.891908 Z-1.000000 I-28.049506 J-124.972969
G02 X1019.762647 Y408.025137 Z-1.000000 I-34.265836 J-123.413412
G02 X1025.786848 Y405.854574 Z-1.000000 I-40.390989 J-121.546641
G02 X1031.695167 Y403.385778 Z-1.000000 I-46.411952 J-119.377337
G02 X1037.472716 Y400.624778 Z-1.000000 I-52.320271 J-116.908542
G02 X1043.105040 Y397.578438 Z-1.000000 I-58.098493 J-114.147198
G02 X1048.578074 Y394.254408 Z-1.000000 I-63.730818 J-111.100858
G02 X1053.878647 Y390.661118 Z-1.000000 I-69.197214 J-107.781090
G02 X1058.992997 Y386.807388 Z-1.000000 I-74.497787 J-104.187800
G02 X1063.907794 Y382.703237 Z-1.000000 I-79.613240 J-100.333194
G02 X1068.611339 Y378.358591 Z-1.000000 I-84.528037 J-96.229043
G02 X1073.091669 Y373.783456 Z-1.000000 I-89.242123 J-91.874160
G02 X1077.337734 Y368.990112 Z-1.000000 I-93.722453 J-87.299025
G02 X1081.339541 Y363.991163 Z-1.000000 I-97.957310 J-82.518988
G02 X1085.086504 Y358.798454 Z-1.000000 I-101.959117 J-77.520039
G02 X1088.568674 Y353.424788 Z-1.000000 I-105.712737 J-72.317599
G02 X1091.777928 Y347.883805 Z-1.000000 I-109.194907 J-66.943932
G02 X1094.706563 Y342.188949 Z-1.000000 I-112.403125 J-61.404848
G02 X1097.346898 Y336.354831 Z-1.000000 I-115.331759 J-55.709991
G02 X1099.692130 Y330.396586 Z-1.000000 I-117.972110 J-49.875836
G02 X1101.736656 Y324.328579 Z-1.000000 I-120.317343 J-43.917591
G02 X1103.475912 Y318.165568 Z-1.000000 I-122.359327 J-37.857799
G02 X1104.904958 Y311.923329 Z-1.000000 I-124.098583 J-31.694789
G02 X1106.019734 Y305.617898 Z-1.000000 I-125.529290 J-25.444353
G02 X1106.817987 Y299.264633 Z-1.000000 I-126.644066 J-19.138922
G02 X1107.297693 Y292.879320 Z-1.000000 I-127.442388 J-12.784978
G02 X1107.457674 Y286.478013 Z-1.000000 I-127.922094 J-6.399666
G02 X1107.377343 Y281.944810 Z-1.000000 I-128.082075 J0.002368
G02 X1107.136619 Y277.417290 Z-1.000000 I-128.001744 J4.535572
G02 X1106.524274 Y271.043307 Z-1.000000 I-127.761412 J9.057569
G02 X1105.594133 Y264.707894 Z-1.000000 I-127.149066 J15.431552
G02 X1104.348409 Y258.426782 Z-1.000000 I-126.218633 J21.768661
G02 X1102.790317 Y252.215780 Z-1.000000 I-124.972909 J28.049773
G02 X1100.923558 Y246.090627 Z-1.000000 I-123.413474 J34.265613
G02 X1098.753008 Y240.066425 Z-1.000000 I-121.546715 J40.390767
G02 X1096.284212 Y234.158106 Z-1.000000 I-119.377337 J46.411952
G02 X1093.523212 Y228.380558 Z-1.000000 I-116.908542 J52.320271
G02 X1090.476898 Y222.747621 Z-1.000000 I-114.150172 J58.092652
G02 X1087.152842 Y217.273988 Z-1.000000 I-111.103858 J63.725589
G02 X1083.559526 Y211.974028 Z-1.000000 I-107.777016 J69.203560
G02 X1079.705821 Y206.860277 Z-1.000000 I-104.183700 J74.503520
G02 X1075.601073 Y201.945445 Z-1.000000 I-100.327891 J79.619923
G02 X1071.255814 Y197.241910 Z-1.000000 I-96.223142 J84.534754
G02 X1066.681264 Y192.761616 Z-1.000000 I-91.879363 J89.236767
G02 X1061.888521 Y188.515540 Z-1.000000 I-87.304813 J93.717061
G02 X1056.888973 Y184.513721 Z-1.000000 I-82.513592 J97.961855
G02 X1051.695652 Y180.766795 Z-1.000000 I-77.514044 J101.963674
G02 X1046.321998 Y177.284625 Z-1.000000 I-72.317697 J105.712670
G02 X1040.781028 Y174.075370 Z-1.000000 I-66.944042 J109.194840
G02 X1035.086771 Y171.146763 Z-1.000000 I-61.408994 J112.400860
G02 X1029.253265 Y168.506401 Z-1.000000 I-55.714737 J115.329467
G02 X1023.294419 Y166.161139 Z-1.000000 I-49.872534 J117.973506
G02 X1017.225801 Y164.116643 Z-1.000000 I-43.913688 J120.318768
G02 X1011.063392 Y162.377412 Z-1.000000 I-37.860172 J122.358593
G02 X1004.821763 Y160.948333 Z-1.000000 I-31.697764 J124.097823
G02 X998.515733 Y159.833528 Z-1.000000 I-25.443176 J125.529529
G02 X992.161865 Y159.035304 Z-1.000000 I-19.137146 J126.644335
G02 X985.776541 Y158.555585 Z-1.000000 I-12.785224 J127.442363
G02 X979.375221 Y158.395592 Z-1.000000 I-6.399900 J127.922082
G01 X979.375221 Y158.395592 Z-1.000000
G00 Z5.000000

(End cutting path id: path3734)                 CODIGO100CERA


(Start cutting path id: path2311-0)             RPAPY
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X414.817658 Y480.259206

G01 Z-1.000000 F100.0(Penetrate)
G03 X414.455182 Y479.708147 Z-1.000000 I3.064578 J-2.410563 F400.000000
G03 X414.080512 Y478.693804 Z-1.000000 I3.327302 J-1.805380
G03 X413.905935 Y477.509205 Z-1.000000 I9.126706 J-1.950192
G03 X413.715245 Y474.089159 Z-1.000000 I64.597887 J-5.317085
G03 X413.619406 Y467.284604 Z-1.000000 I223.938260 J-6.557015
G03 X413.688858 Y441.706036 Z-1.000000 I7942.268961 J8.775613
G01 X413.848558 Y404.790846 Z-1.000000
G01 X431.499368 Y404.613006 Z-1.000000
G03 X440.568969 Y404.580095 Z-1.000000 I7.087348 J703.417586
G03 X445.530610 Y404.654643 Z-1.000000 I-0.572808 J203.274683
G03 X450.435125 Y404.922774 Z-1.000000 I-2.441892 J89.655320
G03 X451.812088 Y405.166306 Z-1.000000 I-0.616358 J7.499519
G03 X457.020625 Y407.779761 Z-1.000000 I-3.943105 J14.355447
G03 X461.030988 Y412.351516 Z-1.000000 I-9.966482 J12.787471
G03 X462.003126 Y414.239376 Z-1.000000 I-16.390398 J9.634327
G03 X462.518288 Y415.932093 Z-1.000000 I-6.520618 J2.909238
G03 X462.706560 Y417.799602 Z-1.000000 I-13.084187 J2.262319
G03 X462.775588 Y422.441706 Z-1.000000 I-156.054795 J4.642104
G03 X462.705068 Y426.784847 Z-1.000000 I-133.777251 J-0.000000
G03 X462.537838 Y428.737863 Z-1.000000 I-18.509434 J-0.601234
G03 X462.098761 Y430.548160 Z-1.000000 I-9.278807 J-1.292124
G03 X461.438788 Y432.041296 Z-1.000000 I-9.802373 J-3.440272
G03 X458.940309 Y435.561269 Z-1.000000 I-15.550591 J-8.391119
G03 X456.000975 Y437.961508 Z-1.000000 I-9.614338 J-8.773849
G03 X452.485679 Y439.468824 Z-1.000000 I-7.373496 J-12.343377
G03 X447.542088 Y440.342726 Z-1.000000 I-6.976206 J-25.044140
G01 X442.834768 Y440.711886 Z-1.000000
G01 X452.959988 Y458.021876 Z-1.000000
G03 X457.903143 Y466.617356 Z-1.000000 I-580.619140 J339.626077
G03 X460.520238 Y471.393298 Z-1.000000 I-179.000876 J101.192934
G03 X462.906895 Y476.264279 Z-1.000000 I-73.837362 J39.198634
G03 X463.085188 Y477.094366 Z-1.000000 I-1.843196 J0.830087
G03 X462.228736 Y479.590340 Z-1.000000 I-4.065258 J0.000000
G03 X460.087700 Y481.323039 Z-1.000000 I-4.781757 J-3.719497
G03 X457.387569 Y481.945078 Z-1.000000 I-2.711352 J-5.598024
G03 X454.805788 Y481.104416 Z-1.000000 I-0.007867 J-4.360663
G03 X454.404104 Y480.680418 Z-1.000000 I0.957346 J-1.309235
G03 X450.634010 Y474.728303 Z-1.000000 I232.050639 J-151.151871
G03 X447.108704 Y468.923662 Z-1.000000 I473.210890 J-291.365840
G03 X442.126468 Y460.511766 Z-1.000000 I847.019588 J-507.358012
G01 X430.531468 Y440.711886 Z-1.000000
G01 X427.299468 Y440.524686 Z-1.000000
G01 X424.067468 Y440.337486 Z-1.000000
G01 X424.067468 Y459.597176 Z-1.000000
G01 X424.067468 Y478.856866 Z-1.000000
G01 X422.547268 Y480.377026 Z-1.000000
G03 X420.631159 Y481.611586 Z-1.000000 I-3.811433 J-3.811329
G03 X418.525563 Y481.977438 Z-1.000000 I-2.018118 J-5.372801
G03 X416.458403 Y481.528430 Z-1.000000 I0.081691 J-5.359026
G03 X414.817658 Y480.259206 Z-1.000000 I1.708384 J-3.903574
G01 X414.817658 Y480.259206 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X451.036388 Y428.353476

G01 Z-1.000000 F100.0(Penetrate)
G02 X451.831649 Y427.380088 Z-1.000000 I-4.434109 J-4.434242 F400.000000
G02 X452.294600 Y426.338194 Z-1.000000 I-2.868381 J-1.898326
G02 X452.479904 Y425.145669 Z-1.000000 I-6.004729 J-1.543720
G02 X452.556588 Y422.440286 Z-1.000000 I-47.684208 J-2.705383
G02 X452.480616 Y419.672762 Z-1.000000 I-50.446290 J0.000000
G02 X452.290463 Y418.514021 Z-1.000000 I-5.450468 J0.299468
G02 X451.806807 Y417.521818 Z-1.000000 I-2.920283 J0.809530
G02 X450.917788 Y416.528516 Z-1.000000 I-5.749556 J4.251437
G02 X450.331796 Y416.018214 Z-1.000000 I-8.535803 J9.210264
G02 X449.766262 Y415.654734 Z-1.000000 I-1.866137 J2.281804
G02 X449.129946 Y415.422921 Z-1.000000 I-1.391197 J2.829540
G02 X447.880935 Y415.203387 Z-1.000000 I-2.241687 J9.090969
G02 X445.381117 Y415.056923 Z-1.000000 I-2.784184 J26.113410
G02 X436.673268 Y415.009776 Z-1.000000 I-8.707849 J804.129355
G01 X424.067468 Y415.009776 Z-1.000000
G01 X424.067468 Y422.441706 Z-1.000000
G01 X424.067468 Y429.873646 Z-1.000000
G01 X436.791868 Y429.873646 Z-1.000000
G02 X445.611353 Y429.827410 Z-1.000000 I0.000000 J-841.182272
G02 X448.102398 Y429.681749 Z-1.000000 I-0.273070 J-26.043386
G02 X449.343870 Y429.460613 Z-1.000000 I-0.957449 J-8.970591
G02 X449.955105 Y429.230215 Z-1.000000 I-0.698314 J-2.778576
G02 X450.490583 Y428.865399 Z-1.000000 I-1.203661 J-2.342136
G02 X451.036388 Y428.353476 Z-1.000000 I-7.728474 J-8.786912
G01 X451.036388 Y428.353476 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X489.373588 Y480.372146

G01 Z-1.000000 F100.0(Penetrate)
G01 X487.848588 Y478.847096 Z-1.000000 F400.000000
G01 X488.008288 Y441.818976 Z-1.000000
G01 X488.167988 Y404.790846 Z-1.000000
G01 X506.094288 Y404.626316 Z-1.000000
G03 X520.031905 Y404.661330 Z-1.000000 I5.471234 J596.129683
G03 X524.901313 Y405.039726 Z-1.000000 I-0.547812 J38.569910
G03 X527.348277 Y405.560431 Z-1.000000 I-2.556647 J18.024414
G03 X528.964610 Y406.179648 Z-1.000000 I-2.724480 J9.530819
G03 X530.456742 Y407.065431 Z-1.000000 I-4.613084 J9.470568
G03 X532.084988 Y408.312976 Z-1.000000 I-16.699812 J23.482267
G03 X533.043664 Y409.227598 Z-1.000000 I-5.816566 J7.056469
G03 X534.197363 Y410.622623 Z-1.000000 I-14.773685 J13.392542
G03 X535.221423 Y412.112218 Z-1.000000 I-16.671912 J12.558342
G03 X535.930088 Y413.417566 Z-1.000000 I-10.230253 J6.398975
G03 X536.677976 Y415.355590 Z-1.000000 I-13.960983 J6.500899
G03 X537.177288 Y417.822204 Z-1.000000 I-11.911202 J3.695003
G03 X537.335353 Y420.577781 Z-1.000000 I-33.214880 J3.287597
G03 X537.398488 Y428.253476 Z-1.000000 I-497.393367 J7.929283
G03 X537.202666 Y438.073857 Z-1.000000 I-237.852360 J0.169276
G03 X536.687900 Y441.883728 Z-1.000000 I-20.525153 J-0.833516
G03 X536.077027 Y443.769740 Z-1.000000 I-11.344089 J-2.632382
G03 X535.297480 Y445.235674 Z-1.000000 I-8.485796 J-3.572291
G03 X534.279068 Y446.571242 Z-1.000000 I-9.711295 J-6.349093
G03 X532.780588 Y448.164806 Z-1.000000 I-26.659066 J-23.567080
G03 X531.169146 Y449.683191 Z-1.000000 I-26.337020 J-26.336837
G03 X529.820502 Y450.708550 Z-1.000000 I-7.416536 J-8.355285
G03 X528.335719 Y451.481620 Z-1.000000 I-4.837132 J-7.477953
G03 X526.349375 Y452.095682 Z-1.000000 I-4.550401 J-11.199748
G03 X522.340733 Y452.597292 Z-1.000000 I-4.807343 J-22.149678
G03 X511.475788 Y452.788786 Z-1.000000 I-10.864945 J-308.131588
G01 X498.386888 Y452.788786 Z-1.000000
G01 X498.386888 Y465.822826 Z-1.000000
G03 X498.341703 Y474.869939 Z-1.000000 I-905.757307 J-0.000000
G03 X498.196863 Y477.410607 Z-1.000000 I-27.104751 J-0.270748
G03 X497.975337 Y478.676804 Z-1.000000 I-9.232428 J-0.962770
G03 X497.745553 Y479.292587 Z-1.000000 I-2.788418 J-0.689753
G03 X497.379496 Y479.831150 Z-1.000000 I-2.323358 J-1.185487
G03 X496.866688 Y480.377026 Z-1.000000 I-8.769241 J-7.724212
G03 X493.122669 Y481.926452 Z-1.000000 I-3.740520 J-3.740340
G03 X489.373588 Y480.372026 Z-1.000000 I0.003509 J-5.306836
G01 X489.373588 Y480.372146 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X525.355788 Y440.740036

G01 Z-1.000000 F100.0(Penetrate)
G02 X525.866375 Y440.195143 Z-1.000000 I-8.119336 J-8.119825 F400.000000
G02 X526.224233 Y439.682003 Z-1.000000 I-2.114366 J-1.855889
G02 X526.459916 Y439.102180 Z-1.000000 I-2.471888 J-1.342567
G02 X526.676700 Y438.031419 Z-1.000000 I-7.032571 J-1.981129
G02 X526.825999 Y435.882353 Z-1.000000 I-19.243660 J-2.416605
G02 X526.875988 Y428.634986 Z-1.000000 I-525.337269 J-7.247367
G03 X526.892070 Y423.263317 Z-1.000000 I897.129283 J0.000000
G02 X526.781005 Y420.365525 Z-1.000000 I-32.743347 J-0.196058
G02 X526.222160 Y417.712401 Z-1.000000 I-10.836858 J0.897219
G02 X525.308975 Y416.364139 Z-1.000000 I-2.963118 J1.023552
G02 X523.831865 Y415.590645 Z-1.000000 I-2.387779 J2.762708
G02 X520.679277 Y415.091874 Z-1.000000 I-4.189603 J16.268586
G02 X517.328101 Y414.995854 Z-1.000000 I-3.124481 J50.519321
G03 X511.111288 Y415.009776 Z-1.000000 I-6.216813 J-1388.058423
G01 X498.386888 Y415.009776 Z-1.000000
G01 X498.386888 Y428.634986 Z-1.000000
G01 X498.386888 Y442.260206 Z-1.000000
G01 X511.111288 Y442.260206 Z-1.000000
G02 X519.930817 Y442.213970 Z-1.000000 I0.000000 J-841.185502
G02 X522.421838 Y442.068309 Z-1.000000 I-0.273065 J-26.042898
G02 X523.663296 Y441.847173 Z-1.000000 I-0.957428 J-8.970317
G02 X524.274519 Y441.616775 Z-1.000000 I-0.698299 J-2.778472
G02 X524.809985 Y441.251961 Z-1.000000 I-1.203659 J-2.342086
G02 X525.355788 Y440.740036 Z-1.000000 I-7.729518 J-8.787955
G01 X525.355788 Y440.740036 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X563.697888 Y480.377026

G01 Z-1.000000 F100.0(Penetrate)
G03 X563.109225 Y479.677921 Z-1.000000 I3.781231 J-3.781276 F400.000000
G03 X562.624250 Y478.874178 Z-1.000000 I5.378350 J-3.793455
G03 X562.270419 Y478.005293 Z-1.000000 I4.932833 J-2.515254
G03 X562.177688 Y477.395646 Z-1.000000 I1.957641 J-0.609647
G03 X562.256602 Y476.759461 Z-1.000000 I2.603847 J-0.000000
G03 X563.889588 Y470.553604 Z-1.000000 I296.889341 J74.804594
G03 X565.557302 Y464.595696 Z-1.000000 I579.551466 J159.013429
G03 X568.005388 Y456.155516 Z-1.000000 I999.729208 J285.397650
G02 X573.820338 Y436.372819 Z-1.000000 I-15208.257602 J-4481.081231
G02 X577.961188 Y422.203966 Z-1.000000 I-7808.263193 J-2289.655169
G03 X579.752301 Y416.170781 Z-1.000000 I584.957351 J170.377895
G03 X581.085325 Y411.860367 Z-1.000000 I349.777244 J105.809591
G03 X582.546888 Y407.424695 Z-1.000000 I182.546023 J57.690637
G03 X582.799688 Y406.893706 Z-1.000000 I2.448424 J0.840003
G03 X584.554846 Y405.321393 Z-1.000000 I3.590666 J2.242437
G03 X586.902050 Y404.743706 Z-1.000000 I2.548023 J5.295576
G03 X589.265741 Y405.170379 Z-1.000000 I0.194344 J5.683907
G03 X591.009888 Y406.612986 Z-1.000000 I-1.462180 J3.543473
G03 X591.548067 Y407.695400 Z-1.000000 I-4.271642 J2.798872
G03 X594.225613 Y415.781122 Z-1.000000 I-174.422573 J62.245333
G03 X596.941928 Y424.848685 Z-1.000000 I-790.553734 J241.762021
G03 X604.318088 Y450.130306 Z-1.000000 I-8703.803194 J2553.136574
G03 X607.431378 Y461.018972 Z-1.000000 I-1380.649330 J400.644949
G03 X609.548438 Y468.676281 Z-1.000000 I-794.338798 J223.736740
G03 X611.628529 Y476.653744 Z-1.000000 I-403.584293 J109.492904
G03 X611.723888 Y477.427616 Z-1.000000 I-3.092458 J0.773872
G03 X610.323724 Y480.497022 Z-1.000000 I-4.064423 J-0.000000
G03 X606.574588 Y481.897206 Z-1.000000 I-3.749137 J-4.319258
G03 X605.400733 Y481.808861 Z-1.000000 I0.000000 J-7.842774
G03 X604.496083 Y481.560754 Z-1.000000 I0.599147 J-3.957936
G03 X603.692407 Y481.101969 Z-1.000000 I1.304927 J-3.219223
G03 X602.915925 Y480.344351 Z-1.000000 I2.732668 J-3.577421
G03 X601.830297 Y478.504120 Z-1.000000 I5.806426 J-4.665795
G03 X600.252288 Y474.029416 Z-1.000000 I43.884430 J-17.991486
G02 X598.627954 Y468.583067 Z-1.000000 I-3155.895978 J938.257351
G02 X598.159188 Y467.033326 Z-1.000000 I-256.457628 J76.727694
G01 X597.685788 Y465.485006 Z-1.000000
G01 X586.965988 Y465.485006 Z-1.000000
G01 X576.246188 Y465.485006 Z-1.000000
G01 X574.202588 Y472.603406 Z-1.000000
G03 X573.090732 Y476.199977 Z-1.000000 I-89.389468 J-25.664007
G03 X572.330225 Y478.246802 Z-1.000000 I-29.049590 J-9.628804
G03 X571.372095 Y480.165439 Z-1.000000 I-13.831343 J-5.708549
G03 X570.815688 Y480.809506 Z-1.000000 I-2.072024 J-1.227643
G03 X569.140462 Y481.674578 Z-1.000000 I-2.928590 J-3.616675
G03 X567.202525 Y481.880779 Z-1.000000 I-1.581750 J-5.655981
G03 X565.300081 Y481.444572 Z-1.000000 I0.360978 J-5.941042
G03 X563.697888 Y480.377026 Z-1.000000 I1.864206 J-4.533908
G01 X563.697888 Y480.377026 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X594.382788 Y454.196946

G01 Z-1.000000 F100.0(Penetrate)
G02 X594.304914 Y453.582642 Z-1.000000 I-2.461895 J-0.000000 F400.000000
G02 X593.263786 Y449.765389 Z-1.000000 I-136.096772 J35.068800
G02 X592.183192 Y446.108924 Z-1.000000 I-261.447198 J75.277487
G02 X590.774900 Y441.542612 Z-1.000000 I-476.757806 J144.536117
G02 X589.319566 Y436.982452 Z-1.000000 I-459.114132 J144.009892
G02 X588.218308 Y433.680024 Z-1.000000 I-207.596111 J67.392051
G02 X587.658734 Y432.082109 Z-1.000000 I-114.473570 J39.190562
G02 X587.321553 Y431.181117 Z-1.000000 I-31.762496 J11.373000
G02 X586.950476 Y430.337170 Z-1.000000 I-10.252328 J4.004319
G01 X586.896188 Y430.329036 Z-1.000000
G02 X586.746275 Y430.565306 Z-1.000000 I0.544771 J0.511350
G02 X586.179163 Y432.118937 Z-1.000000 I28.439220 J11.261312
G02 X585.690315 Y433.651680 Z-1.000000 I53.649391 J17.955086
G02 X585.079488 Y435.757266 Z-1.000000 I86.191457 J26.145396
G03 X583.284957 Y441.998670 Z-1.000000 I-320.106696 J-88.658485
G03 X579.770988 Y453.562946 Z-1.000000 I-1102.356928 J-328.650809
G02 X579.714465 Y453.914850 Z-1.000000 I1.211290 J0.375048
G02 X579.796036 Y454.150316 Z-1.000000 I0.361585 J0.006601
G02 X580.005153 Y454.301824 Z-1.000000 I0.390956 J-0.319543
G02 X580.662725 Y454.476432 Z-1.000000 I1.127596 J-2.921003
G02 X582.048857 Y454.595718 Z-1.000000 I1.611631 J-10.614331
G02 X586.909088 Y454.646766 Z-1.000000 I4.860231 J-231.343326
G02 X590.089364 Y454.605180 Z-1.000000 I-0.000000 J-121.626184
G02 X592.187375 Y454.514631 Z-1.000000 I-1.616807 J-61.811874
G02 X594.287442 Y454.306399 Z-1.000000 I-1.637745 J-27.210957
G02 X594.382788 Y454.196946 Z-1.000000 I-0.015151 J-0.109453
G01 X594.382788 Y454.196946 Z-1.000000
G00 Z5.000000

(End cutting path id: path2311-0)


(Start cutting path id: path2311-9)
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X637.020998 Y480.364686

G01 Z-1.000000 F100.0(Penetrate)
G01 X635.495998 Y478.839636 Z-1.000000 F400.000000
G01 X635.655698 Y441.811516 Z-1.000000
G01 X635.815398 Y404.783396 Z-1.000000
G01 X653.741698 Y404.618866 Z-1.000000
G03 X667.679315 Y404.653880 Z-1.000000 I5.471243 J596.130704
G03 X672.548723 Y405.032274 Z-1.000000 I-0.547811 J38.570004
G03 X674.995687 Y405.552978 Z-1.000000 I-2.556646 J18.024458
G03 X676.612020 Y406.172194 Z-1.000000 I-2.724478 J9.530836
G03 X678.104152 Y407.057974 Z-1.000000 I-4.613078 J9.470581
G03 X679.732398 Y408.305516 Z-1.000000 I-16.699780 J23.482285
G03 X680.691129 Y409.220151 Z-1.000000 I-5.815226 J7.055381
G03 X681.844823 Y410.615163 Z-1.000000 I-14.772660 J13.391753
G03 X682.868884 Y412.104750 Z-1.000000 I-16.672938 J12.559115
G03 X683.577598 Y413.410106 Z-1.000000 I-10.233907 J6.401341
G03 X684.325444 Y415.348123 Z-1.000000 I-13.963133 J6.501427
G03 X684.824748 Y417.814744 Z-1.000000 I-11.911688 J3.695057
G03 X684.982809 Y420.570321 Z-1.000000 I-33.213720 J3.287482
G03 X685.045898 Y428.246016 Z-1.000000 I-497.193023 J7.924654
G03 X684.850078 Y438.066399 Z-1.000000 I-237.855572 J0.169278
G03 X684.335323 Y441.876268 Z-1.000000 I-20.525560 J-0.833521
G03 X683.724461 Y443.762282 Z-1.000000 I-11.344303 J-2.632379
G03 X682.944932 Y445.228214 Z-1.000000 I-8.485890 J-3.572255
G03 X681.926542 Y446.563784 Z-1.000000 I-9.711373 J-6.349003
G03 X680.428098 Y448.157346 Z-1.000000 I-26.659179 J-23.566640
G03 X678.816621 Y449.675731 Z-1.000000 I-26.337730 J-26.338157
G03 X677.467954 Y450.701090 Z-1.000000 I-7.416699 J-8.355637
G03 X675.983152 Y451.474161 Z-1.000000 I-4.837224 J-7.478203
G03 X673.996798 Y452.088222 Z-1.000000 I-4.550444 J-11.199950
G03 X669.988140 Y452.589832 Z-1.000000 I-4.807366 J-22.149881
G03 X659.123198 Y452.781326 Z-1.000000 I-10.864942 J-308.131582
G01 X646.034298 Y452.781326 Z-1.000000
G01 X646.034298 Y465.815366 Z-1.000000
G03 X645.989113 Y474.862479 Z-1.000000 I-905.757307 J-0.000000
G03 X645.844273 Y477.403147 Z-1.000000 I-27.104751 J-0.270748
G03 X645.622747 Y478.669344 Z-1.000000 I-9.232428 J-0.962770
G03 X645.392963 Y479.285127 Z-1.000000 I-2.788418 J-0.689753
G03 X645.026906 Y479.823690 Z-1.000000 I-2.323358 J-1.185487
G03 X644.514098 Y480.369566 Z-1.000000 I-8.769241 J-7.724212
G03 X640.770082 Y481.918989 Z-1.000000 I-3.740527 J-3.740364
G03 X637.020998 Y480.364566 Z-1.000000 I0.003500 J-5.306827
G01 X637.020998 Y480.364686 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X673.003198 Y440.732576

G01 Z-1.000000 F100.0(Penetrate)
G02 X673.513785 Y440.187683 Z-1.000000 I-8.119336 J-8.119825 F400.000000
G02 X673.871643 Y439.674543 Z-1.000000 I-2.114366 J-1.855889
G02 X674.107326 Y439.094720 Z-1.000000 I-2.471888 J-1.342567
G02 X674.324110 Y438.023959 Z-1.000000 I-7.032571 J-1.981129
G02 X674.473409 Y435.874893 Z-1.000000 I-19.243660 J-2.416605
G02 X674.523398 Y428.627526 Z-1.000000 I-525.337269 J-7.247367
G03 X674.539480 Y423.255857 Z-1.000000 I897.129283 J0.000000
G02 X674.428415 Y420.358065 Z-1.000000 I-32.743347 J-0.196058
G02 X673.869570 Y417.704941 Z-1.000000 I-10.836858 J0.897219
G02 X672.956385 Y416.356679 Z-1.000000 I-2.963118 J1.023552
G02 X671.479275 Y415.583185 Z-1.000000 I-2.387779 J2.762708
G02 X668.326687 Y415.084414 Z-1.000000 I-4.189603 J16.268586
G02 X664.975511 Y414.988394 Z-1.000000 I-3.124481 J50.519321
G03 X658.758698 Y415.002316 Z-1.000000 I-6.216813 J-1388.058423
G01 X646.034298 Y415.002316 Z-1.000000
G01 X646.034298 Y428.627526 Z-1.000000
G01 X646.034298 Y442.252746 Z-1.000000
G01 X658.758698 Y442.252746 Z-1.000000
G02 X667.578227 Y442.206510 Z-1.000000 I0.000000 J-841.185502
G02 X670.069248 Y442.060849 Z-1.000000 I-0.273065 J-26.042898
G02 X671.310706 Y441.839713 Z-1.000000 I-0.957428 J-8.970317
G02 X671.921929 Y441.609315 Z-1.000000 I-0.698299 J-2.778472
G02 X672.457395 Y441.244501 Z-1.000000 I-1.203659 J-2.342086
G02 X673.003198 Y440.732576 Z-1.000000 I-7.729518 J-8.787955
G01 X673.003198 Y440.732576 Z-1.000000
G00 Z5.000000

(End cutting path id: path2311-9)


(Start cutting path id: path2311-6)
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X711.091986 Y404.575528

G01 Z-1.000000 F100.0(Penetrate)
G02 X709.284308 Y404.989484 Z-1.000000 I0.059983 J4.415835 F400.000000
G02 X707.624658 Y406.127547 Z-1.000000 I2.599565 J5.570146
G02 X706.442207 Y407.731900 Z-1.000000 I4.078026 J4.243540
G02 X706.014764 Y409.538529 Z-1.000000 I3.598134 J1.805188
G02 X706.072900 Y410.049881 Z-1.000000 I2.277963 J0.000000
G02 X706.478071 Y411.298256 Z-1.000000 I7.323978 J-1.687128
G02 X707.085598 Y412.602330 Z-1.000000 I19.080383 J-8.095401
G02 X708.881376 Y415.999383 Z-1.000000 I109.746960 J-55.842139
G02 X712.606010 Y422.619829 Z-1.000000 I314.958294 J-172.836224
G02 X725.588268 Y445.009761 Z-1.000000 I4634.282256 J-2672.117337
G02 X718.069912 Y457.889472 Z-1.000000 I7776.805000 J4548.238580
G02 X709.907049 Y471.932649 Z-1.000000 I6916.002887 J4029.451507
G02 X708.135966 Y475.487552 Z-1.000000 I27.155093 J15.747537
G02 X707.653626 Y477.401947 Z-1.000000 I6.003746 J2.530631
G02 X707.934521 Y479.215811 Z-1.000000 I3.716678 J0.353117
G02 X709.056970 Y480.878873 Z-1.000000 I4.845688 J-2.060177
G02 X710.663858 Y481.948117 Z-1.000000 I3.464434 J-3.464384
G02 X712.563119 Y482.383833 Z-1.000000 I2.271422 J-5.543755
G02 X714.489767 Y482.174625 Z-1.000000 I0.355120 J-5.705689
G02 X716.063771 Y481.315039 Z-1.000000 I-1.122699 J-3.926683
G02 X716.538572 Y480.778446 Z-1.000000 I-1.894316 J-2.154532
G02 X718.900931 Y477.183934 Z-1.000000 I-100.076005 J-68.344780
G02 X721.096270 Y473.637531 Z-1.000000 I-187.836509 J-118.729583
G02 X723.990952 Y468.773683 Z-1.000000 I-299.850045 J-181.746808
G02 X746.314124 Y430.490142 Z-1.000000 I-22275.044696 J-13014.250995
G02 X752.444224 Y419.794950 Z-1.000000 I-808.551831 J-470.537254
G02 X755.440313 Y414.346073 Z-1.000000 I-243.204568 J-137.275280
G02 X756.376890 Y412.469846 Z-1.000000 I-32.316617 J-17.303704
G02 X757.004916 Y410.735435 Z-1.000000 I-9.631650 J-4.468505
G02 X757.081940 Y410.147282 Z-1.000000 I-2.207039 J-0.588152
G02 X756.568423 Y408.162596 Z-1.000000 I-4.092059 J-0.000000
G02 X755.164213 Y406.460291 Z-1.000000 I-5.291965 J2.934964
G02 X753.244988 Y405.381965 Z-1.000000 I-3.827905 J4.565880
G02 X751.243538 Y405.252994 Z-1.000000 I-1.241443 J3.671103
G02 X749.471269 Y405.859530 Z-1.000000 I1.392021 J6.959927
G02 X748.337249 Y406.700979 Z-1.000000 I1.794891 J3.603857
G02 X748.011770 Y407.127235 Z-1.000000 I1.879211 J1.772311
G02 X743.347670 Y414.856681 Z-1.000000 I740.283642 J451.973684
G02 X738.874420 Y422.394370 Z-1.000000 I1762.373010 J1050.978505
G02 X731.725358 Y434.534933 Z-1.000000 I4782.613826 J2824.453074
G02 X724.456941 Y422.187848 Z-1.000000 I-5050.632890 J2964.867303
G02 X719.872710 Y414.459058 Z-1.000000 I-1856.505844 J1095.938243
G02 X715.088823 Y406.523802 Z-1.000000 I-783.209395 J466.759704
G02 X714.759455 Y406.092225 Z-1.000000 I-2.226592 J1.357807
G02 X713.625466 Y405.250780 Z-1.000000 I-2.928718 J2.762093
G02 X711.853192 Y404.644267 Z-1.000000 I-3.164000 J6.352823
G02 X711.475021 Y404.589851 Z-1.000000 I-0.687597 J3.437259
G02 X711.091986 Y404.575528 Z-1.000000 I-0.331780 J3.743936
G01 X711.091986 Y404.575528 Z-1.000000
G00 Z5.000000

(End cutting path id: path2311-6)               RPAPY

(Change tool to Cylindrical cutter)


G00 Z5.000000
G00 X417.567896 Y186.260963

G01 Z-1.000000 F100.0(Penetrate)
G03 X417.115356 Y185.723466 Z-1.000000 I9.986753 J-8.867499 F400.000000
G03 X416.702714 Y184.921425 Z-1.000000 I1.789232 J-1.427712
G03 X416.512769 Y183.910798 Z-1.000000 I5.437994 J-1.545225
G03 X416.296049 Y180.414970 Z-1.000000 I53.176586 J-5.051256
G03 X416.204385 Y173.380324 Z-1.000000 I218.487899 J-6.364879
G03 X416.324596 Y143.582277 Z-1.000000 I15359.917635 J47.065536
G01 X416.493086 Y109.914837 Z-1.000000
G01 X435.845946 Y110.127257 Z-1.000000
G01 X455.198816 Y110.339647 Z-1.000000
G01 X458.257956 Y112.529987 Z-1.000000
G03 X459.505456 Y113.553448 Z-1.000000 I-7.155576 J9.993975
G03 X460.936870 Y115.018568 Z-1.000000 I-15.815173 J16.883132
G03 X462.232725 Y116.605777 Z-1.000000 I-17.057318 J15.248810
G03 X463.081836 Y117.925957 Z-1.000000 I-9.221413 J6.864168
G03 X464.004685 Y119.875013 Z-1.000000 I-15.490346 J8.527454
G03 X464.533380 Y121.710147 Z-1.000000 I-8.817169 J3.533915
G03 X464.754592 Y123.670520 Z-1.000000 I-14.401894 J2.617808
G03 X464.841496 Y127.606757 Z-1.000000 I-86.058203 J3.869058
G03 X464.750429 Y131.365510 Z-1.000000 I-80.140828 J-0.061164
G03 X464.539906 Y133.375151 Z-1.000000 I-17.810192 J-0.849890
G03 X464.055178 Y135.277915 Z-1.000000 I-10.979630 J-1.783930
G03 X463.321936 Y136.959777 Z-1.000000 I-11.817390 J-4.151264
G03 X460.680831 Y140.750035 Z-1.000000 I-16.506713 J-8.686800
G03 X457.685665 Y143.206208 Z-1.000000 I-9.557018 J-8.599945
G03 X454.092933 Y144.669935 Z-1.000000 I-6.680164 J-11.255465
G03 X449.381306 Y145.233597 Z-1.000000 I-4.711627 J-19.410330
G02 X447.737361 Y145.286724 Z-1.000000 I-0.000000 J25.461422
G02 X446.546150 Y145.409441 Z-1.000000 I1.016389 J15.708987
G02 X445.375716 Y145.662467 Z-1.000000 I1.163899 J8.217451
G02 X445.154916 Y145.832207 Z-1.000000 I0.108231 J0.369269
G02 X445.131249 Y146.103339 Z-1.000000 I0.263157 J0.159569
G02 X446.157034 Y148.456737 Z-1.000000 I29.846669 J-11.609121
G02 X447.300280 Y150.707081 Z-1.000000 I60.050143 J-29.091796
G02 X449.076106 Y153.926167 Z-1.000000 I107.147524 J-57.009161
G03 X458.278920 Y170.145862 Z-1.000000 I-2557.472577 J1461.792463
G03 X461.672440 Y176.332060 Z-1.000000 I-274.758863 J154.746460
G03 X464.464726 Y182.128614 Z-1.000000 I-68.169074 J36.408876
G03 X464.670086 Y183.123098 Z-1.000000 I-2.305279 J0.994484
G03 X461.929010 Y187.265148 Z-1.000000 I-4.500072 J0.000000
G03 X456.899916 Y186.422850 Z-1.000000 I-1.853916 J-4.365567
G03 X456.367746 Y185.795651 Z-1.000000 I2.065420 J-2.291852
G03 X452.745891 Y179.970716 Z-1.000000 I211.876645 J-135.780102
G03 X449.362783 Y174.272885 Z-1.000000 I414.745194 J-250.109773
G03 X444.720426 Y166.226177 Z-1.000000 I704.486052 J-411.798969
G02 X440.096512 Y158.171955 Z-1.000000 I-1131.177607 J644.052542
G02 X436.809229 Y152.549672 Z-1.000000 I-646.206747 J374.057441
G02 X433.319693 Y146.759582 Z-1.000000 I-325.674359 J192.328834
G02 X432.929926 Y146.261087 Z-1.000000 I-2.405454 J1.479170
G02 X431.975283 Y145.660606 Z-1.000000 I-1.740008 J1.707175
G02 X430.206494 Y145.279994 Z-1.000000 I-2.516395 J7.393981
G02 X428.384682 Y145.284514 Z-1.000000 I-0.888001 J9.234605
G02 X426.786236 Y145.670307 Z-1.000000 I0.610394 J6.033328
G02 X426.520450 Y145.873229 Z-1.000000 I0.220936 J0.564902
G02 X426.198520 Y146.531128 Z-1.000000 I1.602780 J1.192005
G02 X426.039641 Y147.327879 Z-1.000000 I5.304246 J1.471927
G02 X425.860616 Y149.313924 Z-1.000000 I29.931399 J3.699154
G02 X425.729076 Y153.256535 Z-1.000000 I83.869590 J4.771695
G02 X425.669666 Y165.295447 Z-1.000000 I1219.757308 J12.038912
G03 X425.605923 Y176.939815 Z-1.000000 I-1063.605148 J0.000000
G03 X425.468182 Y181.258829 Z-1.000000 I-103.218595 J-1.130107
G03 X424.977646 Y185.123035 Z-1.000000 I-26.501238 J-1.400924
G03 X424.570796 Y186.082897 Z-1.000000 I-2.556682 J-0.517528
G03 X423.158934 Y187.255635 Z-1.000000 I-2.779702 J-1.910249
G03 X421.198406 Y187.697139 Z-1.000000 I-2.091629 J-4.714340
G03 X419.198287 Y187.369838 Z-1.000000 I-0.138099 J-5.431048
G03 X417.567896 Y186.260963 Z-1.000000 I1.379618 J-3.781496
G01 X417.567896 Y186.260963 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X453.683286 Y133.596787

G01 Z-1.000000 F100.0(Penetrate)
G02 X454.961540 Y131.466792 Z-1.000000 I-3.260699 J-3.405363 F400.000000
G02 X455.450995 Y127.887244 Z-1.000000 I-12.639762 J-3.551555
G02 X454.973345 Y124.283264 Z-1.000000 I-14.061456 J0.029980
G02 X453.658086 Y121.764107 Z-1.000000 I-6.296338 J1.684406
G02 X453.057018 Y121.089588 Z-1.000000 I-12.826302 J10.824528
G02 X452.532509 Y120.635244 Z-1.000000 I-2.478226 J2.331018
G02 X451.928720 Y120.310605 Z-1.000000 I-1.666971 J2.376556
G02 X450.914826 Y120.013588 Z-1.000000 I-2.152566 J5.468960
G02 X448.851341 Y119.758554 Z-1.000000 I-2.799692 J14.177006
G02 X442.067736 Y119.475227 Z-1.000000 I-18.746868 J367.499116
G02 X437.816934 Y119.385828 Z-1.000000 I-6.016351 J184.964431
G02 X434.103957 Y119.392251 Z-1.000000 I-1.571553 J164.720390
G02 X430.320940 Y119.505266 Z-1.000000 I1.458693 J112.199711
G02 X428.903096 Y119.646377 Z-1.000000 I0.634585 J13.569718
G01 X425.669736 Y120.140567 Z-1.000000
G01 X425.669736 Y127.650837 Z-1.000000
G01 X425.669736 Y135.161117 Z-1.000000
G01 X438.859676 Y135.161117 Z-1.000000
G02 X447.673783 Y135.100429 Z-1.000000 I0.000000 J-640.098306
G02 X450.401704 Y134.938051 Z-1.000000 I-0.411990 J-29.916734
G02 X451.769754 Y134.708901 Z-1.000000 I-1.197438 J-11.347132
G02 X452.478057 Y134.470200 Z-1.000000 I-0.856462 J-3.711626
G02 X453.109558 Y134.090976 Z-1.000000 I-1.290881 J-2.865052
G02 X453.683366 Y133.596787 Z-1.000000 I-4.396032 J-5.684499
G01 X453.683286 Y133.596787 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X507.627796 Y185.496096

G01 Z-1.000000 F100.0(Penetrate)
G03 X506.077820 Y184.079932 Z-1.000000 I5.987440 J-8.109487 F400.000000
G03 X503.992637 Y181.245691 Z-1.000000 I14.341051 J-12.735038
G03 X502.141491 Y177.910354 Z-1.000000 I45.699764 J-27.545212
G03 X497.815856 Y169.217227 Z-1.000000 I469.771061 J-239.177313
G01 X490.964486 Y155.124157 Z-1.000000
G01 X490.970486 Y148.919817 Z-1.000000
G01 X490.976486 Y142.715477 Z-1.000000
G01 X496.544616 Y131.204077 Z-1.000000
G03 X502.252716 Y119.792955 Z-1.000000 I431.980633 J208.953215
G03 X504.696217 Y115.642323 Z-1.000000 I38.949015 J20.134945
G03 X506.174267 Y113.681139 Z-1.000000 I17.394381 J11.571747
G03 X507.287397 Y112.574435 Z-1.000000 I7.156105 J6.084510
G03 X508.571265 Y111.694053 Z-1.000000 I5.129904 J6.104648
G03 X509.953696 Y111.047857 Z-1.000000 I5.575471 J10.125966
G03 X516.355931 Y110.197999 Z-1.000000 I4.943342 J12.699778
G03 X522.327956 Y112.380187 Z-1.000000 I-1.413997 J13.132666
G03 X523.740827 Y113.606021 Z-1.000000 I-4.324057 J6.410962
G03 X525.816594 Y116.390552 Z-1.000000 I-13.678128 J12.362513
G03 X527.665645 Y119.686563 Z-1.000000 I-46.119829 J28.039726
G03 X532.021026 Y128.323377 Z-1.000000 I-469.072224 J241.960812
G01 X538.964516 Y142.411987 Z-1.000000
G01 X538.964516 Y148.768087 Z-1.000000
G01 X538.964516 Y155.124197 Z-1.000000
G01 X532.376506 Y168.705067 Z-1.000000
G03 X529.515152 Y174.475789 Z-1.000000 I-334.187316 J-162.108664
G03 X527.274429 Y178.792159 Z-1.000000 I-220.371489 J-111.659920
G03 X524.845374 Y183.169477 Z-1.000000 I-123.197802 J-65.502103
G03 X524.258776 Y183.973086 Z-1.000000 I-4.069821 J-2.354887
G03 X522.310098 Y185.719963 Z-1.000000 I-9.387792 J-8.511945
G03 X520.262857 Y186.837545 Z-1.000000 I-5.375126 J-7.412505
G03 X517.987576 Y187.442475 Z-1.000000 I-3.824152 J-9.802124
G03 X514.752276 Y187.681910 Z-1.000000 I-3.235300 J-21.738381
G03 X512.284493 Y187.571118 Z-1.000000 I0.000000 J-27.539045
G03 X510.808317 Y187.281554 Z-1.000000 I0.649454 J-7.218374
G03 X509.439847 Y186.684720 Z-1.000000 I1.931468 J-6.295914
G03 X507.627296 Y185.496096 Z-1.000000 I9.778354 J-16.887471
G01 X507.627796 Y185.496096 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X523.258956 Y164.794327

G01 Z-1.000000 F100.0(Penetrate)
G02 X526.236777 Y158.756402 Z-1.000000 I-1386.977419 J-687.791377 F400.000000
G02 X527.849810 Y155.174006 Z-1.000000 I-53.564838 J-26.272798
G02 X529.043224 Y151.545408 Z-1.000000 I-24.420129 J-10.042115
G02 X529.377750 Y148.823812 Z-1.000000 I-10.835854 J-2.713250
G02 X529.039119 Y146.102752 Z-1.000000 I-11.168866 J0.008345
G02 X527.840318 Y142.476701 Z-1.000000 I-25.590731 J6.449305
G02 X526.221910 Y138.897529 Z-1.000000 I-55.109686 J22.763664
G02 X523.235056 Y132.865627 Z-1.000000 I-1387.607237 J683.354570
G02 X519.006395 Y124.534369 Z-1.000000 I-541.207718 J269.459677
G02 X517.615699 Y122.174274 Z-1.000000 I-22.100912 J11.433282
G02 X516.763451 Y121.095601 Z-1.000000 I-7.060586 J4.702477
G02 X516.247684 Y120.683767 Z-1.000000 I-1.646711 J1.533403
G02 X515.636159 Y120.466513 Z-1.000000 I-0.885368 J1.522844
G02 X514.940856 Y120.412137 Z-1.000000 I-0.695302 J4.418206
G02 X514.243959 Y120.465916 Z-1.000000 I-0.000000 J4.542238
G02 X513.634948 Y120.682339 Z-1.000000 I0.266111 J1.713923
G02 X513.123189 Y121.094548 Z-1.000000 I1.115164 J1.908257
G02 X512.276257 Y122.181891 Z-1.000000 I6.210231 J5.710661
G02 X510.899154 Y124.559928 Z-1.000000 I20.882753 J13.680797
G02 X506.692456 Y133.002727 Z-1.000000 I554.348479 J281.478354
G02 X503.762396 Y139.030637 Z-1.000000 I1360.476947 J665.029696
G02 X502.170725 Y142.601435 Z-1.000000 I56.914820 J27.509737
G02 X500.972727 Y146.214428 Z-1.000000 I25.843478 J10.574304
G02 X500.615647 Y148.867664 Z-1.000000 I10.994430 J2.830306
G02 X500.900227 Y151.522118 Z-1.000000 I10.931621 J0.170520
G02 X501.977914 Y154.961788 Z-1.000000 I23.604350 J-5.506849
G02 X503.464624 Y158.358519 Z-1.000000 I49.924250 J-19.827536
G02 X506.208216 Y164.044177 Z-1.000000 I1138.741791 J-545.990643
G02 X508.666203 Y169.050411 Z-1.000000 I580.469420 J-281.895367
G02 X510.465634 Y172.629199 Z-1.000000 I347.842459 J-172.654990
G02 X512.385435 Y176.305931 Z-1.000000 I181.845125 J-92.610602
G02 X512.677276 Y176.746477 Z-1.000000 I2.489636 J-1.332333
G02 X513.573174 Y177.335984 Z-1.000000 I1.368024 J-1.103522
G02 X515.257226 Y177.465940 Z-1.000000 I1.171117 J-4.199589
G02 X515.796651 Y177.324161 Z-1.000000 I-0.236655 J-1.997459
G02 X516.421189 Y176.939234 Z-1.000000 I-0.909587 J-2.174905
G02 X516.950352 Y176.388917 Z-1.000000 I-2.580557 J-3.010930
G02 X517.829809 Y175.147375 Z-1.000000 I-11.244658 J-8.897507
G02 X519.326449 Y172.540651 Z-1.000000 I-30.233986 J-19.091733
G02 X523.258916 Y164.794327 Z-1.000000 I-410.250887 J-213.137603
G01 X523.258956 Y164.794327 Z-1.000000
G00 Z5.000000


(Start cutting path id: path4529)       ROBOT B_INICIO
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X565.612356 Y187.611395

G01 Z-1.000000 F100.0(Penetrate)
G03 X565.456846 Y187.309668 Z-1.000000 I0.447642 J-0.421653 F400.000000
G03 X565.186489 Y185.393668 Z-1.000000 I16.926136 J-3.365441
G03 X565.058251 Y183.380186 Z-1.000000 I49.074424 J-4.136338
G03 X564.915174 Y179.062171 Z-1.000000 I215.920728 J-9.315930
G03 X564.791353 Y170.511380 Z-1.000000 I494.308765 J-11.434184
G03 X564.729986 Y149.465327 Z-1.000000 I3608.892010 J-21.046055
G03 X564.769118 Y133.617141 Z-1.000000 I3209.213406 J0.000005
G03 X564.855839 Y122.851938 Z-1.000000 I1726.842544 J8.527837
G03 X565.056750 Y111.640081 Z-1.000000 I831.201206 J9.287011
G03 X565.158466 Y111.117957 Z-1.000000 I1.592582 J0.039285
G03 X565.355928 Y110.836479 Z-1.000000 I0.646123 J0.243267
G03 X565.911925 Y110.522401 Z-1.000000 I1.080736 J1.264009
G03 X566.592097 Y110.359157 Z-1.000000 I1.292797 J3.887926
G03 X568.226244 Y110.178189 Z-1.000000 I3.089751 J20.431938
G03 X571.477344 Y110.043284 Z-1.000000 I3.948643 J55.916952
G03 X581.217466 Y109.983937 Z-1.000000 I9.542270 J766.781222
G03 X587.889284 Y110.049514 Z-1.000000 I-0.122809 J351.920540
G03 X593.019021 Y110.202921 Z-1.000000 I-4.679279 J242.313184
G03 X598.284204 Y110.515392 Z-1.000000 I-5.679949 J140.223548
G03 X599.443766 Y110.722767 Z-1.000000 I-0.465228 J5.946984
G03 X602.449800 Y111.947248 Z-1.000000 I-3.992572 J14.103615
G03 X605.815100 Y114.125598 Z-1.000000 I-12.488805 J22.982433
G03 X608.807717 Y116.792937 Z-1.000000 I-15.832337 J20.775519
G03 X610.777596 Y119.359907 Z-1.000000 I-10.483356 J10.084194
G03 X613.173938 Y125.255707 Z-1.000000 I-17.419329 J10.514964
G03 X613.906444 Y132.206419 Z-1.000000 I-27.512755 J6.413405
G03 X612.912687 Y139.126025 Z-1.000000 I-28.341866 J-0.539145
G03 X610.290406 Y144.969427 Z-1.000000 I-19.940782 J-5.438523
G02 X607.893324 Y148.839669 Z-1.000000 I83.434683 J54.353746
G02 X607.789286 Y149.221767 Z-1.000000 I0.649641 J0.382098
G02 X607.903181 Y149.595754 Z-1.000000 I0.670963 J0.000000
G02 X609.849496 Y152.290547 Z-1.000000 I40.280223 J-27.042146
G03 X613.130608 Y158.869624 Z-1.000000 I-14.827613 J11.502531
G03 X614.041345 Y166.827491 Z-1.000000 I-25.604009 J6.961294
G03 X612.544842 Y174.703739 Z-1.000000 I-26.770708 J-1.006194
G03 X608.764826 Y181.149209 Z-1.000000 I-18.429021 J-6.476750
G03 X606.845788 Y183.185819 Z-1.000000 I-31.964719 J-28.196980
G03 X605.203279 Y184.582208 Z-1.000000 I-9.789363 J-9.850587
G03 X603.366784 Y185.661268 Z-1.000000 I-6.540982 J-9.030022
G03 X600.918674 Y186.547479 Z-1.000000 I-6.232558 J-13.392627
G03 X595.927459 Y187.378486 Z-1.000000 I-6.898055 J-26.026577
G03 X582.730186 Y188.049404 Z-1.000000 I-23.298004 J-328.149004
G03 X573.339262 Y188.274338 Z-1.000000 I-21.329789 J-694.365071
G03 X569.591362 Y188.258605 Z-1.000000 I-1.506148 J-87.626006
G03 X566.182085 Y187.918441 Z-1.000000 I0.593788 J-23.205932
G03 X565.612356 Y187.611395 Z-1.000000 I0.176866 J-1.010273
G01 X565.612356 Y187.611395 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X599.130206 Y176.647522

G01 Z-1.000000 F100.0(Penetrate)
G02 X601.238326 Y175.034756 Z-1.000000 I-3.885039 J-7.262506 F400.000000
G02 X602.727641 Y172.803993 Z-1.000000 I-5.862216 J-5.526300
G02 X603.522642 Y170.170601 Z-1.000000 I-9.814405 J-4.399592
G02 X603.859316 Y166.098027 Z-1.000000 I-24.463585 J-4.072574
G02 X603.500869 Y161.887967 Z-1.000000 I-24.903395 J-0.000000
G02 X602.651786 Y159.254016 Z-1.000000 I-9.638461 J1.653232
G02 X601.058651 Y157.066476 Z-1.000000 I-6.959802 J3.394774
G02 X598.589426 Y155.293957 Z-1.000000 I-7.385432 J7.682198
G02 X596.872461 Y154.581837 Z-1.000000 I-4.972774 J9.563746
G02 X594.522957 Y154.102743 Z-1.000000 I-3.433039 J10.835231
G02 X591.937283 Y153.975194 Z-1.000000 I-2.672673 J27.908182
G02 X585.368526 Y154.027837 Z-1.000000 I-0.916815 J295.448367
G01 X574.965396 Y154.226887 Z-1.000000
G01 X574.771706 Y165.868677 Z-1.000000
G01 X574.577936 Y177.510460 Z-1.000000
G01 X576.536426 Y177.800094 Z-1.000000
G02 X578.503555 Y177.960668 Z-1.000000 I2.209855 J-14.942532
G02 X587.671416 Y177.979046 Z-1.000000 I5.237210 J-325.874914
G02 X592.839529 Y177.838485 Z-1.000000 I-2.061018 J-170.859316
G02 X595.402430 Y177.635592 Z-1.000000 I-1.482972 J-35.020962
G02 X597.831680 Y177.164893 Z-1.000000 I-1.888014 J-16.247915
G02 X599.130206 Y176.647522 Z-1.000000 I-1.649702 J-6.028750
G01 X599.130206 Y176.647522 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X598.163916 Y142.802667

G01 Z-1.000000 F100.0(Penetrate)
G02 X600.380619 Y141.469486 Z-1.000000 I-3.660859 J-8.596440 F400.000000
G02 X602.053216 Y139.654056 Z-1.000000 I-5.337190 J-6.595489
G02 X603.178775 Y137.430492 Z-1.000000 I-7.948377 J-5.420095
G02 X603.917126 Y134.365837 Z-1.000000 I-14.559953 J-5.129123
G02 X603.858372 Y128.916632 Z-1.000000 I-18.699267 J-2.523303
G02 X602.488375 Y125.051474 Z-1.000000 I-11.015487 J1.729042
G02 X599.793595 Y122.040486 Z-1.000000 I-8.607545 J4.992212
G02 X595.580566 Y119.991997 Z-1.000000 I-7.618501 J10.311994
G02 X594.313873 Y119.755869 Z-1.000000 I-2.026453 J7.355150
G02 X591.312399 Y119.538843 Z-1.000000 I-5.435586 J54.310609
G02 X588.344121 Y119.457098 Z-1.000000 I-3.877186 J86.854591
G02 X584.905895 Y119.466154 Z-1.000000 I-1.373175 J131.343204
G02 X581.463652 Y119.567546 Z-1.000000 I1.972531 J125.450035
G02 X578.812499 Y119.735387 Z-1.000000 I2.858701 J66.177140
G02 X576.128203 Y120.064372 Z-1.000000 I2.909136 J34.852103
G02 X575.483656 Y120.307997 Z-1.000000 I0.276324 J1.705496
G02 X575.204078 Y120.700159 Z-1.000000 I0.338138 J0.536802
G02 X574.811989 Y123.248121 Z-1.000000 I16.247366 J3.804354
G02 X574.678426 Y125.976072 Z-1.000000 I51.504506 J3.888942
G02 X574.612446 Y131.846087 Z-1.000000 I261.084152 J5.870015
G02 X574.655195 Y136.527007 Z-1.000000 I256.298925 J-0.000000
G02 X574.750692 Y139.748173 Z-1.000000 I141.647491 J-2.587413
G02 X574.966603 Y143.062733 Z-1.000000 I69.044095 J-2.833221
G02 X575.083076 Y143.315027 Z-1.000000 I0.411005 J-0.036711
G02 X575.537214 Y143.537447 Z-1.000000 I0.541956 J-0.531725
G02 X578.745114 Y143.785737 Z-1.000000 I4.811985 J-41.323677
G02 X581.893764 Y143.849505 Z-1.000000 I3.314805 J-85.907471
G02 X586.065196 Y143.807242 Z-1.000000 I0.334167 J-172.903037
G02 X590.228341 Y143.666382 Z-1.000000 I-3.975565 J-179.090860
G02 X593.664428 Y143.454549 Z-1.000000 I-4.850869 J-106.658554
G02 X597.159908 Y143.079745 Z-1.000000 I-4.685208 J-60.181929
G02 X598.163916 Y142.802667 Z-1.000000 I-0.530912 J-3.881376
G01 X598.163916 Y142.802667 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)         B_FIM


G00 Z5.000000
G00 X658.613726 Y186.911955

G01 Z-1.000000 F100.0(Penetrate)
G03 X657.420873 Y186.301027 Z-1.000000 I3.836733 J-8.961322 F400.000000
G03 X656.262371 Y185.453173 Z-1.000000 I4.136006 J-6.866824
G03 X655.249068 Y184.418246 Z-1.000000 I6.425861 J-7.305128
G03 X653.915190 Y182.653990 Z-1.000000 I15.970781 J-13.461192
G03 X651.677363 Y178.938836 Z-1.000000 I34.400641 J-23.252819
G03 X646.733096 Y169.335607 Z-1.000000 I330.114146 J-176.035187
G01 X639.943076 Y155.665807 Z-1.000000
G01 X639.925476 Y149.116507 Z-1.000000
G01 X639.907876 Y142.567197 Z-1.000000
G01 X645.779826 Y130.410467 Z-1.000000
G03 X651.003444 Y119.939337 Z-1.000000 I412.843020 J199.411975
G03 X653.266337 Y116.057801 Z-1.000000 I38.190772 J19.664417
G03 X654.627507 Y114.208056 Z-1.000000 I17.710138 J11.606644
G03 X655.680458 Y113.106523 Z-1.000000 I7.935537 J6.531523
G03 X656.885693 Y112.189513 Z-1.000000 I6.092059 J6.756331
G03 X658.209786 Y111.444147 Z-1.000000 I7.510197 J11.792589
G03 X662.292621 Y110.160617 Z-1.000000 I6.182041 J12.529318
G03 X666.064852 Y110.205691 Z-1.000000 I1.748880 J11.507692
G03 X669.611529 Y111.464387 Z-1.000000 I-1.975872 J11.193642
G03 X672.849016 Y114.028227 Z-1.000000 I-6.202547 J11.158249
G03 X673.578050 Y115.005916 Z-1.000000 I-4.617622 J4.203875
G03 X676.109511 Y119.397968 Z-1.000000 I-108.236851 J65.310394
G03 X678.439278 Y123.752274 Z-1.000000 I-187.785553 J103.274883
G03 X681.305876 Y129.405417 Z-1.000000 I-268.414772 J139.661255
G01 X687.880886 Y142.715477 Z-1.000000
G01 X687.894286 Y148.924287 Z-1.000000
G01 X687.908486 Y155.133097 Z-1.000000
G01 X681.083156 Y169.168127 Z-1.000000
G03 X675.174720 Y180.645136 Z-1.000000 I-254.728566 J-123.876497
G03 X672.522764 Y184.406274 Z-1.000000 I-17.475386 J-9.506284
G03 X670.765767 Y185.954795 Z-1.000000 I-7.391668 J-6.615769
G03 X669.112425 Y186.798137 Z-1.000000 I-3.992613 J-5.785049
G03 X667.266132 Y187.239566 Z-1.000000 I-3.025775 J-8.573621
G03 X664.705876 Y187.466171 Z-1.000000 I-3.988036 J-30.481462
G03 X662.853188 Y187.486072 Z-1.000000 I-1.204807 J-25.913661
G03 X661.131570 Y187.386357 Z-1.000000 I0.655107 J-26.222803
G03 X659.404699 Y187.150209 Z-1.000000 I1.748536 J-19.218528
G03 X658.614396 Y186.911955 Z-1.000000 I0.666650 J-3.641187
G01 X658.613726 Y186.911955 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X672.259986 Y164.659097

G01 Z-1.000000 F100.0(Penetrate)
G02 X675.149941 Y158.787004 Z-1.000000 I-1197.585720 J-593.039857 F400.000000
G02 X676.716258 Y155.282552 Z-1.000000 I-49.861377 J-24.387827
G02 X677.857424 Y151.730777 Z-1.000000 I-23.109935 J-9.384310
G02 X678.165855 Y148.996906 Z-1.000000 I-10.841578 J-2.607465
G02 X677.805328 Y146.262139 Z-1.000000 I-11.545381 J0.130889
G02 X676.571778 Y142.538330 Z-1.000000 I-27.530523 J7.053556
G02 X674.927552 Y138.858536 Z-1.000000 I-59.667022 J24.453510
G02 X671.897026 Y132.642997 Z-1.000000 I-1421.455963 J689.216236
G02 X667.822486 Y124.498175 Z-1.000000 I-510.641164 J250.362472
G02 X666.465877 Y122.166988 Z-1.000000 I-21.784140 J11.116720
G02 X665.632777 Y121.099435 Z-1.000000 I-7.036251 J4.632123
G02 X665.125852 Y120.687992 Z-1.000000 I-1.653747 J1.519522
G02 X664.523670 Y120.468052 Z-1.000000 I-0.886313 J1.492331
G02 X663.845136 Y120.412137 Z-1.000000 I-0.678534 J4.089085
G02 X663.168123 Y120.469161 Z-1.000000 I0.000000 J4.047390
G02 X662.560353 Y120.694111 Z-1.000000 I0.292428 J1.723594
G02 X662.048176 Y121.112543 Z-1.000000 I1.185875 J1.974239
G02 X661.197635 Y122.202014 Z-1.000000 I6.468554 J5.926683
G02 X659.806207 Y124.576609 Z-1.000000 I21.283564 J14.066370
G02 X655.613336 Y132.867497 Z-1.000000 I525.619798 J271.022303
G02 X652.654996 Y138.876259 Z-1.000000 I1049.582518 J520.481713
G02 X651.043512 Y142.459101 Z-1.000000 I55.207468 J26.984937
G02 X649.835392 Y146.089073 Z-1.000000 I25.502116 J10.503590
G02 X649.475236 Y148.778054 Z-1.000000 I11.090613 J2.854060
G02 X649.766509 Y151.469046 Z-1.000000 I11.098151 J0.159996
G02 X650.863333 Y154.956844 Z-1.000000 I23.873229 J-5.591165
G02 X652.376787 Y158.398028 Z-1.000000 I49.848013 J-19.870048
G02 X655.162626 Y164.127957 Z-1.000000 I907.820034 J-437.831734
G02 X657.624343 Y169.098998 Z-1.000000 I606.028560 J-297.017233
G02 X659.424806 Y172.653809 Z-1.000000 I363.320640 J-181.783632
G02 X661.342726 Y176.307502 Z-1.000000 I190.033914 J-97.423579
G02 X661.631686 Y176.746477 Z-1.000000 I2.601074 J-1.397592
G02 X662.517894 Y177.335135 Z-1.000000 I1.351934 J-1.073888
G02 X664.200066 Y177.465940 Z-1.000000 I1.171310 J-4.181331
G02 X664.739423 Y177.323761 Z-1.000000 I-0.235455 J-1.987329
G02 X665.365890 Y176.935969 Z-1.000000 I-0.913792 J-2.176117
G02 X665.896409 Y176.381743 Z-1.000000 I-2.605239 J-3.024823
G02 X666.781461 Y175.127517 Z-1.000000 I-11.424880 J-9.001423
G02 X668.288447 Y172.495695 Z-1.000000 I-30.794004 J-19.380060
G02 X672.259986 Y164.659097 Z-1.000000 I-420.494254 J-218.028547
G01 X672.259986 Y164.659097 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X734.536576 Y186.082897

G01 Z-1.000000 F100.0(Penetrate)
G03 X734.284276 Y185.566917 Z-1.000000 I1.612659 J-1.108219 F400.000000
G03 X733.962105 Y184.152028 Z-1.000000 I8.280853 J-2.629679
G03 X733.802834 Y182.626078 Z-1.000000 I20.941762 J-2.957084
G03 X733.629121 Y179.164400 Z-1.000000 I98.360807 J-6.671113
G03 X733.497469 Y172.333262 Z-1.000000 I256.918438 J-8.368265
G03 X733.437706 Y152.345117 Z-1.000000 I3342.569249 J-19.988145
G02 X733.374217 Y127.841210 Z-1.000000 I-4728.748518 J0.000000
G02 X733.002901 Y122.537516 Z-1.000000 I-41.105837 J0.213009
G02 X732.740612 Y121.150829 Z-1.000000 I-13.369071 J1.810582
G02 X732.491689 Y120.425127 Z-1.000000 I-4.093514 J0.998572
G02 X732.090378 Y119.807906 Z-1.000000 I-2.176904 J0.976327
G02 X731.691490 Y119.503539 Z-1.000000 I-1.001181 J0.898534
G02 X730.704505 Y119.295922 Z-1.000000 I-0.888686 J1.774888
G03 X728.968156 Y119.342757 Z-1.000000 I-1.905316 J-38.426989
G02 X725.908171 Y119.562944 Z-1.000000 I0.099658 J22.757673
G02 X724.651774 Y120.000259 Z-1.000000 I0.474996 J3.388112
G02 X723.868463 Y120.907708 Z-1.000000 I0.941400 J1.604418
G02 X723.555246 Y122.518767 Z-1.000000 I3.986699 J1.611059
G03 X723.150560 Y124.632930 Z-1.000000 I-5.724752 J-0.000000
G03 X722.110680 Y126.327611 Z-1.000000 I-5.394426 J-2.143713
G03 X720.561004 Y127.529487 Z-1.000000 I-3.922414 J-3.457483
G03 X718.762966 Y127.955477 Z-1.000000 I-1.775526 J-3.486611
G03 X717.310596 Y127.803883 Z-1.000000 I0.042891 J-7.444034
G03 X716.276253 Y127.417213 Z-1.000000 I0.716054 J-3.492218
G03 X715.439789 Y126.729061 Z-1.000000 I1.525831 J-2.707125
G03 X714.714434 Y125.575162 Z-1.000000 I3.560822 J-3.043312
G03 X714.013817 Y122.939002 Z-1.000000 I8.061881 J-3.553803
G03 X713.672786 Y116.408247 Z-1.000000 I62.361659 J-6.530755
G01 X713.672786 Y109.979917 Z-1.000000
G01 X738.012516 Y109.979917 Z-1.000000
G01 X762.352316 Y109.979917 Z-1.000000
G01 X762.806756 Y111.825337 Z-1.000000
G03 X762.935588 Y112.639938 Z-1.000000 I-4.741199 J1.167327
G03 X763.032586 Y114.803086 Z-1.000000 I-39.284758 J2.845315
G03 X763.030373 Y116.930611 Z-1.000000 I-57.932452 J1.003501
G03 X762.937416 Y119.370887 Z-1.000000 I-65.378626 J-1.268545
G03 X762.348418 Y124.113495 Z-1.000000 I-35.725830 J-2.029012
G03 X761.565890 Y126.155702 Z-1.000000 I-6.148532 J-1.184954
G03 X760.866620 Y127.010963 Z-1.000000 I-3.300147 J-1.984737
G03 X760.046483 Y127.556055 Z-1.000000 I-2.155668 J-2.353857
G03 X759.087853 Y127.853729 Z-1.000000 I-1.612190 J-3.499477
G03 X757.767986 Y127.966487 Z-1.000000 I-1.319867 J-7.668350
G03 X755.907868 Y127.650733 Z-1.000000 I-0.000000 J-5.636884
G03 X754.702467 Y126.861378 Z-1.000000 I0.986108 J-2.820900
G03 X753.915433 Y125.589463 Z-1.000000 I2.853254 J-2.644990
G03 X753.253956 Y123.096987 Z-1.000000 I11.932273 J-4.500714
G02 X752.895486 Y121.585515 Z-1.000000 I-10.278914 J1.639570
G02 X752.491566 Y120.834541 Z-1.000000 I-2.198223 J0.698224
G02 X751.840509 Y120.304225 Z-1.000000 I-1.582814 J1.278385
G02 X750.597126 Y119.807547 Z-1.000000 I-3.209901 J6.230975
G02 X748.346866 Y119.443820 Z-1.000000 I-2.582527 J8.834576
G02 X745.840406 Y119.635707 Z-1.000000 I-0.404113 J11.187246
G01 X743.320166 Y120.117607 Z-1.000000
G01 X743.210786 Y147.964187 Z-1.000000
G03 X743.055301 Y172.729837 Z-1.000000 I-5268.017433 J-20.690709
G03 X742.868700 Y179.715244 Z-1.000000 I-193.276007 J-1.667778
G03 X742.607959 Y183.233694 Z-1.000000 I-60.387213 J-2.706216
G03 X742.381488 Y184.557335 Z-1.000000 I-10.081233 J-1.043672
G03 X741.927537 Y185.706234 Z-1.000000 I-4.287546 J-1.029955
G03 X741.492086 Y186.314298 Z-1.000000 I-3.112627 J-1.769086
G03 X739.897663 Y187.385387 Z-1.000000 I-2.867246 J-2.545907
G03 X737.895052 Y187.693289 Z-1.000000 I-1.820278 J-5.172706
G03 X735.933031 Y187.243134 Z-1.000000 I0.175133 J-5.264192
G03 X734.536526 Y186.082897 Z-1.000000 I1.354323 J-3.050674
G01 X734.536576 Y186.082897 Z-1.000000
G00 Z5.000000

(End cutting path id: path4529)             ROBOT


G00 Z5.000000
G00 X415.180684 Y290.173954                 FRAME

G01 Z-1.000000 F100.0(Penetrate)
G01 X413.834084 Y288.740550 Z-1.000000 F400.000000
G01 X413.834084 Y251.444402 Z-1.000000
G01 X413.834084 Y214.148264 Z-1.000000
G01 X437.113084 Y214.324643 Z-1.000000
G01 X460.391984 Y214.501015 Z-1.000000
G01 X461.886184 Y216.238701 Z-1.000000
G03 X463.036339 Y218.092582 Z-1.000000 I-5.360920 J4.609655
G03 X463.307997 Y219.615649 Z-1.000000 I-3.208737 J1.358079
G03 X462.898058 Y221.112253 Z-1.000000 I-3.551517 J-0.168360
G03 X461.569984 Y222.909582 Z-1.000000 I-6.649488 J-3.524076
G01 X459.759684 Y224.719928 Z-1.000000
G01 X442.097584 Y224.719928 Z-1.000000
G01 X424.435484 Y224.719928 Z-1.000000
G01 X424.244284 Y232.153035 Z-1.000000
G01 X424.052984 Y239.586142 Z-1.000000
G01 X437.145084 Y239.584942 Z-1.000000
G03 X447.484253 Y239.754047 Z-1.000000 I0.025580 J314.592836
G03 X450.478434 Y240.210945 Z-1.000000 I-0.419474 J12.788232
G03 X451.833761 Y240.832437 Z-1.000000 I-1.249355 J4.513113
G03 X452.525719 Y241.543972 Z-1.000000 I-1.327123 J1.982837
G03 X452.939027 Y242.507426 Z-1.000000 I-2.842540 J1.789789
G03 X453.256584 Y244.045170 Z-1.000000 I-16.951519 J4.302295
G03 X453.372447 Y245.422305 Z-1.000000 I-9.127711 J1.461382
G03 X453.227597 Y246.267364 Z-1.000000 I-2.409365 J0.021959
G03 X452.804321 Y247.026254 Z-1.000000 I-2.560681 J-0.930749
G03 X451.737084 Y248.225633 Z-1.000000 I-10.819264 J-8.552733
G01 X449.850384 Y250.112416 Z-1.000000
G01 X437.106484 Y250.112416 Z-1.000000
G01 X424.362684 Y250.112416 Z-1.000000
G01 X424.362684 Y268.722974 Z-1.000000
G03 X424.339904 Y276.981736 Z-1.000000 I-1497.048579 J-0.000000
G03 X424.250920 Y281.708862 Z-1.000000 I-177.678177 J-0.980211
G03 X423.956007 Y286.359226 Z-1.000000 I-74.634924 J-2.398607
G03 X423.579122 Y288.529790 Z-1.000000 I-14.185332 J-1.345063
G03 X422.744208 Y290.382632 Z-1.000000 I-5.686014 J-1.447658
G03 X421.842530 Y291.168407 Z-1.000000 I-1.948504 J-1.325686
G03 X420.595769 Y291.518095 Z-1.000000 I-1.548379 J-3.123100
G03 X418.536384 Y291.607357 Z-1.000000 I-2.059384 J-23.711479
G03 X416.690708 Y291.229661 Z-1.000000 I0.000000 J-4.698446
G03 X415.180684 Y290.173954 Z-1.000000 I1.765497 J-4.133050
G01 X415.180684 Y290.173954 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X489.500084 Y290.173954

G01 Z-1.000000 F100.0(Penetrate)
G01 X488.153484 Y288.740550 Z-1.000000 F400.000000
G01 X488.153484 Y251.465948 Z-1.000000
G01 X488.153484 Y214.191351 Z-1.000000
G01 X506.345784 Y214.191351 Z-1.000000
G03 X520.122257 Y214.335390 Z-1.000000 I-0.000000 J658.892374
G03 X524.922734 Y214.707266 Z-1.000000 I-0.892920 J42.696663
G03 X527.341927 Y215.186991 Z-1.000000 I-2.651338 J19.710077
G03 X528.907969 Y215.739098 Z-1.000000 I-2.611768 J9.905320
G03 X530.364480 Y216.534457 Z-1.000000 I-4.248358 J9.511154
G03 X531.914584 Y217.630789 Z-1.000000 I-14.436276 J22.055479
G03 X533.743704 Y219.375940 Z-1.000000 I-6.965265 J9.131542
G03 X535.649884 Y222.078147 Z-1.000000 I-15.046033 J12.637143
G03 X536.534679 Y223.817509 Z-1.000000 I-14.152022 J8.293704
G03 X537.077184 Y225.467365 Z-1.000000 I-8.537889 J3.721549
G03 X537.359012 Y227.228061 Z-1.000000 I-13.261532 J3.025621
G03 X537.587984 Y230.572918 Z-1.000000 I-67.364330 J6.291704
G03 X537.573798 Y235.315593 Z-1.000000 I-50.884129 J2.219148
G03 X537.166047 Y238.319893 Z-1.000000 I-17.764860 J-0.881269
G03 X536.196337 Y241.138879 Z-1.000000 I-13.121218 J-2.937320
G03 X534.630984 Y243.782238 Z-1.000000 I-15.068285 J-7.138017
G03 X532.271952 Y246.365627 Z-1.000000 I-12.549638 J-9.090982
G03 X529.537047 Y248.206646 Z-1.000000 I-8.683696 J-9.948053
G03 X526.423137 Y249.363658 Z-1.000000 I-6.720618 J-13.318648
G03 X522.218384 Y250.045651 Z-1.000000 I-6.182090 J-24.812055
G01 X517.365484 Y250.422041 Z-1.000000
G01 X527.515984 Y267.763226 Z-1.000000
G03 X532.327662 Y276.126846 Z-1.000000 I-554.747724 J324.718057
G03 X535.012472 Y281.016228 Z-1.000000 I-195.765779 J110.678818
G03 X537.504311 Y286.038119 Z-1.000000 I-84.843771 J45.228259
G03 X537.683084 Y286.835713 Z-1.000000 I-1.773018 J0.816238
G03 X536.782720 Y289.416821 Z-1.000000 I-4.039299 J0.038567
G03 X534.555734 Y291.106562 Z-1.000000 I-4.552459 J-3.687498
G03 X531.778492 Y291.594196 Z-1.000000 I-2.409825 J-5.572285
G03 X529.155584 Y290.523533 Z-1.000000 I0.266895 J-4.401966
G03 X528.672151 Y289.989302 Z-1.000000 I1.959794 J-2.259292
G03 X526.736997 Y287.120132 Z-1.000000 I66.064943 J-46.645675
G03 X524.948362 Y284.268134 Z-1.000000 I119.548499 J-76.961883
G03 X522.676584 Y280.459452 Z-1.000000 I180.510573 J-110.251637
G02 X517.432794 Y271.487480 Z-1.000000 I-2845.073504 J1656.823137
G02 X511.233084 Y260.950617 Z-1.000000 I-3939.894993 J2311.074529
G01 X505.019284 Y250.422041 Z-1.000000
G01 X501.850584 Y250.234843 Z-1.000000
G01 X498.681984 Y250.047645 Z-1.000000
G01 X498.681984 Y268.690608 Z-1.000000
G03 X498.659271 Y276.962713 Z-1.000000 I-1506.317876 J-0.000000
G03 X498.570403 Y281.696811 Z-1.000000 I-178.303083 J-0.979185
G03 X498.275654 Y286.354060 Z-1.000000 I-74.863057 J-2.399999
G03 X497.898909 Y288.527170 Z-1.000000 I-14.216695 J-1.345488
G03 X497.063852 Y290.382330 Z-1.000000 I-5.694569 J-1.447758
G03 X496.162378 Y291.168436 Z-1.000000 I-1.948151 J-1.324121
G03 X494.915494 Y291.518198 Z-1.000000 I-1.547892 J-3.120734
G03 X492.855684 Y291.607357 Z-1.000000 I-2.059810 J-23.748786
G03 X491.010061 Y291.229681 Z-1.000000 I0.000000 J-4.698421
G03 X489.500084 Y290.173954 Z-1.000000 I1.765147 J-4.132344
G01 X489.500084 Y290.173954 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X525.650984 Y238.063629

G01 Z-1.000000 F100.0(Penetrate)
G02 X526.446437 Y237.092718 Z-1.000000 I-4.490763 J-4.490522 F400.000000
G02 X526.902972 Y236.075298 Z-1.000000 I-2.815118 J-1.874331
G02 X527.085587 Y234.913764 Z-1.000000 I-5.684019 J-1.488757
G02 X527.151384 Y232.334845 Z-1.000000 I-42.635647 J-2.378088
G02 X527.059617 Y229.659580 Z-1.000000 I-45.196312 J0.211122
G02 X526.775205 Y228.074356 Z-1.000000 I-7.088160 J0.453595
G02 X526.069026 Y226.763594 Z-1.000000 I-3.536739 J1.059823
G02 X524.811097 Y225.797913 Z-1.000000 I-2.799498 J2.344572
G02 X523.193545 Y225.284603 Z-1.000000 I-2.794406 J6.000501
G02 X519.929208 Y224.886206 Z-1.000000 I-5.104910 J28.255348
G02 X516.575672 Y224.757038 Z-1.000000 I-4.194247 J65.295368
G02 X510.799684 Y224.719928 Z-1.000000 I-5.775988 J449.490076
G01 X498.681984 Y224.719928 Z-1.000000
G01 X498.681984 Y232.151864 Z-1.000000
G01 X498.681984 Y239.583801 Z-1.000000
G01 X511.406384 Y239.583801 Z-1.000000
G02 X520.225910 Y239.537565 Z-1.000000 I0.000000 J-841.184181
G02 X522.716947 Y239.391905 Z-1.000000 I-0.273069 J-26.043234
G02 X523.958416 Y239.170769 Z-1.000000 I-0.957450 J-8.970580
G02 X524.569658 Y238.940369 Z-1.000000 I-0.698338 J-2.778671
G02 X525.105147 Y238.575556 Z-1.000000 I-1.203735 J-2.342310
G02 X525.650984 Y238.063629 Z-1.000000 I-7.730068 J-8.789076
G01 X525.650984 Y238.063629 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X563.992984 Y290.087186

G01 Z-1.000000 F100.0(Penetrate)
G03 X563.406725 Y289.386401 Z-1.000000 I3.644415 J-3.644445 F400.000000
G03 X562.919347 Y288.564389 Z-1.000000 I5.399348 J-3.756811
G03 X562.567061 Y287.676657 Z-1.000000 I5.051415 J-2.518359
G03 X562.472784 Y287.037899 Z-1.000000 I2.116765 J-0.638757
G03 X562.580895 Y286.203384 Z-1.000000 I3.274906 J0.000000
G03 X565.473573 Y275.529869 Z-1.000000 I764.432705 J201.443696
G03 X568.339185 Y265.368854 Z-1.000000 I1574.647531 J438.597878
G03 X572.173589 Y252.045618 Z-1.000000 I3151.860046 J899.886449
G03 X576.063206 Y238.741947 Z-1.000000 I3164.065275 J917.862606
G03 X579.119305 Y228.490156 Z-1.000000 I1618.990182 J477.045971
G03 X582.417955 Y217.810778 Z-1.000000 I798.800910 J240.884761
G03 X582.857194 Y216.768493 Z-1.000000 I5.834346 J1.845013
G03 X583.922437 Y215.541063 Z-1.000000 I3.234580 J1.731217
G03 X585.545957 Y214.722601 Z-1.000000 I3.258379 J4.443936
G03 X587.358039 Y214.458078 Z-1.000000 I1.745551 J5.618703
G03 X589.127594 Y214.820483 Z-1.000000 I-0.053880 J4.764482
G03 X590.840112 Y215.786515 Z-1.000000 I-3.119375 J7.530761
G03 X591.543694 Y216.598772 Z-1.000000 I-1.443593 J1.961300
G03 X591.736916 Y217.060352 Z-1.000000 I-2.331165 J1.247080
G03 X594.317994 Y225.490977 Z-1.000000 I-920.337799 J286.376494
G03 X596.741590 Y233.601359 Z-1.000000 I-1893.192788 J570.153161
G03 X600.237794 Y245.467417 Z-1.000000 I-3472.317567 J1029.528463
G02 X608.401739 Y273.335951 Z-1.000000 I90343.441252 J-26450.623254
G02 X610.208294 Y279.483837 Z-1.000000 I4528.402144 J-1327.329571
G03 X610.948092 Y282.119369 Z-1.000000 I-102.161120 J30.098370
G03 X611.487194 Y284.256220 Z-1.000000 I-77.848591 J20.776717
G03 X611.959077 Y286.462208 Z-1.000000 I-48.301064 J11.485545
G03 X612.019094 Y287.098761 Z-1.000000 I-3.345696 J0.636553
G03 X611.608499 Y288.675243 Z-1.000000 I-3.231749 J-0.000000
G03 X610.385969 Y290.182459 Z-1.000000 I-4.951898 J-2.767156
G03 X608.721824 Y291.225754 Z-1.000000 I-4.118244 J-4.720078
G03 X606.851594 Y291.607357 Z-1.000000 I-1.870230 J-4.392175
G03 X605.565081 Y291.516884 Z-1.000000 I-0.000000 J-9.192281
G03 X604.606357 Y291.257173 Z-1.000000 I0.561411 J-3.971861
G03 X603.767297 Y290.761096 Z-1.000000 I1.275822 J-3.115540
G03 X602.941994 Y289.899050 Z-1.000000 I2.937607 J-3.638481
G03 X601.816921 Y287.826602 Z-1.000000 I6.881079 J-5.077152
G03 X599.994194 Y282.313566 Z-1.000000 I68.441467 J-25.686040
G01 X597.950494 Y275.195163 Z-1.000000
G01 X587.230694 Y275.195163 Z-1.000000
G01 X576.510894 Y275.195163 Z-1.000000
G01 X576.037494 Y276.743484 Z-1.000000
G02 X575.568827 Y278.293198 Z-1.000000 I260.464607 J79.615839
G02 X573.944494 Y283.739569 Z-1.000000 I3157.858534 J944.769369
G03 X573.187937 Y286.046276 Z-1.000000 I-41.684073 J-12.394182
G03 X572.445119 Y287.942495 Z-1.000000 I-33.604788 J-12.070614
G03 X571.552489 Y289.807266 Z-1.000000 I-21.354764 J-9.076097
G03 X571.114794 Y290.397345 Z-1.000000 I-2.208815 J-1.181028
G03 X569.563780 Y291.348970 Z-1.000000 I-2.762332 J-2.762433
G03 X567.613664 Y291.615578 Z-1.000000 I-1.710148 J-5.243534
G03 X565.673264 Y291.196655 Z-1.000000 I0.256589 J-5.891791
G03 X563.992984 Y290.087186 Z-1.000000 I1.870924 J-4.660614
G01 X563.992984 Y290.087186 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X594.447694 Y263.273097

G01 Z-1.000000 F100.0(Penetrate)
G02 X592.244350 Y255.759215 Z-1.000000 I-1064.980317 J308.211144 F400.000000
G02 X590.003207 Y248.287569 Z-1.000000 I-1415.084164 J420.386867
G02 X587.559749 Y240.344035 Z-1.000000 I-911.789519 J276.121906
G02 X587.384894 Y239.950945 Z-1.000000 I-1.777697 J0.555325
G02 X587.291411 Y239.958963 Z-1.000000 I-0.044812 J0.026506
G02 X586.089591 Y243.294094 Z-1.000000 I96.541213 J36.672897
G02 X584.982609 Y246.615668 Z-1.000000 I236.414076 J80.635077
G02 X583.508394 Y251.212616 Z-1.000000 I527.270546 J171.627447
G02 X582.076019 Y255.814050 Z-1.000000 I547.306318 J172.894096
G02 X580.974673 Y259.485672 Z-1.000000 I297.398504 J91.209143
G02 X579.901206 Y263.323035 Z-1.000000 I153.046821 J44.882231
G02 X579.821794 Y263.892425 Z-1.000000 I2.288027 J0.609341
G02 X579.920035 Y264.005743 Z-1.000000 I0.114474 J0.000000
G02 X582.013307 Y264.220476 Z-1.000000 I3.725015 J-26.002179
G02 X584.108028 Y264.313949 Z-1.000000 I3.707704 J-59.570968
G02 X587.287694 Y264.356921 Z-1.000000 I3.179666 J-117.616257
G02 X592.142386 Y264.305670 Z-1.000000 I0.000000 J-229.952987
G02 X593.534244 Y264.186133 Z-1.000000 I-0.228573 J-10.824421
G02 X594.195824 Y264.011798 Z-1.000000 I-0.477523 J-3.154612
G02 X594.409451 Y263.859959 Z-1.000000 I-0.187032 J-0.489338
G02 X594.496505 Y263.623196 Z-1.000000 I-0.283111 J-0.238482
G02 X594.447694 Y263.273097 Z-1.000000 I-1.238719 J-0.005750
G01 X594.447694 Y263.273097 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X638.307494 Y290.082305

G01 Z-1.000000 F100.0(Penetrate)
G01 X636.782494 Y288.557253 Z-1.000000 F400.000000
G01 X636.942194 Y251.529132 Z-1.000000
G01 X637.101894 Y214.501015 Z-1.000000
G01 X643.342894 Y214.325702 Z-1.000000
G01 X649.583794 Y214.150382 Z-1.000000
G01 X655.419794 Y222.529470 Z-1.000000
G02 X657.926840 Y226.081997 Z-1.000000 I288.163805 J-200.698571
G02 X659.646869 Y228.447885 Z-1.000000 I150.817571 J-107.838003
G02 X661.486556 Y230.848462 Z-1.000000 I70.013010 J-51.749386
G02 X661.614794 Y230.910886 Z-1.000000 I0.129310 J-0.102711
G02 X661.742972 Y230.849453 Z-1.000000 I0.000987 J-0.162376
G02 X663.574382 Y228.456475 Z-1.000000 I-67.019915 J-53.189420
G02 X665.286243 Y226.096913 Z-1.000000 I-146.756133 J-108.272271
G02 X667.781494 Y222.552280 Z-1.000000 I-281.421396 J-200.757763
G01 X673.588994 Y214.191351 Z-1.000000
G01 X679.963694 Y214.191351 Z-1.000000
G01 X686.338494 Y214.191351 Z-1.000000
G01 X686.338494 Y251.552741 Z-1.000000
G01 X686.338494 Y288.914129 Z-1.000000
G01 X684.905094 Y290.260746 Z-1.000000
G03 X681.077048 Y291.693939 Z-1.000000 I-3.628241 J-3.862065
G03 X677.334694 Y289.975087 Z-1.000000 I0.202815 J-5.375003
G01 X675.821894 Y288.342817 Z-1.000000
G01 X675.661094 Y259.219446 Z-1.000000
G01 X675.500194 Y230.096085 Z-1.000000
G01 X671.010094 Y236.481011 Z-1.000000
G02 X668.445602 Y240.268919 Z-1.000000 I105.323068 J74.067961
G02 X667.394494 Y242.078044 Z-1.000000 I17.182325 J11.192900
G02 X666.685853 Y243.924668 Z-1.000000 I8.092499 J4.164778
G02 X666.519994 Y245.205546 Z-1.000000 I4.863002 J1.280878
G03 X663.497731 Y249.684874 Z-1.000000 I-4.830561 J-0.000000
G03 X658.130894 Y248.592205 Z-1.000000 I-1.857739 J-4.601800
G03 X657.544882 Y247.891248 Z-1.000000 I3.630938 J-3.630969
G03 X657.057257 Y247.067342 Z-1.000000 I5.401943 J-3.753365
G03 X656.705126 Y246.177658 Z-1.000000 I5.064019 J-2.518833
G03 X656.610694 Y245.535883 Z-1.000000 I2.133597 J-0.641775
G02 X656.419143 Y243.916313 Z-1.000000 I-6.942515 J0.000000
G02 X655.644869 Y241.812427 Z-1.000000 I-9.322753 J2.236548
G02 X654.449349 Y239.749899 Z-1.000000 I-19.163393 J9.730086
G02 X651.407094 Y235.287436 Z-1.000000 I-128.535072 J84.359703
G01 X647.320794 Y229.514233 Z-1.000000
G01 X647.320794 Y259.040622 Z-1.000000
G01 X647.320794 Y288.567020 Z-1.000000
G01 X645.800594 Y290.087186 Z-1.000000
G03 X642.056580 Y291.636609 Z-1.000000 I-3.740528 J-3.740370
G03 X638.307494 Y290.082186 Z-1.000000 I0.003498 J-5.306823
G01 X638.307494 Y290.082305 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X711.261694 Y253.054183

G01 Z-1.000000 F100.0(Penetrate)
G01 X711.421294 Y214.501015 Z-1.000000 F400.000000
G01 X734.125094 Y214.338609 Z-1.000000
G01 X756.828794 Y214.176208 Z-1.000000
G01 X758.743294 Y215.787155 Z-1.000000
G03 X760.206680 Y217.607347 Z-1.000000 I-3.896923 J4.631380
G03 X760.708094 Y219.432704 Z-1.000000 I-3.930091 J2.061118
G03 X760.397841 Y221.294639 Z-1.000000 I-4.352518 J0.231558
G03 X759.137694 Y223.199756 Z-1.000000 I-5.304961 J-2.139663
G01 X757.617494 Y224.719928 Z-1.000000
G01 X739.628794 Y224.719928 Z-1.000000
G01 X721.640194 Y224.719928 Z-1.000000
G01 X721.640194 Y236.177497 Z-1.000000
G01 X721.640194 Y247.635065 Z-1.000000
G01 X729.102894 Y247.635065 Z-1.000000
G03 X735.018127 Y247.842902 Z-1.000000 I0.000000 J84.280237
G03 X737.533694 Y248.296766 Z-1.000000 I-0.830220 J11.799807
G03 X738.734287 Y248.784387 Z-1.000000 I-1.853482 J6.285368
G03 X739.539801 Y249.344453 Z-1.000000 I-1.867555 J3.545301
G03 X740.187924 Y250.086381 Z-1.000000 I-2.689198 J3.003246
G03 X740.737894 Y251.066086 Z-1.000000 I-5.543035 J3.755879
G03 X741.153320 Y252.540443 Z-1.000000 I-4.576113 J2.085103
G03 X741.080994 Y253.916088 Z-1.000000 I-3.942817 J0.482428
G03 X740.565252 Y255.216851 Z-1.000000 I-4.541652 J-1.048105
G03 X739.421594 Y256.770152 Z-1.000000 I-7.836236 J-4.571955
G03 X738.971633 Y257.216459 Z-1.000000 I-4.844872 J-4.434564
G03 X738.518668 Y257.541206 Z-1.000000 I-1.682578 J-1.868627
G03 X738.005614 Y257.760200 Z-1.000000 I-1.271027 J-2.267247
G03 X737.120132 Y257.958737 Z-1.000000 I-1.757686 J-5.765415
G03 X735.341662 Y258.106772 Z-1.000000 I-2.078243 J-14.210603
G03 X729.893194 Y258.163641 Z-1.000000 I-5.448468 J-260.972533
G01 X721.640194 Y258.163641 Z-1.000000
G01 X721.640194 Y269.594406 Z-1.000000
G01 X721.640194 Y281.025171 Z-1.000000
G01 X739.655094 Y281.206808 Z-1.000000
G01 X757.669994 Y281.388444 Z-1.000000
G01 X759.163894 Y283.126130 Z-1.000000
G03 X760.618954 Y286.999842 Z-1.000000 I-4.565329 J3.924982
G03 X759.224894 Y290.260746 Z-1.000000 I-4.422787 J0.037665
G03 X758.725364 Y290.670863 Z-1.000000 I-3.315207 J-3.528709
G03 X758.005269 Y291.033693 Z-1.000000 I-1.529591 J-2.139732
G03 X757.166360 Y291.225046 Z-1.000000 I-1.496692 J-4.627028
G03 X754.972219 Y291.434087 Z-1.000000 I-3.664268 J-26.841468
G03 X750.598475 Y291.564472 Z-1.000000 I-4.848161 J-89.206702
G03 X734.447094 Y291.607357 Z-1.000000 I-16.151380 J-3041.460655
G01 X711.102694 Y291.607357 Z-1.000000
G01 X711.261694 Y253.054183 Z-1.000000
G00 Z5.000000

(End cutting path id: path2311)         FRAME




(Start cutting path id: path2311)               WORK/
(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X420.197784 Y383.732429

G01 Z-1.000000 F100.0(Penetrate)
G03 X419.027972 Y382.253257 Z-1.000000 I9.206992 J-8.483570 F400.000000
G03 X416.022484 Y377.546809 Z-1.000000 I90.972900 J-61.407184
G01 X413.131284 Y372.754669 Z-1.000000
G01 X413.327884 Y342.038943 Z-1.000000
G01 X413.524384 Y311.323251 Z-1.000000
G01 X415.546984 Y309.516566 Z-1.000000
G03 X416.443485 Y308.786199 Z-1.000000 I8.484927 J9.499566
G03 X417.254534 Y308.240599 Z-1.000000 I6.325454 J8.527324
G03 X418.127975 Y307.790425 Z-1.000000 I3.636351 J5.982937
G03 X418.519584 Y307.709887 Z-1.000000 I0.391609 J0.911816
G03 X419.876750 Y307.968885 Z-1.000000 I0.000000 J3.685315
G03 X421.585334 Y308.860539 Z-1.000000 I-3.423351 J8.642635
G03 X423.068831 Y310.088041 Z-1.000000 I-5.004058 J7.557843
G03 X423.769984 Y311.186175 Z-1.000000 I-2.291648 J2.236115
G03 X423.923774 Y311.929715 Z-1.000000 I-2.236073 J0.850171
G03 X424.196384 Y320.449910 Z-1.000000 I-337.846485 J15.074125
G03 X424.310204 Y328.744320 Z-1.000000 I-732.729614 J14.202861
G03 X424.362684 Y341.763057 Z-1.000000 I-1614.754321 J13.018736
G02 X424.413047 Y360.549964 Z-1.000000 I3504.063914 J0.000000
G02 X424.533409 Y366.508745 Z-1.000000 I200.882539 J-1.077040
G02 X424.702213 Y369.529135 Z-1.000000 I72.683914 J-2.547248
G02 X424.859569 Y370.846316 Z-1.000000 I15.649259 J-1.201534
G02 X425.166001 Y372.064749 Z-1.000000 I7.285084 J-1.184429
G02 X425.446484 Y372.685249 Z-1.000000 I3.468935 J-1.194410
G02 X426.142109 Y373.700516 Z-1.000000 I6.429600 J-3.659388
G02 X426.719659 Y374.205927 Z-1.000000 I1.779341 J-1.450617
G02 X427.429347 Y374.483151 Z-1.000000 I1.122933 J-1.827681
G02 X428.552184 Y374.593429 Z-1.000000 I1.134265 J-5.777484
G02 X429.678421 Y374.484068 Z-1.000000 I0.010647 J-5.744198
G02 X430.415359 Y374.194342 Z-1.000000 I-0.431555 J-2.179781
G02 X431.022365 Y373.669231 Z-1.000000 I-1.328957 J-2.149606
G02 X431.801384 Y372.584529 Z-1.000000 I-6.798267 J-5.704528
G02 X432.164473 Y371.907952 Z-1.000000 I-5.963279 J-3.635936
G02 X432.475741 Y371.054703 Z-1.000000 I-4.694177 J-2.195850
G02 X432.661061 Y370.140345 Z-1.000000 I-7.260995 J-1.947603
G02 X432.844284 Y368.341779 Z-1.000000 I-24.632253 J-3.417954
G02 X432.977034 Y364.793418 Z-1.000000 I-63.804228 J-4.163683
G02 X433.030984 Y353.742519 Z-1.000000 I-1165.263247 J-11.214351
G01 X433.030984 Y336.913341 Z-1.000000
G01 X434.634284 Y335.007842 Z-1.000000
G03 X436.421707 Y333.572376 Z-1.000000 I4.534302 J3.815479
G03 X438.247422 Y333.059044 Z-1.000000 I2.129697 J4.071137
G03 X440.111241 Y333.338741 Z-1.000000 I0.292011 J4.403932
G03 X441.940284 Y334.523404 Z-1.000000 I-1.855925 J4.869725
G03 X442.389067 Y335.025474 Z-1.000000 I-4.256277 J4.256146
G03 X442.757673 Y335.650796 Z-1.000000 I-2.021468 J1.612890
G03 X442.976048 Y336.372417 Z-1.000000 I-3.561748 J1.471699
G03 X443.221972 Y338.055679 Z-1.000000 I-16.427179 J3.259596
G03 X443.428764 Y341.418648 Z-1.000000 I-51.065564 J4.827918
G03 X443.724484 Y353.394952 Z-1.000000 I-1544.354880 J44.125151
G02 X444.007423 Y364.266620 Z-1.000000 I1042.367834 J-21.688463
G02 X444.216197 Y368.157237 Z-1.000000 I87.033495 J-2.719377
G02 X444.747213 Y371.603730 Z-1.000000 I22.633630 J-1.723106
G02 X445.195784 Y372.721399 Z-1.000000 I3.874780 J-0.906276
G02 X445.900230 Y373.712428 Z-1.000000 I6.210023 J-3.668344
G02 X446.490584 Y374.213953 Z-1.000000 I1.808307 J-1.530371
G02 X447.209164 Y374.489567 Z-1.000000 I1.144460 J-1.909288
G02 X448.325384 Y374.597339 Z-1.000000 I1.116220 J-5.726566
G02 X449.497638 Y374.478686 Z-1.000000 I0.000000 J-5.850078
G02 X450.209247 Y374.158824 Z-1.000000 I-0.357103 J-1.745962
G02 X450.764423 Y373.571112 Z-1.000000 I-1.426679 J-1.903775
G02 X451.598184 Y372.144709 Z-1.000000 I-12.030754 J-7.989094
G02 X451.963467 Y371.324776 Z-1.000000 I-7.599173 J-3.876791
G02 X452.301384 Y370.149135 Z-1.000000 I-6.602853 J-2.534257
G02 X452.483559 Y368.876075 Z-1.000000 I-12.338704 J-2.415231
G02 X452.675109 Y365.938966 Z-1.000000 I-56.714375 J-5.173559
G02 X452.802745 Y360.137483 Z-1.000000 I-167.750755 J-6.592744
G02 X452.849484 Y340.304336 Z-1.000000 I-4207.911146 J-19.833147
G01 X452.849484 Y310.916637 Z-1.000000
G01 X454.754984 Y309.313259 Z-1.000000
G03 X456.726359 Y308.084726 Z-1.000000 I5.275525 J6.269450
G03 X458.255759 Y307.786864 Z-1.000000 I1.422297 J3.227566
G03 X459.756249 Y308.183292 Z-1.000000 I-0.104283 J3.432618
G03 X461.567684 Y309.520232 Z-1.000000 I-3.529631 J6.677976
G01 X463.377984 Y311.330571 Z-1.000000
G01 X463.377984 Y342.322651 Z-1.000000
G01 X463.377984 Y373.314749 Z-1.000000
G01 X460.510984 Y378.136509 Z-1.000000
G03 X458.957301 Y380.720858 Z-1.000000 I-267.595537 J-159.116393
G03 X457.975131 Y382.206557 Z-1.000000 I-20.035233 J-12.177453
G03 X456.857458 Y383.526020 Z-1.000000 I-8.719099 J-6.252557
G03 X455.872384 Y384.278283 Z-1.000000 I-3.258669 J-3.246064
G03 X454.722199 Y384.740258 Z-1.000000 I-2.274317 J-3.999595
G03 X453.016038 Y385.026475 Z-1.000000 I-2.624460 J-10.416232
G03 X451.236276 Y385.111319 Z-1.000000 I-2.006980 J-23.390637
G03 X448.219384 Y385.125919 Z-1.000000 I-3.016892 J-311.695204
G03 X443.947455 Y385.060215 Z-1.000000 I-0.000000 J-138.908192
G03 X442.209784 Y384.909493 Z-1.000000 I0.481216 J-15.640068
G03 X440.657830 Y384.474723 Z-1.000000 I0.875765 J-6.113441
G03 X439.697284 Y383.905109 Z-1.000000 I1.676962 J-3.922575
G02 X438.617335 Y383.120272 Z-1.000000 I-10.847100 J13.790351
G02 X438.201959 Y382.993964 Z-1.000000 I-0.415376 J0.619851
G02 X437.813898 Y383.140827 Z-1.000000 I-0.000000 J0.586126
G02 X437.040384 Y383.905109 Z-1.000000 I6.453665 J7.305187
G03 X436.314304 Y384.466309 Z-1.000000 I-1.929699 J-1.746344
G03 X434.909147 Y384.921268 Z-1.000000 I-2.046689 J-3.923852
G03 X433.253918 Y385.064411 Z-1.000000 I-2.035014 J-13.890109
G03 X428.707584 Y385.125919 Z-1.000000 I-4.546333 J-167.989622
G03 X423.999073 Y385.065778 Z-1.000000 I0.000000 J-184.345649
G03 X422.384684 Y384.913731 Z-1.000000 I0.303163 J-11.865453
G03 X421.043485 Y384.422498 Z-1.000000 I0.614678 J-3.754773
G03 X420.195484 Y383.732429 Z-1.000000 I1.969620 J-3.286470
G01 X420.197784 Y383.732429 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X507.042984 Y383.988649

G01 Z-1.000000 F100.0(Penetrate)
G03 X504.700075 Y382.630854 Z-1.000000 I7.109206 J-14.967385 F400.000000
G03 X503.224784 Y381.336819 Z-1.000000 I4.963658 J-7.146904
G03 X502.710164 Y380.632899 Z-1.000000 I3.036009 J-2.759634
G03 X500.325972 Y376.357172 Z-1.000000 I131.202615 J-75.962587
G03 X498.117004 Y372.153834 Z-1.000000 I241.542465 J-129.619243
G03 X495.239484 Y366.456369 Z-1.000000 I377.500103 J-194.232774
G03 X490.917542 Y357.579628 Z-1.000000 I626.273059 J-310.412749
G03 X489.516009 Y354.327648 Z-1.000000 I40.848791 J-19.532951
G03 X488.639818 Y351.213082 Z-1.000000 I16.155054 J-6.225277
G03 X488.299584 Y348.446666 Z-1.000000 I23.535466 J-4.298689
G03 X488.352647 Y343.880510 Z-1.000000 I30.128478 J-1.933266
G03 X489.003384 Y340.404760 Z-1.000000 I18.045807 J1.579783
G03 X489.592008 Y338.731934 Z-1.000000 I15.372974 J4.469373
G03 X491.390280 Y334.599648 Z-1.000000 I124.596128 J51.763814
G03 X493.267100 Y330.583960 Z-1.000000 I200.788855 J91.396659
G03 X495.578647 Y325.856627 Z-1.000000 I314.797068 J150.999044
G03 X497.963595 Y321.161769 Z-1.000000 I306.221618 J152.604776
G03 X499.944558 Y317.434379 Z-1.000000 I173.825175 J89.991233
G03 X502.080482 Y313.692100 Z-1.000000 I97.721850 J53.294534
G03 X502.864084 Y312.591585 Z-1.000000 I7.190343 J4.290530
G03 X504.680055 Y310.935968 Z-1.000000 I7.140414 J6.008257
G03 X507.550672 Y309.258064 Z-1.000000 I12.136381 J17.468812
G03 X510.650806 Y308.054330 Z-1.000000 I9.024770 J18.648727
G03 X512.926584 Y307.721988 Z-1.000000 I2.275778 J7.625751
G03 X515.296262 Y308.074009 Z-1.000000 I0.000000 J8.151929
G03 X518.417097 Y309.316468 Z-1.000000 I-5.823463 J19.168212
G03 X521.286891 Y311.062196 Z-1.000000 I-8.970459 J17.978183
G03 X523.008884 Y312.752697 Z-1.000000 I-4.888327 J6.701677
G03 X523.637308 Y313.707397 Z-1.000000 I-6.854156 J5.195870
G03 X525.983559 Y317.978221 Z-1.000000 I-159.409644 J90.354347
G03 X528.178516 Y322.179501 Z-1.000000 I-274.761924 J146.223259
G03 X530.921884 Y327.621618 Z-1.000000 I-392.465719 J201.254282
G03 X535.016680 Y336.046042 Z-1.000000 I-554.291860 J274.627836
G03 X536.367059 Y339.197408 Z-1.000000 I-38.668342 J18.434618
G03 X537.212527 Y342.229699 Z-1.000000 I-15.851733 J6.053817
G03 X537.544584 Y344.962803 Z-1.000000 I-23.319757 J4.219945
G03 X537.600153 Y347.186974 Z-1.000000 I-29.390276 J1.847060
G03 X537.454111 Y349.257910 Z-1.000000 I-18.030974 J-0.230921
G03 X537.071180 Y351.300432 Z-1.000000 I-17.986330 J-2.314907
G03 X536.297297 Y353.886959 Z-1.000000 I-27.272804 J-6.750930
G03 X534.282435 Y358.740003 Z-1.000000 I-45.296363 J-15.961135
G03 X528.918384 Y369.443339 Z-1.000000 I-330.584006 J-158.978647
G03 X523.543365 Y379.155324 Z-1.000000 I-200.158332 J-104.432634
G03 X521.077459 Y382.372781 Z-1.000000 I-14.155156 J-8.295038
G03 X519.457033 Y383.665793 Z-1.000000 I-6.193212 J-6.099563
G03 X517.862250 Y384.395322 Z-1.000000 I-3.701966 J-5.984769
G03 X516.106179 Y384.771038 Z-1.000000 I-2.885116 J-9.193089
G03 X513.621584 Y384.974999 Z-1.000000 I-4.110335 J-34.835626
G03 X511.395182 Y384.996602 Z-1.000000 I-1.407705 J-30.341611
G03 X509.879359 Y384.863031 Z-1.000000 I0.336764 J-12.489628
G03 X508.412136 Y384.525291 Z-1.000000 I1.443138 J-9.625187
G03 X507.042984 Y383.988649 Z-1.000000 I3.149115 J-10.049354
G01 X507.042984 Y383.988649 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X520.885584 Y361.591439

G01 Z-1.000000 F100.0(Penetrate)
G02 X525.036805 Y353.222711 Z-1.000000 I-587.138769 J-296.458033 F400.000000
G02 X526.297447 Y350.346935 Z-1.000000 I-35.086795 J-17.095047
G02 X526.810244 Y348.825875 Z-1.000000 I-15.663580 J-6.127659
G02 X527.010492 Y347.876898 Z-1.000000 I-6.719432 J-1.913514
G02 X527.068494 Y346.918483 Z-1.000000 I-6.097983 J-0.850001
G02 X527.012584 Y345.929513 Z-1.000000 I-12.745253 J0.224460
G02 X526.688623 Y344.487098 Z-1.000000 I-5.873720 J0.561628
G02 X524.975788 Y340.148567 Z-1.000000 I-74.421118 J26.873802
G02 X523.101831 Y335.975832 Z-1.000000 I-126.536510 J54.319805
G02 X520.707984 Y331.030344 Z-1.000000 I-205.672471 J96.502959
G02 X518.188992 Y326.137898 Z-1.000000 I-195.710368 J97.671425
G02 X516.168906 Y322.501659 Z-1.000000 I-96.520200 J51.241860
G02 X515.123213 Y320.763253 Z-1.000000 I-56.898684 J33.042199
G02 X514.410055 Y319.685524 Z-1.000000 I-19.863900 J12.369588
G02 X513.609309 Y318.679279 Z-1.000000 I-8.779786 J6.165012
G02 X513.318284 Y318.489329 Z-1.000000 I-0.501956 J0.451139
G02 X512.657904 Y318.400060 Z-1.000000 I-0.570693 J1.734520
G02 X511.962286 Y318.609593 Z-1.000000 I0.074019 J1.505174
G02 X511.365984 Y319.097840 Z-1.000000 I1.316206 J2.215759
G02 X510.350022 Y320.410684 Z-1.000000 I8.110283 J7.325781
G02 X508.686708 Y323.273588 Z-1.000000 I24.586406 J16.199050
G02 X504.302484 Y332.173345 Z-1.000000 I358.464422 J182.117522
G02 X501.062420 Y339.135358 Z-1.000000 I353.027100 J168.530906
G02 X499.772372 Y342.221561 Z-1.000000 I51.062848 J23.157250
G02 X498.859978 Y345.241149 Z-1.000000 I20.285757 J7.777151
G02 X498.683484 Y346.829817 Z-1.000000 I7.095061 J1.592365
G02 X498.866619 Y348.430908 Z-1.000000 I7.064945 J0.002922
G02 X499.837084 Y351.504492 Z-1.000000 I20.236398 J-4.699518
G02 X501.226078 Y354.670276 Z-1.000000 I53.491011 J-21.581673
G02 X504.884384 Y362.198469 Z-1.000000 I424.281360 J-201.525492
G03 X507.939786 Y368.357110 Z-1.000000 I-4287.451340 J2130.911184
G02 X509.629706 Y371.485091 Z-1.000000 I44.862527 J-22.216902
G02 X510.590985 Y372.977698 Z-1.000000 I18.534087 J-10.880610
G02 X511.270668 Y373.789500 Z-1.000000 I5.185763 J-3.651355
G02 X512.079117 Y374.382346 Z-1.000000 I2.271310 J-2.249675
G02 X512.714734 Y374.523604 Z-1.000000 I0.586989 J-1.140592
G02 X513.821368 Y374.016609 Z-1.000000 I-0.060439 J-1.593161
G02 X515.884888 Y371.247792 Z-1.000000 I-11.614350 J-10.809189
G02 X517.662921 Y367.984229 Z-1.000000 I-49.184229 J-28.912361
G03 X520.885584 Y361.591439 Z-1.000000 I4893.758650 J2462.979784
G01 X520.885584 Y361.591439 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X563.751584 Y383.487929

G01 Z-1.000000 F100.0(Penetrate)
G03 X563.389109 Y382.936871 Z-1.000000 I3.064568 J-2.410555 F400.000000
G03 X563.014439 Y381.922526 Z-1.000000 I3.327311 J-1.805382
G03 X562.839861 Y380.737925 Z-1.000000 I9.126734 J-1.950195
G03 X562.649172 Y377.317876 Z-1.000000 I64.598051 J-5.317091
G03 X562.553333 Y370.513313 Z-1.000000 I223.938727 J-6.557022
G03 X562.622784 Y344.934729 Z-1.000000 I7942.277934 J8.775621
G01 X562.782484 Y308.019551 Z-1.000000
G01 X580.433294 Y307.841711 Z-1.000000
G03 X589.502882 Y307.808806 Z-1.000000 I7.086558 J703.320739
G03 X594.464519 Y307.883361 Z-1.000000 I-0.573122 J203.278679
G03 X599.369036 Y308.151486 Z-1.000000 I-2.442168 J89.662165
G03 X600.745994 Y308.394994 Z-1.000000 I-0.616416 J7.500510
G03 X605.954532 Y311.008452 Z-1.000000 I-3.943114 J14.355453
G03 X609.964894 Y315.580208 Z-1.000000 I-9.966502 J12.787485
G03 X610.937032 Y317.468068 Z-1.000000 I-16.390371 J9.634314
G03 X611.452194 Y319.160786 Z-1.000000 I-6.520618 J2.909238
G03 X611.640466 Y321.028295 Z-1.000000 I-13.084194 J2.262320
G03 X611.709494 Y325.670400 Z-1.000000 I-156.054907 J4.642106
G03 X611.638975 Y330.013543 Z-1.000000 I-133.777277 J-0.000000
G03 X611.471744 Y331.966558 Z-1.000000 I-18.509403 J-0.601234
G03 X611.032668 Y333.776852 Z-1.000000 I-9.278779 J-1.292121
G03 X610.372694 Y335.269985 Z-1.000000 I-9.802337 J-3.440265
G03 X607.874216 Y338.789960 Z-1.000000 I-15.550615 J-8.391131
G03 X604.934882 Y341.190200 Z-1.000000 I-9.614342 J-8.773847
G03 X601.419585 Y342.697517 Z-1.000000 I-7.373488 J-12.343356
G03 X596.475994 Y343.571415 Z-1.000000 I-6.976170 J-25.044026
G01 X591.768694 Y343.940578 Z-1.000000
G01 X601.893894 Y361.250579 Z-1.000000
G03 X606.837051 Y369.846074 Z-1.000000 I-580.629821 J339.631728
G03 X609.454144 Y374.622019 Z-1.000000 I-179.002347 J101.193643
G03 X611.840802 Y379.493002 Z-1.000000 I-73.837682 J39.198773
G03 X612.019094 Y380.323089 Z-1.000000 I-1.843196 J0.830087
G03 X611.162643 Y382.819055 Z-1.000000 I-4.065237 J-0.000000
G03 X609.021607 Y384.551752 Z-1.000000 I-4.781758 J-3.719509
G03 X606.321474 Y385.173792 Z-1.000000 I-2.711365 J-5.598059
G03 X603.739694 Y384.333139 Z-1.000000 I-0.007874 J-4.360673
G03 X603.338011 Y383.909142 Z-1.000000 I0.957345 J-1.309233
G03 X599.567919 Y377.957021 Z-1.000000 I232.047534 J-151.149661
G03 X596.042618 Y372.152373 Z-1.000000 I473.199488 J-291.358210
G03 X591.060394 Y363.740469 Z-1.000000 I846.975164 J-507.329973
G01 X579.465394 Y343.940578 Z-1.000000
G01 X576.233394 Y343.753380 Z-1.000000
G01 X573.001394 Y343.566182 Z-1.000000
G01 X573.001394 Y362.825879 Z-1.000000
G01 X573.001394 Y382.085579 Z-1.000000
G01 X571.481194 Y383.605749 Z-1.000000
G03 X569.565088 Y384.840299 Z-1.000000 I-3.811426 J-3.811356
G03 X567.459489 Y385.206151 Z-1.000000 I-2.018125 J-5.372851
G03 X565.392328 Y384.757146 Z-1.000000 I0.081687 J-5.359054
G03 X563.751584 Y383.487929 Z-1.000000 I1.708367 J-3.903563
G01 X563.751584 Y383.487929 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X599.970294 Y331.582165

G01 Z-1.000000 F100.0(Penetrate)
G02 X600.765555 Y330.608781 Z-1.000000 I-4.434048 J-4.434201 F400.000000
G02 X601.228507 Y329.566887 Z-1.000000 I-2.868378 J-1.898325
G02 X601.413811 Y328.374360 Z-1.000000 I-6.004732 J-1.543721
G02 X601.490494 Y325.668982 Z-1.000000 I-47.684097 J-2.705378
G02 X601.414523 Y322.901455 Z-1.000000 I-50.446408 J0.000000
G02 X601.224369 Y321.742713 Z-1.000000 I-5.450472 J0.299468
G02 X600.740713 Y320.750510 Z-1.000000 I-2.920278 J0.809528
G02 X599.851694 Y319.757211 Z-1.000000 I-5.749508 J4.251408
G02 X599.265703 Y319.246909 Z-1.000000 I-8.535911 J9.210381
G02 X598.700169 Y318.883428 Z-1.000000 I-1.866146 J2.281811
G02 X598.063854 Y318.651615 Z-1.000000 I-1.391202 J2.829539
G02 X596.814844 Y318.432079 Z-1.000000 I-2.241691 J9.090922
G02 X594.315029 Y318.285613 Z-1.000000 I-2.784190 J26.113184
G02 X585.607194 Y318.238464 Z-1.000000 I-8.707835 J804.096537
G01 X573.001394 Y318.238464 Z-1.000000
G01 X573.001394 Y325.670400 Z-1.000000
G01 X573.001394 Y333.102337 Z-1.000000
G01 X585.725794 Y333.102337 Z-1.000000
G02 X594.545265 Y333.056101 Z-1.000000 I-0.000000 J-841.178533
G02 X597.036307 Y332.910441 Z-1.000000 I-0.273070 J-26.043281
G02 X598.277777 Y332.689304 Z-1.000000 I-0.957448 J-8.970559
G02 X598.889012 Y332.458905 Z-1.000000 I-0.698315 J-2.778571
G02 X599.424489 Y332.094089 Z-1.000000 I-1.203663 J-2.342135
G02 X599.970294 Y331.582165 Z-1.000000 I-7.728498 J-8.786926
G01 X599.970294 Y331.582165 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X638.312394 Y383.605749

G01 Z-1.000000 F100.0(Penetrate)
G01 X636.792194 Y382.085579 Z-1.000000 F400.000000
G01 X636.792194 Y346.708064 Z-1.000000
G01 X636.792194 Y311.330571 Z-1.000000
G01 X638.602594 Y309.520232 Z-1.000000
G03 X640.399675 Y308.192295 Z-1.000000 I5.321229 J5.321198
G03 X641.896269 Y307.782227 Z-1.000000 I1.665470 J3.142300
G03 X643.419433 Y308.053592 Z-1.000000 I0.165503 J3.481443
G03 X645.273494 Y309.203105 Z-1.000000 I-2.754423 J6.512612
G01 X647.011194 Y310.696318 Z-1.000000
G01 X647.320794 Y322.832664 Z-1.000000
G01 X647.630494 Y334.969010 Z-1.000000
G01 X656.301094 Y327.478640 Z-1.000000
G03 X673.515082 Y312.789232 Z-1.000000 I1219.033011 J1411.109881
G03 X678.203369 Y309.312569 Z-1.000000 I30.218202 J35.849899
G03 X680.761922 Y308.026890 Z-1.000000 I7.091744 J10.924200
G03 X681.940098 Y307.873676 Z-1.000000 I0.921990 J2.483325
G03 X683.075475 Y308.282377 Z-1.000000 I-0.237092 J2.440040
G03 X684.367794 Y309.258393 Z-1.000000 I-8.613475 J12.748474
G03 X685.897935 Y311.138223 Z-1.000000 I-4.173545 J4.959842
G03 X686.388319 Y312.945544 Z-1.000000 I-3.611056 J1.949984
G03 X686.039172 Y314.800065 Z-1.000000 I-4.304500 J0.149726
G03 X684.635294 Y316.949109 Z-1.000000 I-6.898032 J-2.973117
G03 X683.849223 Y317.718909 Z-1.000000 I-8.847052 J-8.247816
G03 X678.702332 Y322.262570 Z-1.000000 I-368.173074 J-411.865925
G03 X673.681987 Y326.610412 Z-1.000000 I-665.319356 J-763.155300
G03 X666.829594 Y332.472152 Z-1.000000 I-1043.630413 J-1213.071067
G02 X659.995081 Y338.329350 Z-1.000000 I783.891302 J921.605218
G02 X655.457232 Y342.285610 Z-1.000000 I412.446938 J477.659208
G02 X650.805035 Y346.469251 Z-1.000000 I194.753492 J221.243811
G02 X650.727094 Y346.646688 Z-1.000000 I0.163006 J0.177438
G02 X650.805065 Y346.824233 Z-1.000000 I0.241126 J-0.000000
G02 X655.837619 Y351.358683 Z-1.000000 I233.699381 J-254.311074
G02 X660.738803 Y355.650202 Z-1.000000 I490.240573 J-554.940915
G02 X668.124594 Y362.015839 Z-1.000000 I932.350720 J-1074.301351
G03 X675.503992 Y368.396851 Z-1.000000 I-627.492267 J733.128439
G03 X680.531519 Y372.844926 Z-1.000000 I-350.660817 J401.406868
G03 X685.657048 Y377.562180 Z-1.000000 I-174.459441 J194.701936
G03 X685.930294 Y377.970299 Z-1.000000 I-0.760931 J0.804996
G03 X686.262822 Y380.686137 Z-1.000000 I-5.274430 J2.024079
G03 X685.447507 Y382.877968 Z-1.000000 I-5.018242 J-0.619126
G03 X683.733752 Y384.455488 Z-1.000000 I-4.118390 J-2.754418
G03 X681.144594 Y385.125919 Z-1.000000 I-2.589157 J-4.664354
G03 X679.965551 Y384.951141 Z-1.000000 I0.000000 J-4.064265
G03 X678.310257 Y384.203959 Z-1.000000 I2.037118 J-6.720144
G03 X676.713513 Y383.086108 Z-1.000000 I8.980335 J-14.526874
G03 X673.219994 Y380.227039 Z-1.000000 I77.754373 J-98.572599
G03 X667.449474 Y375.327517 Z-1.000000 I21457.757949 J-25277.035145
G03 X657.539694 Y366.910499 Z-1.000000 I67275.776556 J-79223.851637
G01 X647.630494 Y358.492849 Z-1.000000
G01 X647.320794 Y370.626849 Z-1.000000
G03 X647.060096 Y378.893005 Z-1.000000 I-686.952821 J-17.527996
G03 X646.856219 Y381.288379 Z-1.000000 I-25.370669 J-0.952996
G03 X646.614031 Y382.477732 Z-1.000000 I-8.774484 J-1.167420
G03 X646.384738 Y383.039703 Z-1.000000 I-2.486878 J-0.686920
G03 X646.024602 Y383.516676 Z-1.000000 I-1.855105 J-1.026240
G03 X645.549094 Y383.943389 Z-1.000000 I-4.394680 J-4.418915
G03 X643.793163 Y384.870134 Z-1.000000 I-3.230943 J-3.994886
G03 X641.850419 Y385.108883 Z-1.000000 I-1.671409 J-5.576943
G03 X639.941175 Y384.687884 Z-1.000000 I0.268647 J-5.758070
G03 X638.312394 Y383.605749 Z-1.000000 I1.868865 J-4.579780
G01 X638.312394 Y383.605749 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)

G00 Z5.000000
G00 X712.631794 Y383.605749

G01 Z-1.000000 F100.0(Penetrate)
G03 X711.509385 Y381.942670 Z-1.000000 I3.722662 J-3.722715 F400.000000
G03 X711.228557 Y380.128713 Z-1.000000 I3.436077 J-1.460673
G03 X711.710975 Y378.214223 Z-1.000000 I6.486941 J0.616570
G03 X713.482094 Y374.659309 Z-1.000000 I28.929552 J12.194545
G03 X728.231905 Y349.321209 Z-1.000000 I7683.765663 J4455.910039
G03 X739.432707 Y330.231123 Z-1.000000 I5340.883009 J3120.846621
G03 X744.907126 Y320.968130 Z-1.000000 I3364.306336 J1982.052278
G03 X748.209874 Y315.427885 Z-1.000000 I1145.349102 J679.031064
G03 X751.638734 Y309.785154 Z-1.000000 I469.145026 J281.217326
G03 X751.912494 Y309.428095 Z-1.000000 I1.880372 J1.158224
G03 X753.046380 Y308.586489 Z-1.000000 I2.931146 J2.764456
G03 X754.818494 Y307.979734 Z-1.000000 I3.162501 J6.345309
G03 X756.820048 Y308.108710 Z-1.000000 I0.760054 J3.800220
G03 X758.739644 Y309.187214 Z-1.000000 I-1.909196 J5.645691
G03 X760.144002 Y310.889718 Z-1.000000 I-3.887389 J4.637084
G03 X760.657094 Y312.874395 Z-1.000000 I-3.576186 J1.983202
G03 X760.579936 Y313.462543 Z-1.000000 I-2.285674 J-0.000719
G03 X759.951746 Y315.196942 Z-1.000000 I-10.265346 J-2.737092
G03 X759.015117 Y317.073165 Z-1.000000 I-33.268893 J-15.436253
G03 X756.019007 Y322.522009 Z-1.000000 I-246.287111 J-131.875698
G03 X749.888928 Y333.217163 Z-1.000000 I-814.861027 J-459.944745
G03 X727.565794 Y371.500689 Z-1.000000 I-22273.719778 J-12962.161552
G03 X724.671183 Y376.364572 Z-1.000000 I-303.060942 J-177.065445
G03 X722.475969 Y379.910983 Z-1.000000 I-190.113817 J-115.227044
G03 X720.113768 Y383.505519 Z-1.000000 I-102.457393 J-64.757923
G03 X719.638994 Y384.042099 Z-1.000000 I-2.368705 J-1.617528
G03 X718.064956 Y384.901654 Z-1.000000 I-2.696444 J-3.066811
G03 X716.138132 Y385.110804 Z-1.000000 I-1.571476 J-5.497289
G03 X714.238696 Y384.675012 Z-1.000000 I0.372416 J-5.980522
G03 X712.631794 Y383.605749 Z-1.000000 I1.857593 J-4.533681
G01 X712.631794 Y383.605749 Z-1.000000
G00 Z5.000000

(Change tool to Cylindrical cutter)         WORK/


(Footer)
M5
G00 X0.0000 Y0.0000
M2
(Using default footer. To add your own footer create file "footer" in the output dir.)
(end)
%
"""

BORBOLETA = """
%
(Header)
(Generated by gcodetools from Inkscape.)
(Using default header. To add your own header create file "header" in the output dir.)
M3
(Header end.)
G21 (All units in mm)

(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X158.910410 Y456.633840
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X121.002175 Y431.983438 Z-1.000000 I-581.344960 J852.538737 F400.000000
G02 X119.295180 Y432.091370 Z-1.000000 I-0.777552 J1.255086
G03 X118.020209 Y432.742514 Z-1.000000 I-2.174292 J-2.683568
G03 X117.719285 Y432.510584 Z-1.000000 I-0.062169 J-0.230524
G03 X117.883821 Y431.679790 Z-1.000000 I2.246561 J0.013235
G03 X119.315980 Y428.561970 Z-1.000000 I34.987407 J14.183492
G02 X120.779617 Y425.217547 Z-1.000000 I-25.555136 J-13.176309
G02 X138.030090 Y377.134214 Z-1.000000 I-5290.807439 J-1925.276728
G03 X155.126623 Y328.990173 Z-1.000000 I18536.907582 J6555.573122
G03 X155.340330 Y328.588497 Z-1.000000 I1.468624 J0.523676
G03 X155.703398 Y328.381078 Z-1.000000 I0.416046 J0.306784
G03 X168.094380 Y327.274450 Z-1.000000 I46.895010 J455.161571
G02 X180.473989 Y326.106136 Z-1.000000 I-25.184543 J-333.030849
G02 X181.118760 Y325.800183 Z-1.000000 I-0.123734 J-1.093137
G02 X181.048028 Y325.382456 Z-1.000000 I-0.180003 J-0.184373
G02 X158.028460 Y314.987867 Z-1.000000 I-388.329611 J829.297460
G01 X134.394820 Y304.705987 Z-1.000000
G01 X129.719310 Y289.131438 Z-1.000000
G02 X125.037706 Y273.558722 Z-1.000000 I-19952.311174 J5989.769559
G02 X119.327380 Y254.613017 Z-1.000000 I-29474.083944 J8873.253150
G01 X113.610960 Y235.669157 Z-1.000000
G01 X90.377145 Y208.535883 Z-1.000000
G03 X67.916065 Y181.265669 Z-1.000000 I701.196766 J-600.425027
G03 X67.701931 Y179.906450 Z-1.000000 I1.097709 J-0.869412
G02 X68.083353 Y178.367420 Z-1.000000 I-6.507089 J-2.429448
G02 X72.001790 Y146.641550 Z-1.000000 I-2799.860700 J-361.913515
G03 X75.981549 Y114.957428 Z-1.000000 I2052.169451 J241.675601
G03 X76.374997 Y113.874600 Z-1.000000 I2.666057 J0.355823
G03 X77.162706 Y113.317753 Z-1.000000 I1.069143 J0.676831
G03 X78.509530 Y113.387640 Z-1.000000 I0.550098 J2.411378
G02 X81.321010 Y114.058080 Z-1.000000 I4.965210 J-14.591356
G02 X176.826300 Y127.262010 Z-1.000000 I2396.550314 J-16982.476844
G03 X213.126972 Y132.281449 Z-1.000000 I-868.119735 J6412.030987
G03 X213.326760 Y132.414450 Z-1.000000 I-0.038796 J0.274834
G02 X213.872117 Y133.161022 Z-1.000000 I4.208762 J-2.501953
G02 X251.549160 Y175.582013 Z-1.000000 I26727.717506 J-23700.808802
G02 X288.928289 Y217.534074 Z-1.000000 I26784.426658 J-23827.194990
G02 X291.207990 Y219.961644 Z-1.000000 I48.928228 J-43.663700
G03 X294.102890 Y223.047339 Z-1.000000 I-58.162681 J57.467166
G03 X322.184600 Y254.741091 Z-1.000000 I-6115.155382 J5446.516901
G02 X349.699162 Y284.686116 Z-1.000000 I738.171758 J-650.645461
G02 X350.610210 Y284.422474 Z-1.000000 I0.386544 J-0.370190
G02 X350.382574 Y282.478962 Z-1.000000 I-3.075899 J-0.624820
G02 X347.282940 Y277.235846 Z-1.000000 I-27.595698 J12.776292
G01 X343.531690 Y272.136325 Z-1.000000
G01 X349.634720 Y266.267754 Z-1.000000
G01 X355.737760 Y260.399172 Z-1.000000
G01 X361.840790 Y266.267754 Z-1.000000
G01 X367.943820 Y272.136325 Z-1.000000
G01 X364.192580 Y277.235846 Z-1.000000
G02 X361.099065 Y282.481621 Z-1.000000 I24.230453 J17.824110
G02 X360.865310 Y284.463653 Z-1.000000 I2.934548 J1.350892
G02 X361.788871 Y284.732735 Z-1.000000 I0.531846 J-0.105944
G02 X389.282130 Y254.782282 Z-1.000000 I-699.683843 J-669.874476
G03 X417.363049 Y223.055135 Z-1.000000 I5585.887281 J4915.641239
G03 X420.278270 Y219.961644 Z-1.000000 I55.938099 J49.794175
G02 X422.497513 Y217.609539 Z-1.000000 I-41.209510 J-41.104730
G02 X459.873537 Y175.653249 Z-1.000000 I-25184.525099 J-22472.818039
G02 X497.512513 Y133.259036 Z-1.000000 I-25140.021147 J-22358.042575
G02 X498.124070 Y132.429600 Z-1.000000 I-4.296159 J-3.807804
G03 X498.312766 Y132.305432 Z-1.000000 I0.225066 J0.136566
G03 X560.540800 Y123.723050 Z-1.000000 I2813.172814 J20167.503925
G02 X622.807985 Y115.107002 Z-1.000000 I-2135.942841 J-15665.519755
G02 X628.710610 Y114.152090 Z-1.000000 I-19.889139 J-141.661772
G03 X634.381885 Y113.557307 Z-1.000000 I6.700621 J36.555320
G03 X635.163000 Y113.973300 Z-1.000000 I0.024793 J0.894796
G03 X635.516830 Y114.949304 Z-1.000000 I-2.040792 J1.291985
G03 X639.473710 Y146.641550 Z-1.000000 I-2266.377128 J299.057725
G02 X643.392161 Y178.367421 Z-1.000000 I2803.905563 J-330.204459
G02 X643.773580 Y179.906450 Z-1.000000 I6.888604 J-0.890431
G03 X643.559450 Y181.265670 Z-1.000000 I-1.311834 J0.489812
G03 X621.098370 Y208.535883 Z-1.000000 I-723.655347 J-573.152769
G01 X597.864550 Y235.669157 Z-1.000000
G01 X592.148130 Y254.613017 Z-1.000000
G02 X586.437809 Y273.558725 Z-1.000000 I29519.850233 J8907.771230
G02 X581.755650 Y289.133305 Z-1.000000 I19923.531982 J5998.055386
G01 X577.079580 Y304.709720 Z-1.000000
G01 X553.467500 Y315.010220 Z-1.000000
G02 X530.445822 Y325.418772 Z-1.000000 I380.814058 J872.950100
G02 X530.377760 Y325.820670 Z-1.000000 I0.105116 J0.224514
G02 X530.997821 Y326.114371 Z-1.000000 I0.737336 J-0.755276
G02 X543.381170 Y327.274450 Z-1.000000 I38.539209 J-344.716095
G03 X555.772146 Y328.381079 Z-1.000000 I-34.503793 J456.265078
G03 X556.135200 Y328.588497 Z-1.000000 I-0.052975 J0.514168
G03 X556.348907 Y328.990173 Z-1.000000 I-1.254920 J0.925354
G03 X573.445440 Y377.134214 Z-1.000000 I-18510.294906 J6600.337794
G02 X590.695913 Y425.217547 Z-1.000000 I5308.057912 J-1877.193395
G02 X592.159550 Y428.561970 Z-1.000000 I27.018774 J-9.831886
G03 X593.593506 Y431.684212 Z-1.000000 I-33.560764 J17.303904
G03 X593.755891 Y432.511729 Z-1.000000 I-2.055745 J0.833094
G03 X593.454312 Y432.742152 Z-1.000000 I-0.238360 J0.000599
G03 X592.168550 Y432.081300 Z-1.000000 I0.935810 J-3.401946
G02 X590.449338 Y431.973131 Z-1.000000 I-0.935667 J1.154815
G02 X552.608950 Y456.623760 Z-1.000000 I538.550066 J868.080014
G03 X514.602355 Y482.145261 Z-1.000000 I-1586.253834 J-2321.187258
G03 X513.360140 Y482.518120 Z-1.000000 I-1.242215 J-1.882849
G03 X512.017640 Y482.378439 Z-1.000000 I-0.000000 J-6.521336
G03 X468.104790 Y472.997820 Z-1.000000 I1485.924354 J-7063.429649
G02 X424.182457 Y463.608436 Z-1.000000 I-1605.864194 J7404.599967
G02 X419.174210 Y462.689360 Z-1.000000 I-20.076277 J95.295033
G03 X414.949149 Y460.797293 Z-1.000000 I1.363814 J-8.708850
G03 X411.321930 Y456.251530 Z-1.000000 I8.819990 J-10.757793
G03 X404.820274 Y442.609448 Z-1.000000 I372.356487 J-185.831085
G03 X376.458330 Y379.542989 Z-1.000000 I7931.948999 J-3605.028460
G02 X363.947756 Y351.966664 Z-1.000000 I-1906.953769 J848.503314
G02 X363.200860 Y350.874869 Z-1.000000 I-3.615463 J1.671960
G02 X362.661851 Y351.040131 Z-1.000000 I-0.227605 J0.219289
G02 X360.797850 Y364.725326 Z-1.000000 I186.897768 J32.426068
G03 X359.323371 Y379.542655 Z-1.000000 I-32578.983059 J-3234.492450
G03 X357.968540 Y393.092349 Z-1.000000 I-27400.142027 J-2732.872241
G03 X356.156343 Y406.139933 Z-1.000000 I-172.039458 J-17.245152
G03 X355.557390 Y406.641710 Z-1.000000 I-0.598953 J-0.106587
G03 X354.903036 Y406.239972 Z-1.000000 I-0.000000 J-0.733778
G03 X354.487890 Y404.534030 Z-1.000000 I3.421264 J-1.736058
G02 X354.294981 Y401.384024 Z-1.000000 I-29.614171 J0.232683
G02 X351.405832 Y377.051023 Z-1.000000 I-3281.844450 J377.327740
G02 X348.268125 Y352.230552 Z-1.000000 I-3216.836127 J394.051356
G02 X348.020100 Y351.734332 Z-1.000000 I-0.839254 J0.109388
G02 X347.716951 Y351.788235 Z-1.000000 I-0.132275 J0.135491
G02 X339.745840 Y369.063637 Z-1.000000 I652.418456 J311.511640
G03 X306.692269 Y442.553763 Z-1.000000 I-10469.269891 J-4664.573408
G03 X300.190940 Y456.188630 Z-1.000000 I-366.759550 J-166.509585
G01 X297.397300 Y461.775430 Z-1.000000
G01 X283.697620 Y464.427720 Z-1.000000
G02 X270.024017 Y467.206242 Z-1.000000 I143.587296 J741.653117
G02 X234.968900 Y474.799080 Z-1.000000 I1039.064157 J4881.938956
G03 X199.922639 Y482.359662 Z-1.000000 I-872.962525 J-3961.516794
G03 X198.404910 Y482.518120 Z-1.000000 I-1.517729 J-7.189245
G03 X196.999345 Y482.098560 Z-1.000000 I0.000000 J-2.564166
G03 X158.910780 Y456.633840 Z-1.000000 I1371.297942 J-2092.320931
G01 X158.910410 Y456.633840 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X204.921370 Y463.881460
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X211.048234 Y447.327745 Z-1.000000 I-151.933101 J-65.644198 F400.000000
G02 X210.354970 Y446.386500 Z-1.000000 I-0.693264 J-0.215315
G03 X209.319676 Y446.298811 Z-1.000000 I-0.000000 J-6.155397
G03 X192.775150 Y443.386980 Z-1.000000 I270.670451 J-1586.360820
G02 X176.227577 Y440.503855 Z-1.000000 I-216.170548 J1191.772931
G02 X174.755790 Y440.376010 Z-1.000000 I-1.544166 J9.241157
G02 X174.384265 Y440.992190 Z-1.000000 I-0.003077 J0.418240
G02 X185.101850 Y459.634760 Z-1.000000 I329.212191 J-176.861535
G02 X197.106599 Y478.582253 Z-1.000000 I672.539511 J-412.830725
G02 X197.687610 Y478.888430 Z-1.000000 I0.566229 J-0.370132
G02 X198.223618 Y478.542654 Z-1.000000 I-0.014830 J-0.611326
G02 X204.921370 Y463.881460 Z-1.000000 I-350.940901 J-169.182694
G01 X204.921370 Y463.881460 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X225.419550 Y466.361500
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X246.517932 Y454.094730 Z-1.000000 I-170.666280 J-317.817171 F400.000000
G02 X246.391980 Y453.199120 Z-1.000000 I-0.263358 J-0.419625
G02 X240.413623 Y451.803220 Z-1.000000 I-16.491492 J57.129744
G02 X227.892589 Y449.756575 Z-1.000000 I-71.378232 J397.355959
G02 X214.755011 Y448.081121 Z-1.000000 I-49.604550 J336.614505
G02 X213.901770 Y448.411270 Z-1.000000 I-0.106874 J0.991431
G02 X213.090392 Y449.666464 Z-1.000000 I3.714207 J3.290765
G02 X201.946570 Y474.990700 Z-1.000000 I1550.787374 J697.531392
G02 X200.873039 Y478.492051 Z-1.000000 I15.475558 J6.660129
G02 X201.203800 Y478.882220 Z-1.000000 I0.322820 J0.061618
G02 X202.091516 Y478.654924 Z-1.000000 I-0.049213 J-2.039363
G02 X225.419550 Y466.361500 Z-1.000000 I-775.955049 J-1500.734737
G01 X225.419550 Y466.361500 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X246.192530 Y470.475460
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X284.827319 Y462.283484 Z-1.000000 I700.327835 J3207.668007 F400.000000
G03 X287.166160 Y462.044750 Z-1.000000 I2.344835 J11.396009
G02 X288.829587 Y461.758254 Z-1.000000 I-0.002531 J-4.986961
G02 X288.901840 Y461.429590 Z-1.000000 I-0.066217 J-0.186831
G02 X287.957277 Y460.926147 Z-1.000000 I-1.335670 J1.368172
G02 X272.283861 Y457.822210 Z-1.000000 I-155.712836 J745.152697
G02 X256.910972 Y455.101482 Z-1.000000 I-145.851901 J779.313804
G02 X253.605640 Y454.823770 Z-1.000000 I-3.352639 J20.094260
G02 X250.935575 Y455.423899 Z-1.000000 I-0.011960 J6.186619
G02 X228.315370 Y466.858220 Z-1.000000 I227.630271 J478.406359
G02 X206.661817 Y478.695984 Z-1.000000 I698.484552 J1303.386065
G02 X206.715030 Y478.902090 Z-1.000000 I0.053577 J0.096089
G02 X207.573633 Y478.812395 Z-1.000000 I0.008527 J-4.072682
G02 X246.192530 Y470.475460 Z-1.000000 I-1760.166257 J-8247.172877
G01 X246.192530 Y470.475460 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X504.909230 Y478.488390
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X501.473816 Y476.221202 Z-1.000000 I-13.283581 J16.391892 F400.000000
G02 X480.884221 Y465.610009 Z-1.000000 I-553.321461 J1048.364644
G02 X459.757169 Y455.249895 Z-1.000000 I-519.408746 J1032.491853
G02 X457.871450 Y454.825410 Z-1.000000 I-1.873088 J3.920179
G02 X454.848550 Y455.077435 Z-1.000000 I0.058556 J18.957384
G02 X440.111955 Y457.644384 Z-1.000000 I132.313947 J803.185149
G02 X425.086787 Y460.556043 Z-1.000000 I141.549802 J770.669365
G02 X424.245490 Y461.001850 Z-1.000000 I0.337074 J1.652824
G02 X424.328792 Y461.299371 Z-1.000000 I0.122995 J0.125985
G02 X462.377552 Y469.958813 Z-1.000000 I1215.236996 J-5251.714588
G02 X500.079198 Y478.258704 Z-1.000000 I1215.946906 J-5433.570715
G02 X503.675560 Y478.777040 Z-1.000000 I5.490222 J-25.357263
G02 X504.864358 Y478.683728 Z-1.000000 I0.288679 J-3.941531
G02 X504.909230 Y478.488390 Z-1.000000 I-0.025416 J-0.108661
G01 X504.909230 Y478.488390 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X509.528920 Y474.990700
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X498.378071 Y449.669563 Z-1.000000 I-1510.340153 J650.003354 F400.000000
G02 X497.542050 Y448.380340 Z-1.000000 I-4.624419 J2.083115
G02 X495.459431 Y447.624431 Z-1.000000 I-1.741043 J1.549893
G02 X465.613960 Y453.110110 Z-1.000000 I64.077217 J432.551215
G02 X465.423059 Y454.162191 Z-1.000000 I0.124466 J0.565945
G02 X485.969540 Y466.315210 Z-1.000000 I157.185266 J-242.300066
G02 X509.383979 Y478.654901 Z-1.000000 I804.981971 J-1499.060972
G02 X510.271690 Y478.882220 Z-1.000000 I0.936919 J-1.811820
G02 X510.602450 Y478.492037 Z-1.000000 I0.007929 J-0.328564
G02 X509.528920 Y474.990700 Z-1.000000 I-16.548579 J3.158645
G01 X509.528920 Y474.990700 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X526.373640 Y459.634760
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X537.091234 Y440.991950 Z-1.000000 I-318.440459 J-195.470867 F400.000000
G02 X536.719700 Y440.376010 Z-1.000000 I-0.368310 J-0.197861
G02 X535.247915 Y440.503604 Z-1.000000 I0.068759 J9.345410
G02 X518.700350 Y443.386980 Z-1.000000 I200.080395 J1197.176755
G03 X502.155813 Y446.298811 Z-1.000000 I-287.219687 J-1583.476694
G03 X501.120520 Y446.386500 Z-1.000000 I-1.035293 J-6.067706
G02 X500.410138 Y446.719302 Z-1.000000 I0.000000 J0.924574
G02 X500.077340 Y447.638760 Z-1.000000 I1.103746 J0.919458
G02 X500.428777 Y449.420039 Z-1.000000 I4.689980 J0.000000
G02 X506.611419 Y463.981653 Z-1.000000 I613.185522 J-251.756011
G02 X513.199281 Y478.505158 Z-1.000000 I589.302324 J-258.551648
G02 X513.787880 Y478.888430 Z-1.000000 I0.603130 J-0.282642
G02 X514.368908 Y478.583119 Z-1.000000 I0.016360 J-0.674389
G02 X526.373640 Y459.634760 Z-1.000000 I-661.565177 J-432.411583
G01 X526.373640 Y459.634760 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X188.434580 Y468.366570
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X171.124376 Y441.008786 Z-1.000000 I-698.966193 J423.104494 F400.000000
G02 X169.480940 Y439.685200 Z-1.000000 I-2.896052 J1.913813
G02 X166.150468 Y438.717344 Z-1.000000 I-7.254516 J18.749241
G02 X146.595831 Y434.992580 Z-1.000000 I-236.188169 J1186.771905
G02 X126.643376 Y431.545419 Z-1.000000 I-203.206271 J1116.707952
G02 X126.330750 Y431.833200 Z-1.000000 I-0.043576 J0.266360
G02 X126.696062 Y432.469937 Z-1.000000 I0.856350 J-0.068147
G02 X159.275713 Y454.970195 Z-1.000000 I2793.430851 J-4009.960803
G02 X191.964446 Y477.158051 Z-1.000000 I2772.803892 J-4049.918997
G02 X193.180480 Y477.602320 Z-1.000000 I1.507831 J-2.240799
G02 X193.408941 Y477.269978 Z-1.000000 I0.024766 J-0.227671
G02 X188.434580 Y468.366570 Z-1.000000 I-130.252961 J66.931422
G01 X188.434580 Y468.366570 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X552.084820 Y454.817210
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X583.896873 Y432.447881 Z-1.000000 I-830.179931 J-1214.427349 F400.000000
G02 X583.730490 Y431.933840 Z-1.000000 I-0.166383 J-0.230093
G02 X581.068522 Y432.154213 Z-1.000000 I-0.000000 J16.187636
G02 X564.247509 Y435.089878 Z-1.000000 I185.012772 J1109.760556
G02 X547.662301 Y438.240139 Z-1.000000 I206.417166 J1131.959437
G02 X543.862810 Y439.203860 Z-1.000000 I7.121683 J36.049037
G02 X540.481983 Y440.788541 Z-1.000000 I4.076384 J13.095477
G02 X538.822110 Y442.469450 Z-1.000000 I2.985115 J4.607762
G02 X517.692194 Y477.110526 Z-1.000000 I1180.493111 J743.826743
G02 X518.029100 Y477.700570 Z-1.000000 I0.336906 J0.198838
G02 X518.559925 Y477.538793 Z-1.000000 I0.000000 J-0.951766
G02 X552.084820 Y454.817210 Z-1.000000 I-2827.478388 J-4207.937809
G01 X552.084820 Y454.817210 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X304.986220 Y423.202030
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X313.946638 Y386.777290 Z-1.000000 I20709.439622 J5075.168902 F400.000000
G03 X316.733550 Y375.628730 Z-1.000000 I1942.900640 J479.762766
G02 X319.112674 Y364.867789 Z-1.000000 I-176.688872 J-44.707411
G02 X318.831600 Y364.561433 Z-1.000000 I-0.254837 J-0.048310
G02 X317.895759 Y365.092545 Z-1.000000 I0.134952 J1.327842
G02 X286.568605 Y407.566819 Z-1.000000 I5044.997158 J3753.757662
G02 X255.625170 Y450.265285 Z-1.000000 I5087.790814 J3719.666509
G02 X255.193490 Y451.606150 Z-1.000000 I1.866629 J1.340865
G02 X255.756953 Y452.331009 Z-1.000000 I0.747974 J0.000000
G02 X271.138935 Y455.895690 Z-1.000000 I90.466653 J-355.403533
G02 X286.107717 Y458.755358 Z-1.000000 I89.520115 J-427.981841
G02 X292.897060 Y459.492530 Z-1.000000 I9.220269 J-53.285094
G02 X294.913520 Y458.106666 Z-1.000000 I0.091216 J-2.027205
G02 X304.986220 Y423.202030 Z-1.000000 I-424.423449 J-141.384843
G01 X304.986220 Y423.202030 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X436.936160 Y456.463280
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X454.892499 Y452.967653 Z-1.000000 I-83.368163 J-476.112498 F400.000000
G02 X456.113940 Y452.351760 Z-1.000000 I-0.566589 J-2.642790
G02 X456.224018 Y451.415354 Z-1.000000 I-0.467179 J-0.529591
G02 X425.379857 Y408.205542 Z-1.000000 I-3589.292006 J2529.505855
G02 X393.989680 Y365.330827 Z-1.000000 I-3599.071683 J2602.086221
G02 X392.643890 Y364.563251 Z-1.000000 I-1.545137 J1.145514
G02 X392.362868 Y364.868794 Z-1.000000 I-0.026814 J0.257344
G02 X394.741940 Y375.628730 Z-1.000000 I179.139953 J-33.965696
G03 X397.528852 Y386.777290 Z-1.000000 I-1940.113733 J490.911307
G03 X406.489270 Y423.202030 Z-1.000000 I-20700.479045 J5111.594285
G02 X416.048533 Y458.830651 Z-1.000000 I821.834140 J-201.403812
G02 X417.120830 Y459.634760 Z-1.000000 I1.072297 J-0.312910
G02 X418.812323 Y459.500252 Z-1.000000 I-0.000000 J-10.702884
G02 X436.936160 Y456.463280 Z-1.000000 I-198.403786 J-1239.618233
G01 X436.936160 Y456.463280 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X281.717440 Y411.429720
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X307.732363 Y376.162123 Z-1.000000 I-37599.354742 J-27762.158050 F400.000000
G02 X312.609095 Y369.323174 Z-1.000000 I-216.386376 J-159.459466
G02 X314.459168 Y366.279304 Z-1.000000 I-25.427156 J-17.538874
G02 X314.348124 Y366.132267 Z-1.000000 I-0.089732 J-0.047683
G02 X313.784686 Y366.401867 Z-1.000000 I0.281332 J1.311522
G03 X311.604060 Y368.101308 Z-1.000000 I-45.189454 J-55.735777
G02 X303.265417 Y374.574189 Z-1.000000 I182.407469 J243.592686
G02 X260.667345 Y408.971772 Z-1.000000 I9243.247264 J11490.441219
G02 X216.698436 Y444.756610 Z-1.000000 I8966.266694 J11061.776149
G02 X216.421950 Y445.252410 Z-1.000000 I0.487930 J0.597089
G02 X216.945812 Y445.987418 Z-1.000000 I0.655731 J0.086831
G02 X250.875640 Y452.261810 Z-1.000000 I192.700473 J-947.182354
G02 X251.370870 Y452.064334 Z-1.000000 I0.084548 J-0.507677
G02 X281.717440 Y411.429720 Z-1.000000 I-3716.273827 J-2807.020800
G01 X281.717440 Y411.429720 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X479.413340 Y448.681080
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X494.762360 Y445.689934 Z-1.000000 I-60.633477 J-352.016943 F400.000000
G02 X495.055410 Y445.275830 Z-1.000000 I-0.079968 J-0.367335
G02 X494.773026 Y444.764947 Z-1.000000 I-0.780006 J0.097656
G02 X450.851431 Y409.000153 Z-1.000000 I-8668.181769 J10600.263545
G02 X408.303042 Y374.635322 Z-1.000000 I-8938.506818 J11023.578821
G02 X399.878560 Y368.101308 Z-1.000000 I-187.872908 J233.531671
G03 X397.694622 Y366.402253 Z-1.000000 I42.989136 J-57.510653
G02 X397.129350 Y366.132215 Z-1.000000 I-0.847121 J1.046623
G02 X397.017884 Y366.279805 Z-1.000000 I-0.021355 J0.099759
G02 X398.866079 Y369.322102 Z-1.000000 I27.153228 J-14.413038
G02 X403.742132 Y376.158972 Z-1.000000 I220.368859 J-152.009770
G02 X429.758050 Y411.417650 Z-1.000000 I37472.577049 J-27622.234893
G02 X460.331627 Y451.360097 Z-1.000000 I1179.411619 J-871.097363
G02 X461.833520 Y451.865380 Z-1.000000 I1.140119 J-0.904125
G03 X463.698435 Y451.457696 Z-1.000000 I6.587441 J25.664392
G03 X479.413340 Y448.681080 Z-1.000000 I330.061091 J1822.198199
G01 X479.413340 Y448.681080 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X314.676350 Y420.492160
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X327.121910 Y392.791252 Z-1.000000 F400.000000
G01 X324.740210 Y378.404747 Z-1.000000
G02 X322.224322 Y364.088235 Z-1.000000 I-724.682801 J119.971562
G02 X321.999930 Y363.668186 Z-1.000000 I-0.775182 J0.144146
G02 X321.760563 Y363.735288 Z-1.000000 I-0.100290 J0.102731
G02 X311.372100 Y405.153410 Z-1.000000 I4103.514662 J1051.252449
G03 X301.023239 Y446.965267 Z-1.000000 I-10865.577663 J-2667.154523
G03 X300.481100 Y448.795270 Z-1.000000 I-21.314370 J-5.319087
G02 X300.608662 Y449.392027 Z-1.000000 I0.559259 J0.192465
G02 X301.045070 Y449.397470 Z-1.000000 I0.220817 J-0.206799
G02 X301.996901 Y448.033067 Z-1.000000 I-3.535133 J-3.480379
G02 X314.676350 Y420.492160 Z-1.000000 I-1490.518233 J-702.902945
G01 X314.676350 Y420.492160 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X400.428800 Y406.362580
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X389.706940 Y363.739147 Z-1.000000 I-4414.799668 J1087.875946 F400.000000
G02 X389.470650 Y363.672968 Z-1.000000 I-0.137282 J0.035239
G02 X389.249376 Y364.087309 Z-1.000000 I0.543103 J0.556292
G02 X386.734820 Y378.404747 Z-1.000000 I732.087953 J135.955351
G01 X384.352670 Y392.791252 Z-1.000000
G01 X396.090160 Y418.685580 Z-1.000000
G03 X407.761920 Y444.607717 Z-1.000000 I-5157.224842 J2337.692452
G03 X408.958000 Y447.456830 Z-1.000000 I-61.068081 J27.312479
G02 X410.139843 Y450.008007 Z-1.000000 I22.022261 J-8.652566
G02 X410.336183 Y449.983232 Z-1.000000 I0.093397 J-0.050210
G02 X410.363992 Y449.364231 Z-1.000000 I-1.534116 J-0.379047
G02 X409.509150 Y444.627278 Z-1.000000 I-91.132808 J14.000438
G02 X407.409354 Y435.195428 Z-1.000000 I-328.313368 J68.142166
G02 X400.428800 Y406.362580 Z-1.000000 I-3546.780113 J843.428967
G01 X400.428800 Y406.362580 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X220.949380 Y431.020780
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X228.662901 Y416.886018 Z-1.000000 I7798.505953 J4246.572189 F400.000000
G03 X236.855520 Y401.936050 Z-1.000000 I8738.071942 J4778.765604
G02 X243.157498 Y389.114005 Z-1.000000 I-138.140310 J-75.855064
G02 X242.636920 Y388.687783 Z-1.000000 I-0.320324 J-0.139786
G02 X175.120720 Y436.812073 Z-1.000000 I2606.117291 J3727.687749
G02 X175.049090 Y437.798260 Z-1.000000 I0.380249 J0.523313
G02 X175.714199 Y438.154621 Z-1.000000 I0.948683 J-0.971757
G02 X192.145184 Y441.544788 Z-1.000000 I254.415791 J-1191.554034
G02 X208.189266 Y444.632470 Z-1.000000 I246.927952 J-1239.848584
G02 X211.398150 Y445.037530 Z-1.000000 I5.086592 J-27.383071
G02 X212.804337 Y444.353798 Z-1.000000 I0.106127 J-1.569608
G02 X220.949380 Y431.020780 Z-1.000000 I-129.815136 J-88.457501
G01 X220.949380 Y431.020780 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X515.997080 Y442.132660
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X532.713815 Y438.887691 Z-1.000000 I-128.174017 J-704.980785 F400.000000
G02 X534.959750 Y438.217320 Z-1.000000 I-2.744583 J-13.292583
G02 X535.143750 Y436.853847 Z-1.000000 I-0.289905 J-0.733275
G02 X473.097890 Y391.697633 Z-1.000000 I-1474.296924 J1960.519440
G01 X466.413340 Y386.990850 Z-1.000000
G01 X474.619370 Y401.934930 Z-1.000000
G03 X482.812590 Y416.886012 Z-1.000000 I-8730.400152 J4793.996005
G03 X490.526110 Y431.020780 Z-1.000000 I-7792.898790 J4261.854192
G02 X498.353994 Y444.953245 Z-1.000000 I529.550212 J-288.359827
G02 X498.725670 Y445.172340 Z-1.000000 I0.380309 J-0.220356
G02 X499.220658 Y445.132782 Z-1.000000 I0.038426 J-2.635852
G02 X515.997080 Y442.132660 Z-1.000000 I-511.092726 J-2906.392169
G01 X515.997080 Y442.132660 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X273.367910 Y396.550784
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X316.058385 Y361.779409 Z-1.000000 I-4580.933331 J-5667.822532 F400.000000
G02 X316.925340 Y360.798745 Z-1.000000 I-2.613296 J-3.183830
G02 X316.707703 Y360.432643 Z-1.000000 I-0.201817 J-0.127766
G02 X314.171830 Y360.981511 Z-1.000000 I0.587101 J8.845050
G03 X310.699099 Y362.258911 Z-1.000000 I-86.327263 J-229.329593
G03 X280.483760 Y373.085268 Z-1.000000 I-6235.666751 J-17355.578546
G02 X250.481482 Y384.239563 Z-1.000000 I422.044083 J1181.120016
G02 X248.799340 Y385.327362 Z-1.000000 I1.770532 J4.582410
G02 X247.646913 Y386.829023 Z-1.000000 I4.875132 J4.934384
G02 X232.659556 Y413.419175 Z-1.000000 I2988.757578 J1702.111503
G02 X217.830066 Y440.286680 Z-1.000000 I2989.873255 J1667.783578
G02 X217.566510 Y441.319160 Z-1.000000 I1.890588 J1.032480
G02 X217.773192 Y441.423274 Z-1.000000 I0.129564 J0.000000
G02 X224.043290 Y436.521760 Z-1.000000 I-91.197224 J-123.122154
G03 X230.542477 Y431.209409 Z-1.000000 I632.062214 J766.639879
G03 X273.367910 Y396.550784 Z-1.000000 I26966.939116 J33277.524127
G01 X273.367910 Y396.550784 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X479.037250 Y413.768390
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X463.941778 Y386.940566 Z-1.000000 I-1482.866729 J816.718148 F400.000000
G02 X462.691830 Y385.327362 Z-1.000000 I-6.450650 J3.707267
G02 X460.996509 Y384.237655 Z-1.000000 I-3.498217 J3.578774
G02 X430.991730 Y373.085268 Z-1.000000 I-454.837288 J1177.771235
G03 X400.774063 Y362.265114 Z-1.000000 I8135.056033 J-22766.544877
G03 X397.065750 Y360.911837 Z-1.000000 I123.435774 J-344.001974
G01 X393.364730 Y359.538547 Z-1.000000
G01 X395.832080 Y361.962304 Z-1.000000
G02 X399.452611 Y365.204089 Z-1.000000 I36.847858 J-37.510229
G02 X445.918110 Y402.938011 Z-1.000000 I9614.783725 J-11792.155474
G02 X493.231303 Y441.048154 Z-1.000000 I9448.211641 J-11681.394379
G02 X493.599820 Y441.039370 Z-1.000000 I0.179043 J-0.223201
G02 X493.656660 Y440.723954 Z-1.000000 I-0.171148 J-0.193672
G02 X479.037250 Y413.768390 Z-1.000000 I-2087.339570 J1114.631059
G01 X479.037250 Y413.768390 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X172.411830 Y426.213000
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X173.896149 Y412.510976 Z-1.000000 I-87.003149 J-16.356326 F400.000000
G02 X172.845900 Y411.742470 Z-1.000000 I-0.788979 J-0.023666
G02 X125.989940 Y428.706950 Z-1.000000 I823.721744 J2348.313112
G02 X126.001240 Y429.235330 Z-1.000000 I0.097918 J0.262217
G02 X128.763847 Y429.945825 Z-1.000000 I7.023533 J-21.583357
G02 X146.522803 Y433.235221 Z-1.000000 I297.463079 J-1556.374644
G02 X164.315891 Y436.321637 Z-1.000000 I279.757224 J-1559.960137
G02 X167.166210 Y436.617690 Z-1.000000 I3.788246 J-22.603111
G02 X169.317970 Y435.352753 Z-1.000000 I0.094292 J-2.302229
G02 X172.411830 Y426.213000 Z-1.000000 I-30.748090 J-15.501932
G01 X172.411830 Y426.213000 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X565.599670 Y433.039570
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X585.753246 Y429.210174 Z-1.000000 I-300.152380 J-1634.607301 F400.000000
G02 X586.145580 Y429.004090 Z-1.000000 I-0.149487 J-0.761085
G02 X586.096595 Y428.773118 Z-1.000000 I-0.097607 J-0.099979
G02 X562.475058 Y420.114274 Z-1.000000 I-1078.270848 J2904.999422
G02 X538.808113 Y411.653431 Z-1.000000 I-1080.412099 J2984.832597
G02 X537.674840 Y411.459260 Z-1.000000 I-1.133273 J3.210064
G02 X537.079042 Y412.113776 Z-1.000000 I0.000000 J0.598432
G02 X539.063660 Y426.213000 Z-1.000000 I154.183234 J-14.513667
G02 X541.828152 Y435.807100 Z-1.000000 I51.875869 J-9.752434
G02 X543.215070 Y436.751400 Z-1.000000 I1.386918 J-0.546351
G02 X545.369120 Y436.574017 Z-1.000000 I0.000000 J-13.167512
G02 X565.599670 Y433.039570 Z-1.000000 I-194.515681 J-1173.038788
G01 X565.599670 Y433.039570 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X208.005550 Y411.137520
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X239.615004 Y388.348656 Z-1.000000 I-2801.326686 J-3918.916359 F400.000000
G02 X239.772590 Y388.039233 Z-1.000000 I-0.224986 J-0.309423
G02 X239.589476 Y387.908742 Z-1.000000 I-0.138052 J0.000000
G02 X208.622390 Y398.796160 Z-1.000000 I960.428991 J2781.232140
G01 X177.472170 Y409.943590 Z-1.000000
G01 X176.007410 Y417.024450 Z-1.000000
G02 X173.540241 Y434.216663 Z-1.000000 I138.472711 J28.644663
G02 X174.558940 Y434.850170 Z-1.000000 I0.672356 J0.054634
G02 X176.196935 Y433.781841 Z-1.000000 I-12.945033 J-21.637594
G02 X208.005550 Y411.137520 Z-1.000000 I-5816.587841 J-8204.259360
G01 X208.005550 Y411.137520 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X536.530540 Y422.670680
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X534.170447 Y411.027782 Z-1.000000 I-211.855996 J36.883974 F400.000000
G02 X533.621720 Y410.112870 Z-1.000000 I-1.689674 J0.391390
G02 X531.612956 Y408.928610 Z-1.000000 I-4.164683 J4.768432
G02 X502.038426 Y398.450548 Z-1.000000 I-764.765348 J2111.588428
G02 X473.281659 Y388.735270 Z-1.000000 I-644.802672 J1861.168705
G02 X473.049420 Y389.127577 Z-1.000000 I-0.072996 J0.221682
G02 X476.733024 Y392.149669 Z-1.000000 I25.479094 J-27.300259
G02 X506.741285 Y413.518065 Z-1.000000 I2422.586097 J-3370.353353
G02 X537.548578 Y435.035883 Z-1.000000 I2325.571696 J-3296.733749
G02 X538.067080 Y434.992130 Z-1.000000 I0.232703 J-0.336490
G02 X538.279627 Y434.384862 Z-1.000000 I-0.505429 J-0.517734
G02 X536.530540 Y422.670680 Z-1.000000 I-243.304868 J30.341069
G01 X536.530540 Y422.670680 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X150.948230 Y417.503850
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X172.401178 Y409.678522 Z-1.000000 I-326.381636 J-928.086470 F400.000000
G02 X173.672890 Y408.914510 Z-1.000000 I-1.468045 J-3.883986
G02 X174.099763 Y407.795218 Z-1.000000 I-0.846310 J-0.963809
G02 X172.670076 Y398.854772 Z-1.000000 I-121.313706 J14.815006
G02 X170.651836 Y390.161019 Z-1.000000 I-134.760312 J26.703243
G02 X169.381630 Y387.258284 Z-1.000000 I-10.116937 J2.697782
G02 X168.272340 Y387.081909 Z-1.000000 I-0.632313 J0.400293
G02 X147.357520 Y405.474970 Z-1.000000 I293.277508 J354.574663
G02 X126.453201 Y425.256145 Z-1.000000 I1585.074636 J1696.008871
G02 X126.274790 Y425.674160 Z-1.000000 I0.400496 J0.418015
G02 X126.482932 Y425.876810 Z-1.000000 I0.202722 J0.000000
G02 X127.816890 Y425.632960 Z-1.000000 I-0.118113 J-4.416697
G02 X129.358152 Y425.093916 Z-1.000000 I-178.516743 J-512.897064
G02 X150.948230 Y417.503900 Z-1.000000 I-35225.644212 J-100234.925504
G01 X150.948230 Y417.503850 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X585.200700 Y425.545310
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X584.821528 Y424.669092 Z-1.000000 I-1.201999 J0.000000 F400.000000
G02 X564.052405 Y405.431146 Z-1.000000 I-1416.668635 J1508.593360
G02 X543.047042 Y386.505017 Z-1.000000 I-1376.167670 J1506.234530
G02 X542.485010 Y386.518212 Z-1.000000 I-0.273933 J0.308297
G02 X541.751802 Y387.737942 Z-1.000000 I1.961820 J2.009536
G02 X539.335625 Y396.720921 Z-1.000000 I199.956694 J58.599334
G02 X537.328335 Y405.759137 Z-1.000000 I205.426379 J50.364992
G02 X537.331850 Y407.536650 Z-1.000000 I4.417153 J0.880024
G02 X538.440718 Y408.883770 Z-1.000000 I1.823092 J-0.370723
G02 X561.418360 Y417.979566 Z-1.000000 I277.455394 J-667.332234
G02 X584.365510 Y426.145875 Z-1.000000 I239.167293 J-635.731256
G02 X585.200700 Y425.545310 Z-1.000000 I0.201669 J-0.600565
G01 X585.200700 Y425.545310 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X164.938130 Y386.403711
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X168.441820 Y382.797943 Z-1.000000 F400.000000
G01 X165.651650 Y370.597907 Z-1.000000
G02 X162.567053 Y359.470456 Z-1.000000 I-123.522839 J28.250025
G02 X161.969320 Y359.335330 Z-1.000000 I-0.339212 J0.110906
G02 X161.197884 Y360.362631 Z-1.000000 I4.001755 J3.808357
G02 X143.529100 Y390.984671 Z-1.000000 I3615.185085 J2106.354499
G03 X125.977299 Y421.694380 Z-1.000000 I-125306.220620 J-71597.022042
G03 X125.282530 Y422.900950 Z-1.000000 I-195.000122 J-111.481912
G02 X125.452217 Y423.038587 Z-1.000000 I0.095661 J0.055481
G02 X142.703516 Y407.328660 Z-1.000000 I-1479.360199 J-1641.833832
G02 X159.646880 Y391.598408 Z-1.000000 I-1804.114389 J-1960.237200
G02 X164.938130 Y386.403711 Z-1.000000 I-107.557637 J-114.848951
G01 X164.938130 Y386.403711 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X586.037230 Y422.599850
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X585.504154 Y421.690864 Z-1.000000 I42.388539 J-25.469640 F400.000000
G03 X567.946390 Y390.984671 Z-1.000000 I48238.596657 J-27603.145110
G02 X550.277606 Y360.362631 Z-1.000000 I-3632.853869 J2075.732459
G02 X549.506170 Y359.335330 Z-1.000000 I-4.773191 J2.781056
G02 X548.908231 Y359.470775 Z-1.000000 I-0.258569 J0.246078
G02 X545.776240 Y370.777012 Z-1.000000 I124.275366 J40.512991
G01 X542.938470 Y383.156152 Z-1.000000
G01 X564.378010 Y403.305510 Z-1.000000
G02 X585.736589 Y422.849027 Z-1.000000 I743.440177 J-791.042696
G02 X586.037230 Y422.599850 Z-1.000000 I0.131362 J-0.147463
G01 X586.037230 Y422.599850 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X146.369880 Y382.527137
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X161.856150 Y355.401583 Z-1.000000 F400.000000
G01 X159.233950 Y344.890027 Z-1.000000
G02 X156.363147 Y334.696712 Z-1.000000 I-170.947094 J42.643913
G02 X155.929490 Y334.378458 Z-1.000000 I-0.433657 J0.136326
G02 X155.590853 Y334.614015 Z-1.000000 I0.000000 J0.361190
G02 X142.122433 Y371.289145 Z-1.000000 I5534.518127 J2053.283820
G02 X129.395581 Y406.634061 Z-1.000000 I5856.352361 J2128.694971
G02 X126.598460 Y415.072420 Z-1.000000 I185.842950 J66.285415
G02 X127.073226 Y415.285426 Z-1.000000 I0.250678 J0.076871
G02 X146.369880 Y382.527137 Z-1.000000 I-1196.254281 J-726.730198
G01 X146.369880 Y382.527137 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X570.697300 Y374.940747
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X556.000374 Y334.695331 Z-1.000000 I-2728.295542 J973.519875 F400.000000
G02 X555.543730 Y334.378458 Z-1.000000 I-0.456644 J0.170596
G02 X555.111534 Y334.695671 Z-1.000000 I0.000000 J0.453036
G02 X552.246170 Y344.871419 Z-1.000000 I168.058775 J52.814555
G01 X549.628620 Y355.364368 Z-1.000000
G01 X567.038610 Y385.786245 Z-1.000000
G02 X584.503518 Y415.814974 Z-1.000000 I2155.653003 J-1233.648826
G02 X584.809720 Y415.855570 Z-1.000000 I0.169037 J-0.099902
G02 X584.907682 Y415.442151 Z-1.000000 I-0.277230 J-0.284007
G02 X570.697300 Y374.940747 Z-1.000000 I-3829.412143 J1320.849551
G01 X570.697300 Y374.940747 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X210.396320 Y396.272330
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X242.827533 Y384.622099 Z-1.000000 I-2435.700748 J-6831.322289 F400.000000
G02 X246.557800 Y383.185022 Z-1.000000 I-33.221838 J-91.794946
G03 X250.291847 Y381.758922 Z-1.000000 I32.359108 J79.126267
G03 X275.549080 Y372.882678 Z-1.000000 I1284.277715 J3614.022427
G02 X300.746265 Y363.927665 Z-1.000000 I-625.720907 J-1800.550167
G02 X301.714640 Y363.376279 Z-1.000000 I-1.130396 J-3.111317
G02 X301.591554 Y363.101396 Z-1.000000 I-0.098948 J-0.120693
G02 X238.458421 Y373.135383 Z-1.000000 I2124.684010 J13571.991768
G02 X175.254383 Y383.477602 Z-1.000000 I2263.977228 J14034.062980
G02 X172.437740 Y384.231248 Z-1.000000 I2.464383 J14.850504
G02 X171.799753 Y385.344480 Z-1.000000 I0.352747 J0.941587
G02 X175.746550 Y403.329640 Z-1.000000 I204.030401 J-35.348312
G02 X177.121570 Y407.623380 Z-1.000000 I44.093988 J-11.753545
G02 X177.443150 Y407.846100 Z-1.000000 I0.321580 J-0.120801
G02 X177.921764 Y407.764494 Z-1.000000 I-0.000000 J-1.444328
G02 X210.396320 Y396.272330 Z-1.000000 I-2409.630485 J-6860.762823
G01 X210.396320 Y396.272330 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X535.728950 Y403.329640
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X539.708957 Y385.327362 Z-1.000000 I-207.626521 J-55.343895 F400.000000
G02 X539.095380 Y384.252843 Z-1.000000 I-0.953112 J-0.168193
G02 X536.195118 Y383.476116 Z-1.000000 I-5.450670 J14.549465
G02 X472.989279 Y373.108495 Z-1.000000 I-2270.095717 J13641.708460
G02 X409.845818 Y363.056233 Z-1.000000 I-2131.448624 J13185.387765
G02 X409.718110 Y363.338702 Z-1.000000 I-0.024898 J0.158847
G02 X410.736981 Y363.923292 Z-1.000000 I2.183911 J-2.626115
G02 X435.926410 Y372.848667 Z-1.000000 I602.972770 J-1661.717933
G03 X461.179337 Y381.690160 Z-1.000000 I-1147.212666 J3317.134301
G03 X463.684030 Y382.686019 Z-1.000000 I-11.899369 J33.575907
G02 X467.165243 Y384.085864 Z-1.000000 I22.347937 J-50.547542
G02 X499.616121 Y395.798923 Z-1.000000 I2971.179232 J-8180.804965
G02 X532.657966 Y407.575382 Z-1.000000 I2905.991271 J-8101.254356
G02 X533.647360 Y407.782480 Z-1.000000 I1.262839 J-3.566179
G02 X534.234387 Y407.438009 Z-1.000000 I0.043282 J-0.598666
G02 X535.728950 Y403.329640 Z-1.000000 I-22.429849 J-10.485675
G01 X535.728950 Y403.329640 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X358.995910 Y364.794818
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X361.640055 Y337.411221 Z-1.000000 I-1785.647663 J-186.240603 F400.000000
G02 X361.523240 Y336.344006 Z-1.000000 I-2.723296 J-0.241915
G02 X361.228678 Y336.344462 Z-1.000000 I-0.147209 J0.046615
G02 X358.976610 Y344.473622 Z-1.000000 I120.835924 J37.852410
G03 X356.436987 Y353.237118 Z-1.000000 I-98.312527 J-23.740811
G03 X355.862720 Y353.648657 Z-1.000000 I-0.574267 J-0.194900
G03 X355.277030 Y353.236078 Z-1.000000 I0.000000 J-0.622007
G03 X351.663810 Y341.905879 Z-1.000000 I161.388005 J-57.708147
G02 X349.140154 Y333.396690 Z-1.000000 I-298.576603 J83.923053
G02 X348.841462 Y333.407753 Z-1.000000 I-0.147841 J0.046158
G02 X348.763782 Y334.602005 Z-1.000000 I3.577717 J0.832367
G02 X350.192730 Y348.228914 Z-1.000000 I1175.047777 J-116.329779
G03 X350.584994 Y351.842924 Z-1.000000 I-832.312413 J92.167408
G03 X352.783310 Y372.589205 Z-1.000000 I-27431.866961 J2917.223039
G02 X355.220147 Y392.653189 Z-1.000000 I641.339103 J-67.712784
G02 X355.555660 Y392.768272 Z-1.000000 I0.196766 J-0.027033
G02 X355.895124 Y392.082920 Z-1.000000 I-0.806686 J-0.826309
G02 X358.995910 Y364.794818 Z-1.000000 I-1477.825077 J-181.747602
G01 X358.995910 Y364.794818 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X345.508370 Y350.486960
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X345.009018 Y350.367546 Z-1.000000 I-0.349587 J0.358089 F400.000000
G02 X334.836354 Y353.705612 Z-1.000000 I121.885090 J388.610992
G02 X324.825802 Y357.279394 Z-1.000000 I135.860107 J396.366131
G02 X323.803910 Y357.926473 Z-1.000000 I1.015212 J2.733707
G02 X323.609292 Y358.483077 Z-1.000000 I0.462958 J0.474201
G02 X325.401761 Y371.607536 Z-1.000000 I593.585027 J-74.384075
G02 X327.412236 Y384.314305 Z-1.000000 I629.191332 J-93.038702
G02 X328.152280 Y387.333124 Z-1.000000 I20.905883 J-3.524824
G02 X328.820304 Y387.391364 Z-1.000000 I0.346349 J-0.112391
G02 X337.340270 Y370.011900 Z-1.000000 I-268.016040 J-142.168097
G02 X345.610153 Y350.987549 Z-1.000000 I-683.092662 J-308.249912
G02 X345.508370 Y350.486960 Z-1.000000 I-0.419263 J-0.175395
G01 X345.508370 Y350.486960 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X385.915250 Y372.568803
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X387.849303 Y358.459834 Z-1.000000 I-389.113143 J-60.526558 F400.000000
G02 X387.655200 Y357.910479 Z-1.000000 I-0.654552 J-0.077696
G02 X386.652938 Y357.276261 Z-1.000000 I-1.994707 J2.043208
G02 X376.637493 Y353.704845 Z-1.000000 I-148.593378 J400.876506
G02 X366.462628 Y350.364345 Z-1.000000 I-134.794075 J393.404509
G02 X365.972870 Y350.481360 Z-1.000000 I-0.147024 J0.468074
G02 X365.871258 Y350.975977 Z-1.000000 I0.314733 J0.322404
G02 X373.897080 Y369.639022 Z-1.000000 I673.946387 J-278.765477
G02 X382.401017 Y387.489996 Z-1.000000 I368.202475 J-164.455093
G02 X382.984040 Y387.568040 Z-1.000000 I0.318332 J-0.161340
G02 X383.446000 Y386.645198 Z-1.000000 I-1.429006 J-1.292383
G02 X385.915250 Y372.568803 Z-1.000000 I-362.613507 J-70.863631
G01 X385.915250 Y372.568803 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X245.940960 Y369.937469
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X313.782786 Y358.593524 Z-1.000000 I-2649.979747 J-16056.588905 F400.000000
G02 X313.792900 Y358.444384 Z-1.000000 I-0.012837 J-0.075783
G02 X312.595043 Y358.253130 Z-1.000000 I-1.254190 J4.008392
G02 X238.963881 Y357.354146 Z-1.000000 I-410.063958 J30570.395554
G02 X165.189647 Y356.631400 Z-1.000000 I-336.035885 J30535.228016
G02 X164.675590 Y356.803371 Z-1.000000 I-0.007154 J0.832912
G02 X164.068822 Y358.478268 Z-1.000000 I1.006163 J1.311861
G02 X170.127950 Y381.027530 Z-1.000000 I277.088973 J-62.366850
G02 X171.004589 Y381.828663 Z-1.000000 I1.127108 J-0.353141
G02 X174.032620 Y381.874816 Z-1.000000 I1.628776 J-7.506110
G03 X177.477821 Y381.269757 Z-1.000000 I34.101217 J184.060892
G03 X245.940960 Y369.937469 Z-1.000000 I12062.208296 J72660.429534
G01 X245.940960 Y369.937469 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X544.186140 Y369.465867
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X546.933162 Y357.200766 Z-1.000000 I-317.888581 J-77.637862 F400.000000
G02 X546.772760 Y356.779270 Z-1.000000 I-0.412822 J-0.084168
G02 X545.932897 Y356.503616 Z-1.000000 I-0.828598 J1.107302
G02 X473.150803 Y357.239189 Z-1.000000 I151.138376 J18555.699150
G02 X400.354191 Y358.260597 Z-1.000000 I223.806103 J18545.486614
G02 X399.647990 Y358.554767 Z-1.000000 I0.016546 J1.034478
G02 X399.762833 Y358.924421 Z-1.000000 I0.151810 J0.155503
G02 X468.935358 Y370.687703 Z-1.000000 I2395.250077 J-13875.710480
G02 X538.235082 Y382.120580 Z-1.000000 I2368.162530 J-14138.735385
G02 X540.052410 Y382.071757 Z-1.000000 I0.780288 J-4.802900
G02 X540.771041 Y381.422991 Z-1.000000 I-0.207789 J-0.952559
G02 X544.186140 Y369.465867 Z-1.000000 I-152.830719 J-50.116554
G01 X544.186140 Y369.465867 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X272.464900 Y349.801146
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X237.922090 Y343.777859 Z-1.000000 F400.000000
G01 X207.080290 Y349.014355 Z-1.000000
G01 X176.238500 Y354.250851 Z-1.000000
G01 X207.697130 Y354.884263 Z-1.000000
G02 X239.157337 Y355.390265 Z-1.000000 I78.201141 J-3883.827894
G02 X273.081730 Y355.671053 Z-1.000000 I54.341466 J-4515.964080
G01 X307.007710 Y355.824433 Z-1.000000
G01 X272.464900 Y349.801146 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X335.211540 Y351.617073
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X341.669823 Y348.548262 Z-1.000000 I-15.741335 J-41.457590 F400.000000
G02 X341.550520 Y347.951351 Z-1.000000 I-0.160891 J-0.278221
G02 X338.343360 Y347.584323 Z-1.000000 I-13.203752 J101.180875
G02 X289.578460 Y342.810677 Z-1.000000 I-4397.012359 J44665.962326
G02 X240.330083 Y338.044353 Z-1.000000 I-4331.131948 J44494.903789
G02 X239.231500 Y337.991620 Z-1.000000 I-1.098583 J11.417076
G02 X238.789843 Y338.347125 Z-1.000000 I0.000000 J0.452096
G02 X238.770550 Y340.099298 Z-1.000000 I4.218733 J0.922645
G02 X240.004402 Y341.368051 Z-1.000000 I1.589069 J-0.311025
G02 X278.646930 Y349.065602 Z-1.000000 I176.070117 J-783.047160
G02 X318.126067 Y355.381449 Z-1.000000 I254.661423 J-1465.293033
G02 X322.257980 Y355.442533 Z-1.000000 I2.298823 J-15.721409
G02 X326.320385 Y354.613134 Z-1.000000 I-2.792402 J-24.040771
G02 X335.211540 Y351.617073 Z-1.000000 I-35.100894 J-118.856720
G01 X335.211540 Y351.617073 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X430.816440 Y349.235047
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X472.128485 Y340.495495 Z-1.000000 I-82.820605 J-493.505931 F400.000000
G02 X472.656560 Y338.521177 Z-1.000000 I-0.290779 J-1.135557
G02 X471.958882 Y338.271211 Z-1.000000 I-0.621820 J0.636930
G02 X420.100920 Y343.016373 Z-1.000000 I374.540368 J4378.940501
G02 X368.784571 Y348.312624 Z-1.000000 I441.346315 J4527.540452
G02 X368.679010 Y348.618376 Z-1.000000 I0.019368 J0.177785
G02 X369.213563 Y348.959822 Z-1.000000 I1.088460 J-1.114888
G02 X378.449343 Y352.396854 Z-1.000000 I243.832542 J-641.084324
G02 X387.581974 Y355.649596 Z-1.000000 I239.537377 J-658.094993
G02 X389.046880 Y356.031997 Z-1.000000 I3.120534 J-8.957098
G02 X389.664192 Y356.036325 Z-1.000000 I0.321391 J-1.814228
G02 X430.816440 Y349.235047 Z-1.000000 I-1349.834564 J-8295.295199
G01 X430.816440 Y349.235047 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X500.077340 Y354.988731
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X535.236990 Y354.250851 Z-1.000000 F400.000000
G01 X504.395200 Y349.014355 Z-1.000000
G01 X473.553410 Y343.777859 Z-1.000000
G01 X439.010590 Y349.801146 Z-1.000000
G01 X404.467780 Y355.824433 Z-1.000000
G01 X434.692740 Y355.775533 Z-1.000000
G02 X464.915915 Y355.569267 Z-1.000000 I-4.687634 J-2901.188510
G02 X500.077340 Y354.988743 Z-1.000000 I-47.278994 J-3928.740693
G01 X500.077340 Y354.988731 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X197.519340 Y348.730506
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X228.353462 Y343.218335 Z-1.000000 I-257.676218 J-1530.392172 F400.000000
G02 X228.669560 Y342.836823 Z-1.000000 I-0.072183 J-0.381512
G02 X228.377555 Y342.486710 Z-1.000000 I-0.355895 J0.000000
G02 X194.886011 Y336.529632 Z-1.000000 I-683.016933 J3742.893396
G02 X161.754773 Y330.936930 Z-1.000000 I-662.556559 J3824.062173
G02 X158.802400 Y330.731644 Z-1.000000 I-2.571025 J15.643063
G02 X158.238511 Y331.396574 Z-1.000000 I0.014023 J0.583458
G02 X160.514260 Y342.874424 Z-1.000000 I104.099336 J-14.675567
G02 X164.102044 Y354.074409 Z-1.000000 I97.257382 J-24.980623
G02 X164.996820 Y354.479457 Z-1.000000 I0.653297 J-0.252341
G03 X166.402292 Y354.100295 Z-1.000000 I2.838280 J7.726413
G03 X197.519340 Y348.730506 Z-1.000000 I674.198682 J3814.027233
G01 X197.519340 Y348.730506 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X550.977990 Y342.809170
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X553.237194 Y331.396432 Z-1.000000 I-100.685260 J-25.861087 F400.000000
G02 X552.673090 Y330.731644 Z-1.000000 I-0.578147 J-0.081144
G02 X549.735140 Y330.935874 Z-1.000000 I-0.379744 J15.771232
G02 X516.595272 Y336.528550 Z-1.000000 I632.956040 J3851.616553
G02 X483.096396 Y342.483743 Z-1.000000 I653.087092 J3770.911248
G02 X482.805930 Y342.831969 Z-1.000000 I0.063503 J0.348226
G02 X483.119445 Y343.212316 Z-1.000000 I0.387470 J0.000000
G02 X514.572990 Y349.008803 Z-1.000000 I311.368248 J-1601.344500
G03 X546.338162 Y354.551676 Z-1.000000 I-8763.537394 J50315.993849
G03 X547.112280 Y354.697354 Z-1.000000 I-5.242353 J29.987044
G02 X547.605953 Y354.408950 Z-1.000000 I0.084950 J-0.421310
G02 X550.977990 Y342.809170 Z-1.000000 I-182.450591 J-59.328097
G01 X550.977990 Y342.809170 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X362.423650 Y323.061503
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X363.198730 Y314.740457 Z-1.000000 I-34.096359 J-7.372591 F400.000000
G02 X361.850400 Y301.984721 Z-1.000000 I-82.149107 J2.234342
G02 X357.451327 Y285.026099 Z-1.000000 I-122.492024 J22.724609
G02 X355.737750 Y283.794182 Z-1.000000 I-1.713577 J0.575820
G02 X354.098600 Y284.973095 Z-1.000000 I0.000000 J1.728986
G02 X349.652480 Y301.927538 Z-1.000000 I123.528404 J41.454191
G02 X348.250426 Y314.629800 Z-1.000000 I79.300595 J15.181572
G02 X349.005710 Y322.980062 Z-1.000000 I34.738171 J1.067210
G02 X352.917837 Y339.146508 Z-1.000000 I309.160768 J-66.257408
G02 X355.335640 Y346.422333 Z-1.000000 I64.857988 J-17.513085
G02 X356.376079 Y346.378951 Z-1.000000 I0.512638 J-0.203524
G02 X362.423650 Y323.061503 Z-1.000000 I-282.440140 J-85.696130
G01 X362.423650 Y323.061503 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X347.942640 Y344.750727
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X347.520400 Y343.026248 Z-1.000000 I6.166584 J-2.423825 F400.000000
G03 X346.408620 Y330.298077 Z-1.000000 I341.221022 J-36.217611
G01 X345.531650 Y317.517032 Z-1.000000
G01 X340.456900 Y317.629523 Z-1.000000
G02 X335.386688 Y317.839470 Z-1.000000 I2.928668 J132.054993
G02 X315.643410 Y319.135007 Z-1.000000 I121.502088 J2002.711555
G01 X295.904660 Y320.528014 Z-1.000000
G01 X319.344430 Y333.352464 Z-1.000000
G02 X346.535052 Y346.151649 Z-1.000000 I118.928364 J-217.370220
G02 X347.942640 Y344.750740 Z-1.000000 I0.400597 J-1.005096
G01 X347.942640 Y344.750727 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X391.620710 Y333.475167
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X415.569400 Y320.528001 Z-1.000000 F400.000000
G01 X396.448200 Y319.135500 Z-1.000000
G02 X377.321680 Y317.858090 Z-1.000000 I-115.853825 J1590.837919
G02 X371.668180 Y317.630016 Z-1.000000 I-8.427222 J138.710667
G01 X366.009390 Y317.517032 Z-1.000000
G01 X364.572800 Y330.464199 Z-1.000000
G02 X364.142112 Y345.222800 Z-1.000000 I90.315697 J10.021193
G02 X365.405900 Y346.422333 Z-1.000000 I1.263788 J-0.065976
G02 X367.545593 Y345.922753 Z-1.000000 I-0.000000 J-4.831928
G02 X391.620710 Y333.475167 Z-1.000000 I-325.554188 J-659.166507
G01 X391.620710 Y333.475167 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X315.156210 Y332.999048
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X292.462920 Y320.806805 Z-1.000000 F400.000000
G01 X267.659850 Y328.495928 Z-1.000000
G02 X245.867010 Y335.944024 Z-1.000000 I118.776485 J383.141943
G02 X245.940960 Y336.622255 Z-1.000000 I0.123359 J0.329697
G02 X300.355423 Y343.135308 Z-1.000000 I177.296283 J-1250.689277
G02 X332.297980 Y345.062940 Z-1.000000 I41.896929 J-428.646943
G02 X334.375297 Y345.056259 Z-1.000000 I0.911828 J-39.440168
G02 X334.973947 Y344.889445 Z-1.000000 I-0.038273 J-1.294949
G02 X335.045191 Y344.579485 Z-1.000000 I-0.105858 J-0.187499
G02 X333.976404 Y343.501813 Z-1.000000 I-3.559293 J2.461126
G02 X330.720164 Y341.467449 Z-1.000000 I-20.162117 J28.648661
G02 X315.156210 Y332.999048 Z-1.000000 I-807.575873 J1465.695659
G01 X315.156210 Y332.999048 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X422.853160 Y340.976948
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X466.131041 Y336.633145 Z-1.000000 I-884.311017 J-9028.270311 F400.000000
G02 X467.208550 Y336.417559 Z-1.000000 I-0.592035 J-5.759531
G02 X467.213933 Y336.094698 Z-1.000000 I-0.048875 J-0.162290
G02 X443.631990 Y328.450763 Z-1.000000 I-319.775426 J946.326330
G01 X418.998240 Y320.802348 Z-1.000000
G01 X396.327150 Y333.010153 Z-1.000000
G02 X376.408136 Y344.524088 Z-1.000000 I180.846712 J335.850182
G02 X376.605510 Y345.217945 Z-1.000000 I0.197374 J0.318856
G02 X379.548387 Y345.083294 Z-1.000000 I-0.000000 J-32.226447
G02 X422.853160 Y340.976948 Z-1.000000 I-641.609794 J-6996.694924
G01 X422.853160 Y340.976948 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X236.071580 Y339.821145
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X235.471595 Y338.527807 Z-1.000000 I-1.693963 J0.000000 F400.000000
G02 X233.295820 Y337.473252 Z-1.000000 I-3.146289 J3.719652
G02 X229.292334 Y336.860207 Z-1.000000 I-8.303898 J40.849644
G02 X210.632769 Y335.036287 Z-1.000000 I-163.624172 J1577.592575
G02 X191.348716 Y333.390668 Z-1.000000 I-138.249424 J1506.253315
G02 X190.836300 Y333.576167 Z-1.000000 I-0.051854 J0.657253
G02 X190.896506 Y333.776126 Z-1.000000 I0.082272 J0.084272
G02 X210.908621 Y337.490655 Z-1.000000 I385.316131 J-2020.135063
G02 X230.433880 Y340.924369 Z-1.000000 I391.563729 J-2169.333104
G02 X234.529500 Y341.462773 Z-1.000000 I8.969220 J-52.381745
G02 X235.579054 Y341.064799 Z-1.000000 I0.117222 J-1.273807
G02 X236.071580 Y339.821145 Z-1.000000 I-1.323881 J-1.243654
G01 X236.071580 Y339.821145 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X500.987560 Y337.280430
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X520.562105 Y333.803943 Z-1.000000 I-149.580109 J-899.064711 F400.000000
G02 X520.632800 Y333.569916 Z-1.000000 I-0.025577 J-0.135417
G02 X520.125381 Y333.386314 Z-1.000000 I-0.455894 J0.466977
G02 X500.827595 Y335.037578 Z-1.000000 I120.838481 J1525.789262
G02 X482.151240 Y336.865014 Z-1.000000 I146.843904 J1597.091019
G02 X478.179680 Y337.473252 Z-1.000000 I4.269437 J41.148298
G02 X476.031377 Y338.550260 Z-1.000000 I0.916747 J4.509739
G02 X475.403910 Y339.951712 Z-1.000000 I1.251342 J1.401452
G02 X476.083392 Y341.016022 Z-1.000000 I1.173285 J0.000000
G02 X478.179680 Y341.239565 Z-1.000000 I1.376418 J-2.966643
G03 X480.969109 Y340.683116 Z-1.000000 I12.510107 J55.442217
G03 X500.987560 Y337.280430 Z-1.000000 I495.176768 J2852.603571
G01 X500.987560 Y337.280430 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X261.361860 Y328.427964
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X274.942541 Y323.974735 Z-1.000000 I-317.169717 J-990.183432 F400.000000
G02 X278.236075 Y322.722531 Z-1.000000 I-14.160228 J-42.201505
G02 X279.438767 Y322.027947 Z-1.000000 I-2.259145 J-5.300314
G02 X279.418183 Y321.846973 Z-1.000000 I-0.063468 J-0.084439
G02 X278.964114 Y321.753828 Z-1.000000 I-0.439570 J0.989506
G02 X276.782750 Y321.855141 Z-1.000000 I0.442856 J33.069104
G03 X241.598470 Y324.424103 Z-1.000000 I-213.785615 J-2685.764686
G03 X184.547110 Y327.982907 Z-1.000000 I-468.971573 J-7059.029366
G02 X172.499253 Y328.951648 Z-1.000000 I16.013389 J274.554104
G02 X172.402670 Y329.228557 Z-1.000000 I0.016520 J0.161060
G02 X173.153393 Y329.584288 Z-1.000000 I0.893159 J-0.914877
G02 X200.456729 Y332.493200 Z-1.000000 I278.354440 J-2483.071846
G02 X226.763146 Y335.017348 Z-1.000000 I266.244805 J-2636.433194
G02 X233.604240 Y335.409100 Z-1.000000 I9.270591 J-101.962404
G02 X239.049059 Y334.800424 Z-1.000000 I0.482875 J-20.337743
G02 X261.361860 Y328.427964 Z-1.000000 I-89.015429 J-353.932124
G01 X261.361860 Y328.427964 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X508.547490 Y332.622894
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X538.512897 Y329.481413 Z-1.000000 I-210.138890 J-2148.921391 F400.000000
G02 X539.097510 Y329.204457 Z-1.000000 I-0.110678 J-0.989117
G02 X539.005497 Y328.942727 Z-1.000000 I-0.106870 J-0.109468
G02 X525.692670 Y327.896456 Z-1.000000 I-34.422882 J352.779523
G03 X511.800288 Y327.019857 Z-1.000000 I121.061002 J-2029.101858
G03 X471.086060 Y324.263060 Z-1.000000 I1159.820799 J-17431.036749
G02 X430.435707 Y321.648444 Z-1.000000 I-308.023341 J4471.636330
G02 X429.954600 Y321.805121 Z-1.000000 I-0.041253 J0.690329
G02 X430.049780 Y322.311680 Z-1.000000 I0.187153 J0.227056
G02 X450.715030 Y328.856042 Z-1.000000 I302.370838 J-918.902209
G02 X471.470212 Y334.942661 Z-1.000000 I307.903524 J-1011.510856
G02 X474.787070 Y335.452109 Z-1.000000 I3.725201 J-13.201383
G02 X478.480987 Y335.352621 Z-1.000000 I0.970248 J-32.601485
G02 X508.547490 Y332.622894 Z-1.000000 I-181.722632 J-2168.526858
G01 X508.547490 Y332.622894 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X202.112540 Y310.899790
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X213.939063 Y297.390589 Z-1.000000 I-149.349095 J-142.677804 F400.000000
G02 X213.676360 Y296.470541 Z-1.000000 I-0.457731 J-0.366832
G02 X148.784100 Y274.761620 Z-1.000000 I-706.950537 J2005.372677
G02 X148.067960 Y275.689690 Z-1.000000 I-0.188438 J0.594930
G02 X150.980816 Y279.793058 Z-1.000000 I39.270818 J-24.791599
G02 X168.753550 Y301.913865 Z-1.000000 I2090.982918 J-1661.778581
G02 X187.261055 Y324.438782 Z-1.000000 I2001.142536 J-1625.364721
G02 X187.910250 Y324.743357 Z-1.000000 I0.649195 J-0.539586
G02 X188.742548 Y324.404827 Z-1.000000 I-0.000000 J-1.192395
G02 X202.112540 Y310.899790 Z-1.000000 I-372.321138 J-381.968737
G01 X202.112540 Y310.899790 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X259.181140 Y303.560783
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X301.126067 Y290.208541 Z-1.000000 I-1008.576856 J-3240.917362 F400.000000
G02 X301.409480 Y289.816121 Z-1.000000 I-0.127037 J-0.390301
G02 X301.036210 Y289.137777 Z-1.000000 I-0.811214 J0.004514
G02 X278.679148 Y275.026484 Z-1.000000 I-986.374999 J1537.985490
G02 X256.180316 Y261.274257 Z-1.000000 I-972.891019 J1566.382704
G02 X254.870930 Y260.910820 Z-1.000000 I-1.309386 J2.177002
G02 X253.807801 Y261.310115 Z-1.000000 I-0.000000 J1.614946
G02 X249.017960 Y266.029467 Z-1.000000 I37.270765 J42.617816
G03 X244.301867 Y271.049405 Z-1.000000 I-122.963926 J-110.795928
G03 X225.126230 Y290.422902 Z-1.000000 I-1852.148947 J-1814.056382
G02 X205.876003 Y309.726753 Z-1.000000 I6357.275907 J6358.866574
G02 X198.521930 Y317.182321 Z-1.000000 I940.915788 J935.461513
G01 X191.197240 Y324.666964 Z-1.000000
G01 X204.073450 Y320.685102 Z-1.000000
G02 X216.946933 Y316.694425 Z-1.000000 I-2906.323280 J-9398.221825
G02 X259.181140 Y303.560783 Z-1.000000 I-31469.676370 J-101272.942396
G01 X259.181140 Y303.560783 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X268.763890 Y320.525002
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X330.453636 Y316.508496 Z-1.000000 I730.935343 J10750.721954 F400.000000
G03 X336.307420 Y316.321907 Z-1.000000 I6.006642 J96.526392
G02 X341.724048 Y315.894712 Z-1.000000 I-0.052452 J-35.218823
G02 X342.115590 Y315.409354 Z-1.000000 I-0.072119 J-0.458788
G02 X341.716051 Y314.713261 Z-1.000000 I-0.895667 J0.051381
G02 X323.179100 Y302.840511 Z-1.000000 I-373.171138 J562.225611
G01 X304.294360 Y291.174946 Z-1.000000
G01 X252.603160 Y307.369935 Z-1.000000
G02 X200.928788 Y323.612164 Z-1.000000 I8749.496330 J27926.692733
G02 X199.678260 Y324.059061 Z-1.000000 I5.290209 J16.776418
G02 X199.804171 Y324.632780 Z-1.000000 I0.110701 J0.276381
G02 X268.763890 Y320.525002 Z-1.000000 I-210.392554 J-4112.872868
G01 X268.763890 Y320.525002 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X511.797230 Y324.059061
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X510.546702 Y323.612164 Z-1.000000 I-6.540765 J16.329590 F400.000000
G02 X458.883270 Y307.373355 Z-1.000000 I-8799.645502 J27905.585706
G01 X407.203000 Y291.181799 Z-1.000000
G01 X388.057830 Y303.145028 Z-1.000000
G02 X369.387165 Y315.041719 Z-1.000000 I665.826721 J1065.547548
G02 X369.418800 Y315.253723 Z-1.000000 I0.064176 J0.098786
G02 X369.934478 Y315.345505 Z-1.000000 I0.699301 J-2.434481
G02 X433.459070 Y319.908465 Z-1.000000 I2711.376222 J-37302.749080
G02 X511.644077 Y324.753879 Z-1.000000 I309.448547 J-4360.008918
G02 X511.797230 Y324.059061 Z-1.000000 I0.019078 J-0.360083
G01 X511.797230 Y324.059061 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X512.992280 Y317.220524
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X505.599618 Y309.726616 Z-1.000000 I-962.454773 J942.058258 F400.000000
G02 X486.349260 Y290.422902 Z-1.000000 I-6405.739777 J6368.783782
G03 X467.173621 Y271.049402 Z-1.000000 I1832.966964 J-1833.423531
G03 X462.457540 Y266.029467 Z-1.000000 I118.242184 J-115.810325
G02 X457.667688 Y261.310115 Z-1.000000 I-42.061209 J37.899156
G02 X456.604560 Y260.910820 Z-1.000000 I-1.063128 J1.215650
G02 X454.643967 Y261.459814 Z-1.000000 I0.000000 J3.775376
G02 X432.667706 Y275.149063 Z-1.000000 I608.333086 J1001.082106
G02 X411.083183 Y289.281233 Z-1.000000 I600.928735 J941.367510
G02 X411.024120 Y290.007425 Z-1.000000 I0.259555 J0.386608
G02 X411.782264 Y290.454499 Z-1.000000 I1.375126 J-1.465559
G02 X464.796791 Y307.452274 Z-1.000000 I4817.324678 J-14933.628876
G02 X517.580796 Y324.170654 Z-1.000000 I4775.902044 J-14987.002546
G02 X519.777480 Y324.663820 Z-1.000000 I4.022307 J-12.777486
G02 X519.865451 Y324.496530 Z-1.000000 I0.013690 J-0.099576
G02 X512.992280 Y317.220524 Z-1.000000 I-208.315533 J189.897830
G01 X512.992280 Y317.220524 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X543.507660 Y300.956705
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X562.399609 Y277.117885 Z-1.000000 I-4023.661044 J-3208.103353 F400.000000
G02 X563.396840 Y275.706576 Z-1.000000 I-13.047491 J-10.277333
G02 X562.739196 Y274.854211 Z-1.000000 I-0.484652 J-0.305951
G02 X497.813510 Y296.469939 Z-1.000000 I698.917611 J2207.605415
G02 X497.553458 Y297.378656 Z-1.000000 I0.190855 J0.546187
G02 X509.410840 Y310.894407 Z-1.000000 I163.158779 J-131.180236
G02 X522.795028 Y324.416122 Z-1.000000 I399.716567 J-382.266240
G02 X523.598750 Y324.743357 Z-1.000000 I0.803722 J-0.823393
G02 X524.369048 Y324.383576 Z-1.000000 I0.000000 J-1.004504
G02 X543.507660 Y300.956705 Z-1.000000 I-975.234523 J-816.250129
G01 X543.507660 Y300.956705 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X182.220050 Y322.033485
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X180.033333 Y318.945360 Z-1.000000 I-29.883440 J18.842337 F400.000000
G02 X163.058119 Y297.684616 Z-1.000000 I-2628.343580 J2081.141096
G02 X145.543348 Y276.117051 Z-1.000000 I-2559.199477 J2060.400885
G02 X144.861610 Y275.631000 Z-1.000000 I-1.235923 J1.012381
G02 X144.272248 Y275.909981 Z-1.000000 I-0.158213 J0.427785
G02 X140.081330 Y289.481457 Z-1.000000 I204.586226 J70.609751
G02 X136.758775 Y303.383258 Z-1.000000 I214.868605 J58.701924
G02 X137.100310 Y304.000095 Z-1.000000 I0.530915 J0.109010
G03 X137.984681 Y304.358260 Z-1.000000 I-5.983277 J16.044666
G03 X159.583930 Y313.843554 Z-1.000000 I-4199.361669 J9591.827972
G02 X181.685401 Y322.697769 Z-1.000000 I141.660656 J-321.595168
G02 X182.220050 Y322.033485 Z-1.000000 I0.153359 J-0.423868
G01 X182.220050 Y322.033485 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X553.125230 Y313.282298
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X574.695090 Y303.920939 Z-1.000000 I-2867.434796 J-6636.495337 F400.000000
G02 X574.974280 Y303.757507 Z-1.000000 I-0.546745 J-1.254186
G02 X575.207306 Y303.080156 Z-1.000000 I-0.387819 J-0.512179
G02 X571.351204 Y289.385626 Z-1.000000 I-475.699717 J126.557246
G02 X567.117796 Y275.873863 Z-1.000000 I-461.653459 J137.222833
G02 X566.613890 Y275.631000 Z-1.000000 I-0.369099 J0.121630
G02 X565.907950 Y276.134381 Z-1.000000 I0.573721 J1.551280
G02 X548.376084 Y297.736302 Z-1.000000 I2454.294857 J2009.792245
G02 X531.402980 Y319.020588 Z-1.000000 I2525.770509 J2031.579496
G02 X529.125860 Y322.238160 Z-1.000000 I29.130710 J23.030764
G02 X529.554416 Y322.765319 Z-1.000000 I0.303215 J0.191278
G02 X553.125230 Y313.282298 Z-1.000000 I-175.296173 J-469.747875
G01 X553.125230 Y313.282298 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X348.442720 Y298.519385
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X349.944630 Y288.941120 Z-1.000000 F400.000000
G01 X324.184630 Y259.945816 Z-1.000000
G02 X298.412632 Y231.288234 Z-1.000000 I-2380.255865 J2114.666321
G02 X298.088460 Y231.278671 Z-1.000000 I-0.166704 J0.151753
G02 X298.037116 Y231.645332 Z-1.000000 I0.211399 J0.216528
G02 X321.462620 Y272.185019 Z-1.000000 I4202.158758 J-2401.142762
G02 X344.987536 Y310.542713 Z-1.000000 I910.360053 J-531.934394
G02 X346.056860 Y310.430404 Z-1.000000 I0.506360 J-0.325622
G02 X346.742619 Y308.045156 Z-1.000000 I-14.122894 J-5.351532
G02 X348.442720 Y298.519385 Z-1.000000 I-225.094084 J-45.088018
G01 X348.442720 Y298.519385 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X390.217010 Y271.858798
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X413.464065 Y231.691398 Z-1.000000 I-3847.644229 J-2253.651474 F400.000000
G02 X413.409380 Y231.300507 Z-1.000000 I-0.279982 J-0.160102
G02 X413.063672 Y231.310549 Z-1.000000 I-0.168002 J0.172069
G02 X387.270260 Y259.969000 Z-1.000000 I2208.624564 J2013.760878
G01 X361.489640 Y288.987513 Z-1.000000
G01 X363.025290 Y298.133632 Z-1.000000
G02 X365.826875 Y312.304377 Z-1.000000 I247.058047 J-41.481570
G02 X366.284770 Y312.439371 Z-1.000000 I0.266766 J-0.060781
G02 X366.610753 Y312.023778 Z-1.000000 I-1.398510 J-1.432606
G02 X390.217010 Y271.858798 Z-1.000000 I-13419.312552 J-7913.983470
G01 X390.217010 Y271.858798 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X329.203160 Y300.876974
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X320.597212 Y292.647673 Z-1.000000 I1769.135057 J-1858.721363 F400.000000
G03 X285.965570 Y259.339250 Z-1.000000 I28952.769322 J-30137.610825
G01 X251.353050 Y226.010885 Z-1.000000
G01 X233.181560 Y228.248769 Z-1.000000
G02 X215.135755 Y230.748222 Z-1.000000 I73.767139 J598.985674
G02 X214.330160 Y231.150416 Z-1.000000 I0.224289 J1.457152
G02 X214.376597 Y231.738538 Z-1.000000 I0.268204 J0.274717
G02 X229.445420 Y241.846051 Z-1.000000 I209.777700 J-296.461393
G02 X335.042016 Y308.455657 Z-1.000000 I10826.875390 J-17046.902363
G02 X336.605200 Y308.979539 Z-1.000000 I1.869446 J-2.984045
G02 X336.785003 Y308.615578 Z-1.000000 I0.018961 J-0.217026
G02 X329.203160 Y300.876974 Z-1.000000 I-114.541849 J104.638140
G01 X329.203160 Y300.876974 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X338.692330 Y305.172061
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X336.306080 Y301.223865 Z-1.000000 I137.289666 J-85.671715 F400.000000
G03 X313.147710 Y261.513014 Z-1.000000 I13865.716326 J-8112.745254
G02 X290.030120 Y221.778803 Z-1.000000 I-48725.193064 J28322.002761
G02 X289.877920 Y221.554735 Z-1.000000 I-1.689524 J0.983897
G02 X289.559461 Y221.420851 Z-1.000000 I-0.279797 J0.219843
G02 X274.689558 Y223.139134 Z-1.000000 I131.631043 J1204.326634
G02 X260.345016 Y224.968875 Z-1.000000 I154.766509 J1270.459338
G02 X257.044000 Y225.554542 Z-1.000000 I5.143981 J38.588880
G02 X256.609688 Y226.981410 Z-1.000000 I0.185796 J0.836085
G02 X297.138330 Y267.639022 Z-1.000000 I918.658257 J-875.215321
G02 X340.373135 Y308.909161 Z-1.000000 I9363.219973 J-9765.674605
G02 X340.725890 Y309.062220 Z-1.000000 I0.385665 J-0.405815
G02 X340.828115 Y308.904121 Z-1.000000 I0.006360 J-0.107986
G02 X338.692330 Y305.172061 Z-1.000000 I-50.011894 J26.143664
G01 X338.692330 Y305.172061 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X446.927340 Y236.382998
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X454.871132 Y226.846299 Z-1.000000 I-44.904024 J-45.480566 F400.000000
G02 X454.329370 Y225.499875 Z-1.000000 I-0.722586 J-0.491459
G02 X450.955185 Y224.922574 Z-1.000000 I-9.089847 J42.978678
G02 X436.695366 Y223.127636 Z-1.000000 I-177.594645 J1353.352921
G02 X421.893512 Y221.430241 Z-1.000000 I-154.365200 J1280.729133
G02 X421.597570 Y221.554735 Z-1.000000 I-0.035794 J0.328908
G02 X421.445335 Y221.778817 Z-1.000000 I1.535554 J1.206966
G02 X398.327780 Y261.513014 Z-1.000000 I48803.858088 J28420.920992
G03 X375.181358 Y301.231299 Z-1.000000 I-17667.304224 J-10269.280811
G03 X372.643070 Y305.473158 Z-1.000000 I-204.022686 J-119.204879
G02 X373.043682 Y305.807993 Z-1.000000 I0.225010 J0.137861
G02 X398.431326 Y282.689942 Z-1.000000 I-1083.490818 J-1215.360591
G02 X423.629318 Y258.974743 Z-1.000000 I-1995.011392 J-2144.993824
G02 X446.927340 Y236.382998 Z-1.000000 I-1242.056920 J-1304.195899
G01 X446.927340 Y236.382998 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X423.785220 Y278.567827
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X496.992946 Y231.832511 Z-1.000000 I-3377.017040 J-5370.574974 F400.000000
G02 X497.068310 Y231.075214 Z-1.000000 I-0.264906 J-0.408761
G02 X496.164941 Y230.625050 Z-1.000000 I-1.151768 J1.179816
G02 X478.196360 Y228.199245 Z-1.000000 I-80.160905 J526.011080
G01 X460.087200 Y226.067972 Z-1.000000
G01 X423.641790 Y261.161838 Z-1.000000
G03 X387.174240 Y296.232555 Z-1.000000 I-27710.367108 J-28777.551512
G03 X380.424590 Y302.671018 Z-1.000000 I-940.496597 J-979.195965
G02 X374.333668 Y308.815020 Z-1.000000 I94.874790 J100.146281
G02 X374.451760 Y309.086319 Z-1.000000 I0.118092 J0.109948
G02 X375.190202 Y308.876375 Z-1.000000 I0.000000 J-1.403644
G02 X423.785220 Y278.567827 Z-1.000000 I-4109.517309 J-6643.079979
G01 X423.785220 Y278.567827 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X142.301580 Y273.159983
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X141.830668 Y271.719983 Z-1.000000 I-2.387108 J-0.016363 F400.000000
G02 X130.405350 Y256.763019 Z-1.000000 I-565.806654 J420.365853
G02 X118.621803 Y242.178769 Z-1.000000 I-530.405494 J416.496009
G02 X117.946290 Y242.143632 Z-1.000000 I-0.353974 J0.294224
G02 X117.837783 Y242.551961 Z-1.000000 I0.288147 J0.295152
G02 X125.384056 Y268.574517 Z-1.000000 I3025.764559 J-863.335129
G02 X132.921538 Y293.779458 Z-1.000000 I3131.280412 J-922.672926
G02 X134.477830 Y298.246833 Z-1.000000 I55.753920 J-16.918136
G02 X135.033990 Y298.241368 Z-1.000000 I0.277033 J-0.109314
G02 X138.622273 Y287.739515 Z-1.000000 I-193.093451 J-71.840294
G02 X141.839947 Y276.598970 Z-1.000000 I-261.923496 J-81.685160
G02 X142.301580 Y273.159983 Z-1.000000 I-13.278370 J-3.532901
G01 X142.301580 Y273.159983 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X585.669790 Y270.090842
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X593.650505 Y242.609894 Z-1.000000 I-1575.847508 J-472.539628 F400.000000
G02 X593.524330 Y242.138886 Z-1.000000 I-0.459850 J-0.129219
G02 X593.141608 Y242.156394 Z-1.000000 I-0.183179 J0.187618
G02 X580.907740 Y256.938473 Z-1.000000 I500.510657 J426.683550
G01 X568.699000 Y272.136229 Z-1.000000
G01 X572.269080 Y285.786720 Z-1.000000
G02 X576.075865 Y298.432868 Z-1.000000 I174.632917 J-45.672529
G02 X576.623270 Y298.540920 Z-1.000000 I0.304914 J-0.104095
G02 X577.194781 Y297.527088 Z-1.000000 I-2.146869 J-1.878219
G02 X585.669790 Y270.090842 Z-1.000000 I-1654.911070 J-526.226258
G01 X585.669790 Y270.090842 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X215.811630 Y290.719410
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G03 X215.229707 Y286.797820 Z-1.000000 I150.316377 J-24.309330 F400.000000
G03 X213.820660 Y275.983476 Z-1.000000 I1142.089015 J-154.306667
G02 X212.367461 Y265.234222 Z-1.000000 I-563.904812 J70.761838
G02 X212.238850 Y265.068425 Z-1.000000 I-0.210500 J0.030506
G02 X212.006588 Y265.035932 Z-1.000000 I-0.178242 J0.427732
G02 X180.556360 Y268.787333 Z-1.000000 I983.961164 J8382.841937
G02 X151.642406 Y272.949016 Z-1.000000 I78.604893 J648.642946
G02 X151.565060 Y273.623949 Z-1.000000 I0.058151 J0.348562
G02 X156.528820 Y275.491920 Z-1.000000 I30.287350 J-72.953510
G02 X184.597082 Y284.886689 Z-1.000000 I1411.555436 J-4170.604005
G02 X213.425265 Y294.315130 Z-1.000000 I1346.689788 J-4068.821001
G02 X214.846650 Y294.568754 Z-1.000000 I1.637534 J-5.067508
G02 X215.673773 Y293.949994 Z-1.000000 I0.033238 J-0.817775
G02 X215.811630 Y290.719410 Z-1.000000 I-7.779488 J-1.950203
G01 X215.811630 Y290.719410 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X249.779430 Y261.156286
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X248.303137 Y260.994033 Z-1.000000 I-1.128040 J3.466397 F400.000000
G02 X232.245383 Y262.698161 Z-1.000000 I76.757831 J799.785106
G02 X216.146672 Y264.736885 Z-1.000000 I91.981759 J790.911260
G02 X215.258880 Y265.170967 Z-1.000000 I0.214116 J1.562820
G02 X215.021139 Y265.797513 Z-1.000000 I0.555781 J0.569268
G02 X216.360030 Y280.426089 Z-1.000000 I381.278003 J-27.521196
G01 X217.994600 Y295.160421 Z-1.000000
G01 X234.517600 Y278.363550 Z-1.000000
G02 X249.962484 Y261.829966 Z-1.000000 I-312.461592 J-307.367410
G02 X249.779430 Y261.156286 Z-1.000000 I-0.312178 J-0.276884
G01 X249.779430 Y261.156286 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X496.212090 Y265.166571
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X495.286701 Y264.712879 Z-1.000000 I-1.152877 J1.180911 F400.000000
G02 X479.286636 Y262.660199 Z-1.000000 I-104.031841 J747.514420
G02 X463.326912 Y260.956049 Z-1.000000 I-88.873834 J756.736998
G02 X461.794040 Y261.124394 Z-1.000000 I-0.362067 J3.766162
G02 X461.610168 Y261.796090 Z-1.000000 I0.128971 J0.396320
G02 X476.915910 Y278.303680 Z-1.000000 I326.887600 J-287.738789
G01 X493.292150 Y295.074765 Z-1.000000
G01 X495.045560 Y280.407131 Z-1.000000
G02 X496.471623 Y265.855012 Z-1.000000 I-340.952694 J-40.758240
G02 X496.212090 Y265.166571 Z-1.000000 I-0.866851 J-0.066350
G01 X496.212090 Y265.166571 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X528.789760 Y284.256402
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X559.791615 Y273.673657 Z-1.000000 I-729.994921 J-2189.200601 F400.000000
G02 X560.824430 Y273.066582 Z-1.000000 I-1.017691 J-2.913495
G02 X560.680967 Y272.588393 Z-1.000000 I-0.182904 J-0.205741
G02 X530.291050 Y268.646371 Z-1.000000 I-148.603979 J1026.508456
G02 X499.265657 Y265.511872 Z-1.000000 I-131.315336 J1144.651904
G02 X498.843680 Y265.898586 Z-1.000000 I-0.033790 J0.386714
G03 X498.714549 Y267.924999 Z-1.000000 I-15.964518 J-0.000000
G03 X495.744490 Y290.117216 Z-1.000000 I-1925.483913 J-246.399135
G02 X495.751260 Y294.078760 Z-1.000000 I14.006684 J1.956842
G02 X496.391290 Y294.633671 Z-1.000000 I0.640030 J-0.091648
G02 X497.637517 Y294.438958 Z-1.000000 I-0.000000 J-4.085503
G02 X528.789760 Y284.256402 Z-1.000000 I-841.440548 J-2627.025349
G01 X528.789760 Y284.256402 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X362.324460 Y276.885250
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X365.604220 Y272.385091 Z-1.000000 F400.000000
G01 X360.691170 Y267.588666 Z-1.000000
G01 X355.778130 Y262.792241 Z-1.000000
G01 X350.881990 Y267.459712 Z-1.000000
G01 X345.985840 Y272.127184 Z-1.000000
G01 X349.319710 Y276.730582 Z-1.000000
G02 X353.114948 Y280.534069 Z-1.000000 I13.870846 J-10.045515
G02 X355.840520 Y281.438595 Z-1.000000 I2.781804 J-3.823611
G02 X358.569831 Y280.588243 Z-1.000000 I0.055047 J-4.628529
G02 X362.324460 Y276.885250 Z-1.000000 I-9.414596 J-13.300867
G01 X362.324460 Y276.885250 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X175.409010 Y251.275719
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X204.874465 Y232.465473 Z-1.000000 I-1341.667527 J-2134.148779 F400.000000
G02 X205.166720 Y231.987164 Z-1.000000 I-0.352499 J-0.543823
G02 X204.889317 Y231.692224 Z-1.000000 I-0.266743 J-0.027042
G02 X169.144880 Y233.420433 Z-1.000000 I83.270975 J2092.801966
G03 X133.055001 Y235.323446 Z-1.000000 I-245.988627 J-4321.908037
G03 X125.349540 Y235.546466 Z-1.000000 I-9.549473 J-196.714454
G02 X118.198468 Y236.172840 Z-1.000000 I0.430919 J46.053361
G02 X117.639090 Y236.833319 Z-1.000000 I0.110237 J0.660479
G02 X118.168764 Y238.396370 Z-1.000000 I2.571093 J-0.000000
G02 X130.776574 Y254.403472 Z-1.000000 I590.307683 J-451.980309
G02 X143.835882 Y270.069449 Z-1.000000 I564.411608 J-457.222753
G02 X144.754020 Y270.304922 Z-1.000000 I0.627238 J-0.537975
G02 X145.669782 Y269.853897 Z-1.000000 I-1.838254 J-4.887588
G02 X175.409010 Y251.275719 Z-1.000000 I-3250.137073 J-5235.785425
G01 X175.409010 Y251.275719 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X184.583090 Y266.437513
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X207.610488 Y263.433506 Z-1.000000 I-119.474594 J-1005.600797 F400.000000
G02 X208.379060 Y263.055316 Z-1.000000 I-0.193878 J-1.364062
G02 X208.265638 Y262.065235 Z-1.000000 I-0.440381 J-0.451088
G02 X196.553565 Y255.460522 Z-1.000000 I-109.681752 J180.810768
G02 X184.174686 Y249.440070 Z-1.000000 I-109.935915 J210.306828
G02 X182.210620 Y249.376003 Z-1.000000 I-1.059731 J2.349894
G02 X179.868036 Y250.523424 Z-1.000000 I4.723617 J12.608802
G02 X164.992550 Y259.856538 Z-1.000000 I586.671841 J951.581958
G02 X149.989606 Y269.610111 Z-1.000000 I582.615912 J912.595146
G02 X149.714550 Y270.122035 Z-1.000000 I0.338858 J0.511924
G02 X149.956520 Y270.355341 Z-1.000000 I0.233461 J0.000000
G02 X155.574500 Y269.908328 Z-1.000000 I-2.392018 J-65.588884
G03 X161.435605 Y269.197340 Z-1.000000 I220.372791 J1792.151291
G03 X184.583090 Y266.437513 Z-1.000000 I3346.826623 J27972.414952
G01 X184.583090 Y266.437513 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X561.760940 Y270.080556
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X561.500839 Y269.597350 Z-1.000000 I-0.578892 J0.000000 F400.000000
G02 X546.462215 Y259.842633 Z-1.000000 I-635.113770 J962.670638
G02 X531.549604 Y250.488192 Z-1.000000 I-638.630236 J1001.524151
G02 X529.328870 Y249.399970 Z-1.000000 I-6.691061 J10.844411
G02 X527.430040 Y249.461819 Z-1.000000 I-0.874397 J2.334053
G02 X514.991251 Y255.492365 Z-1.000000 I101.966089 J226.161980
G02 X503.193362 Y262.087563 Z-1.000000 I102.377675 J196.989368
G02 X503.081690 Y263.040936 Z-1.000000 I0.311674 J0.519734
G02 X503.834150 Y263.411495 Z-1.000000 I0.943370 J-0.966360
G02 X530.579080 Y266.967504 Z-1.000000 I197.533601 J-1383.307384
G02 X561.403883 Y270.403741 Z-1.000000 I164.646866 J-1336.991492
G02 X561.760940 Y270.080556 Z-1.000000 I0.032265 J-0.323185
G01 X561.760940 Y270.080556 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X580.807300 Y254.296937
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X593.475344 Y237.921053 Z-1.000000 I-463.118836 J-371.346918 F400.000000
G02 X593.836400 Y236.833319 Z-1.000000 I-1.457949 J-1.087734
G02 X593.277070 Y236.172154 Z-1.000000 I-0.670435 J0.000000
G02 X586.125950 Y235.537602 Z-1.000000 I-7.635615 J45.437747
G03 X578.420425 Y235.309937 Z-1.000000 I2.128282 J-202.547374
G03 X542.330610 Y233.408956 Z-1.000000 I216.185280 J-4447.773108
G02 X506.592573 Y231.688371 Z-1.000000 I-118.191924 J2082.932707
G02 X506.318150 Y231.984550 Z-1.000000 I-0.010501 J0.265493
G02 X506.658224 Y232.531207 Z-1.000000 I0.749375 J-0.087076
G02 X536.171626 Y251.425244 Z-1.000000 I2368.824213 J-3667.720686
G02 X565.752523 Y270.030733 Z-1.000000 I2350.691554 J-3704.550098
G02 X566.928500 Y270.444354 Z-1.000000 I1.516064 J-2.431828
G02 X567.625854 Y270.173152 Z-1.000000 I0.094395 J-0.789448
G02 X580.807300 Y254.296937 Z-1.000000 I-458.617675 J-394.183793
G01 X580.807300 Y254.296937 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X235.249780 Y260.395306
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X249.746409 Y258.486336 Z-1.000000 I-54.618500 J-470.768270 F400.000000
G02 X250.401770 Y258.161744 Z-1.000000 I-0.172915 J-1.173012
G02 X250.356715 Y257.675873 Z-1.000000 I-0.219353 J-0.224684
G02 X230.822540 Y244.875703 Z-1.000000 I-772.500185 J1157.597156
G02 X211.038997 Y232.374407 Z-1.000000 I-778.504826 J1210.094420
G02 X209.744990 Y232.005519 Z-1.000000 I-1.294007 J2.085149
G02 X208.670362 Y232.303472 Z-1.000000 I-0.000000 J2.086910
G02 X197.145510 Y239.407986 Z-1.000000 I303.272088 J504.864096
G02 X185.829293 Y246.756747 Z-1.000000 I299.732996 J473.940346
G02 X185.792380 Y247.153811 Z-1.000000 I0.142313 J0.213478
G02 X186.238739 Y247.494774 Z-1.000000 I1.623853 J-1.663157
G02 X199.723050 Y255.338625 Z-1.000000 I908.691099 J-1546.612415
G02 X213.438371 Y262.149634 Z-1.000000 I62.086759 J-107.809273
G02 X216.949670 Y262.608524 Z-1.000000 I2.586679 J-6.129358
G03 X220.655221 Y262.124464 Z-1.000000 I27.224585 J193.983063
G03 X235.249780 Y260.395306 Z-1.000000 I363.604373 J3006.462605
G01 X235.249780 Y260.395306 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X509.540620 Y256.516913
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X521.621464 Y249.640590 Z-1.000000 I-547.668698 J-976.236095 F400.000000
G02 X523.879040 Y248.237302 Z-1.000000 I-20.468837 J-35.447402
G02 X523.993489 Y246.791033 Z-1.000000 I-0.512841 J-0.768246
G02 X514.308590 Y239.382946 Z-1.000000 I-44.310238 J47.893830
G02 X502.468147 Y232.205885 Z-1.000000 I-253.901557 J405.521656
G02 X501.730500 Y232.005519 Z-1.000000 I-0.737647 J1.257641
G02 X500.401021 Y232.384675 Z-1.000000 I-0.000000 J2.520429
G02 X480.644144 Y244.881811 Z-1.000000 I736.246932 J1185.807469
G02 X461.144669 Y257.685991 Z-1.000000 I730.469575 J1133.681456
G02 X461.098610 Y258.186037 Z-1.000000 I0.179788 J0.268705
G02 X461.783550 Y258.527627 Z-1.000000 I0.874218 J-0.895439
G02 X478.717930 Y260.881975 Z-1.000000 I94.449153 J-617.274892
G03 X495.755940 Y263.036915 Z-1.000000 I-784.634969 J6272.141235
G03 X496.587060 Y263.166721 Z-1.000000 I-1.921702 J15.029872
G02 X497.370229 Y263.051706 Z-1.000000 I0.222843 J-1.206512
G02 X509.540620 Y256.516913 Z-1.000000 I-170.630600 J-332.382712
G01 X509.540620 Y256.516913 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X160.372290 Y231.978553
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X203.169432 Y229.407803 Z-1.000000 I-377.498381 J-6642.011966 F400.000000
G02 X203.532540 Y229.246388 Z-1.000000 I-0.036104 J-0.570336
G02 X203.484234 Y229.029299 Z-1.000000 I-0.091367 J-0.093588
G02 X138.234560 Y206.479204 Z-1.000000 I-4149.318944 J11900.557247
G02 X72.676539 Y184.157888 Z-1.000000 I-6105.060835 J17823.217389
G02 X72.463883 Y184.220310 Z-1.000000 I-0.061847 J0.182746
G02 X72.473524 Y184.912492 Z-1.000000 I0.426410 J0.340219
G02 X93.585334 Y209.563308 Z-1.000000 I1790.920393 J-1512.440281
G02 X115.047879 Y233.936000 Z-1.000000 I1789.449899 J-1554.149716
G02 X116.035290 Y234.388484 Z-1.000000 I1.011044 J-0.902697
G02 X117.514610 Y234.361224 Z-1.000000 I0.359903 J-20.621580
G02 X160.372290 Y231.978553 Z-1.000000 I-942.774369 J-17344.551144
G01 X160.372290 Y231.978553 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X618.224310 Y209.122156
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X639.013188 Y184.431097 Z-1.000000 I-2898.296954 J-2461.346675 F400.000000
G02 X639.013630 Y184.216770 Z-1.000000 I-0.128699 J-0.107429
G02 X638.792262 Y184.153964 Z-1.000000 I-0.156302 J0.129383
G02 X573.199620 Y206.497942 Z-1.000000 I5902.649045 J17435.175022
G02 X509.418343 Y228.883895 Z-1.000000 I1491.699615 J4352.153118
G02 X509.452890 Y229.399140 Z-1.000000 I0.090781 J0.252694
G02 X512.576738 Y229.829428 Z-1.000000 I4.459535 J-20.821242
G02 X549.453585 Y232.017497 Z-1.000000 I330.681598 J-5261.328843
G02 X585.875243 Y233.926016 Z-1.000000 I298.179482 J-5341.890292
G02 X591.537090 Y233.981004 Z-1.000000 I3.531360 J-72.092884
G01 X597.256660 Y233.812100 Z-1.000000
G01 X618.224310 Y209.122156 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X216.433740 Y228.392357
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X220.033850 Y228.392357 Z-1.000000 F400.000000
G01 X220.033850 Y218.842647 Z-1.000000
G01 X220.033850 Y209.292939 Z-1.000000
G01 X213.557070 Y208.022942 Z-1.000000
G02 X207.068720 Y206.815604 Z-1.000000 I-65.799568 J335.575158
G02 X147.864050 Y196.453660 Z-1.000000 I-4918.971584 J27931.031559
G03 X88.677565 Y186.025033 Z-1.000000 I2335.352957 J-13427.197285
G03 X86.180462 Y185.444760 Z-1.000000 I4.357052 J-24.412861
G02 X84.612656 Y185.277022 Z-1.000000 I-1.262584 J4.390241
G02 X84.504469 Y185.557010 Z-1.000000 I0.011060 J0.165170
G02 X86.233800 Y186.666897 Z-1.000000 I3.195936 J-3.077441
G02 X147.787459 Y207.966988 Z-1.000000 I2868.961835 J-8191.224491
G02 X209.576210 Y228.829830 Z-1.000000 I2803.336674 J-8200.614197
G02 X211.190630 Y228.864006 Z-1.000000 I0.862314 J-2.585947
G03 X212.856204 Y228.552955 Z-1.000000 I2.466387 J8.591849
G03 X216.433740 Y228.392380 Z-1.000000 I3.577536 J39.772670
G01 X216.433740 Y228.392357 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X564.161510 Y207.748572
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X602.885864 Y194.387747 Z-1.000000 I-2590.416274 J-7570.733196 F400.000000
G02 X617.842700 Y189.069036 Z-1.000000 I-379.956784 J-1092.172430
G02 X625.284276 Y186.282426 Z-1.000000 I-138.750708 J-381.860019
G02 X627.134783 Y185.486264 Z-1.000000 I-9.485269 J-24.595097
G02 X627.087172 Y185.288843 Z-1.000000 I-0.044626 J-0.093690
G02 X625.295030 Y185.535400 Z-1.000000 I0.241499 J8.391896
G03 X622.809446 Y186.065263 Z-1.000000 I-8.887294 J-35.595341
G03 X563.611450 Y196.466260 Z-1.000000 I-3579.647688 J-20200.148922
G02 X504.405446 Y206.836851 Z-1.000000 I5467.008534 J31385.499778
G02 X497.918420 Y208.036999 Z-1.000000 I66.583371 J378.027631
G01 X491.441640 Y209.292939 Z-1.000000
G01 X491.441640 Y218.769939 Z-1.000000
G01 X491.441640 Y228.246951 Z-1.000000
G01 X496.067910 Y228.711688 Z-1.000000
G02 X500.674943 Y228.933845 Z-1.000000 I4.440310 J-44.201245
G02 X502.544690 Y228.687358 Z-1.000000 I-0.027871 J-7.426228
G02 X504.376008 Y228.134931 Z-1.000000 I-7.063130 J-26.726222
G02 X564.161510 Y207.748555 Z-1.000000 I-9697.163919 J-28535.958052
G01 X564.161510 Y207.748572 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X255.810330 Y223.695776
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X279.229985 Y220.168092 Z-1.000000 I-61.666634 J-488.896927 F400.000000
G02 X279.250090 Y219.586719 Z-1.000000 I-0.051330 J-0.292809
G02 X276.145123 Y218.922414 Z-1.000000 I-11.960307 J48.314090
G02 X251.492480 Y214.568107 Z-1.000000 I-550.507609 J3044.827011
G03 X226.828746 Y210.263962 Z-1.000000 I1066.148960 J-6182.092937
G03 X224.660120 Y209.830324 Z-1.000000 I8.512900 J-48.212586
G02 X223.471774 Y210.559189 Z-1.000000 I-0.218434 J0.977040
G02 X222.501190 Y218.267818 Z-1.000000 I30.126675 J7.708629
G01 X222.501190 Y227.187969 Z-1.000000
G01 X225.893800 Y227.114629 Z-1.000000
G02 X229.274895 Y226.885367 Z-1.000000 I-0.795233 J-36.774267
G02 X255.810330 Y223.695757 Z-1.000000 I-259.693860 J-2272.453678
G01 X255.810330 Y223.695776 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X488.974300 Y218.370528
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X488.974300 Y209.553074 Z-1.000000 F400.000000
G01 X459.674600 Y214.708407 Z-1.000000
G02 X430.386576 Y219.909748 Z-1.000000 I1619.238689 J9202.765640
G02 X429.758050 Y220.070497 Z-1.000000 I0.784253 J4.375556
G02 X429.777839 Y220.279035 Z-1.000000 I0.034179 J0.101965
G02 X455.266095 Y223.609511 Z-1.000000 I401.950494 J-2976.945614
G02 X480.550556 Y226.704019 Z-1.000000 I412.416176 J-3264.906659
G02 X484.964870 Y227.044105 Z-1.000000 I6.349979 J-53.603818
G01 X488.974300 Y227.187969 Z-1.000000
G01 X488.974300 Y218.370528 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X263.989750 Y208.978316
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X246.878890 Y200.529937 Z-1.000000 F400.000000
G01 X237.157390 Y203.902098 Z-1.000000
G02 X227.746786 Y207.495477 Z-1.000000 I53.478613 J154.173003
G02 X227.435880 Y207.962258 Z-1.000000 I0.194950 J0.466781
G02 X227.767261 Y208.366080 Z-1.000000 I0.411740 J0.000000
G02 X251.345442 Y212.875220 Z-1.000000 I313.652611 J-1576.182204
G02 X274.420287 Y216.944186 Z-1.000000 I304.331014 J-1658.379197
G02 X278.633260 Y217.357553 Z-1.000000 I5.063171 J-29.927375
G02 X278.775958 Y216.849643 Z-1.000000 I0.007625 J-0.271859
G02 X263.989750 Y208.978316 Z-1.000000 I-136.154044 J237.940464
G01 X263.989750 Y208.978316 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X460.908270 Y212.605667
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X483.675002 Y208.330801 Z-1.000000 I-124.398522 J-725.273524 F400.000000
G02 X484.039610 Y207.884254 Z-1.000000 I-0.091146 J-0.446547
G02 X483.699579 Y207.373846 Z-1.000000 I-0.553094 J-0.000000
G02 X474.322510 Y203.827960 Z-1.000000 I-56.934517 J136.391188
G01 X464.605410 Y200.525635 Z-1.000000
G01 X447.490140 Y209.089867 Z-1.000000
G02 X433.876548 Y216.553688 Z-1.000000 I92.527281 J184.911701
G02 X434.075900 Y217.113616 Z-1.000000 I0.155560 J0.260067
G02 X437.770631 Y216.533535 Z-1.000000 I-25.208898 J-172.620562
G02 X460.908270 Y212.605667 Z-1.000000 I-1138.632118 J-6777.378044
G01 X460.908270 Y212.605667 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X268.644690 Y197.379370
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X251.780161 Y178.442396 Z-1.000000 I-5089.654114 J4515.669136 F400.000000
G02 X250.111580 Y176.646700 Z-1.000000 I-47.391380 J42.363524
G02 X249.231569 Y176.920583 Z-1.000000 I-0.371082 J0.358398
G02 X248.408300 Y186.903500 Z-1.000000 I60.114518 J9.982917
G01 X248.408300 Y198.923883 Z-1.000000
G01 X265.988130 Y207.566831 Z-1.000000
G02 X283.613561 Y216.041540 Z-1.000000 I493.029710 J-1002.821847
G02 X284.521230 Y216.279133 Z-1.000000 I1.094232 J-2.327688
G02 X284.685067 Y215.958507 Z-1.000000 I0.014194 J-0.194920
G02 X268.644690 Y197.379370 Z-1.000000 I-695.361197 J584.129211
G01 X268.644690 Y197.379370 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X445.168070 Y207.635929
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X463.067190 Y198.923369 Z-1.000000 F400.000000
G01 X463.067190 Y186.903250 Z-1.000000
G02 X462.243927 Y176.920584 Z-1.000000 I-60.935231 J-0.000000
G02 X461.363910 Y176.646700 Z-1.000000 I-0.508932 J0.084517
G02 X456.412525 Y181.990410 Z-1.000000 I126.570854 J122.244048
G02 X441.713491 Y198.660501 Z-1.000000 I2097.100747 J1863.956925
G02 X426.548282 Y216.159822 Z-1.000000 I1884.891485 J1648.799879
G02 X426.634960 Y216.348482 Z-1.000000 I0.086678 J0.074418
G02 X427.237909 Y216.213825 Z-1.000000 I0.000000 J-1.417232
G02 X445.168070 Y207.635929 Z-1.000000 I-624.183534 J-1327.743607
G01 X445.168070 Y207.635929 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X230.520050 Y204.307972
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X238.828684 Y200.835973 Z-1.000000 I-25.097946 J-71.737843 F400.000000
G02 X239.176339 Y197.826222 Z-1.000000 I-0.790961 J-1.616317
G02 X233.234364 Y194.007989 Z-1.000000 I-19.651965 J24.049999
G02 X194.916970 Y175.215670 Z-1.000000 I-7695.638201 J15642.886547
G02 X145.818219 Y151.638681 Z-1.000000 I-1692.733303 J3462.181339
G02 X145.152760 Y151.750440 Z-1.000000 I-0.252266 J0.534995
G02 X144.981438 Y152.672370 Z-1.000000 I0.568744 J0.582573
G02 X155.478011 Y174.888692 Z-1.000000 I976.020029 J-447.553549
G02 X166.474211 Y196.857137 Z-1.000000 I981.094244 J-477.345749
G02 X167.575340 Y197.835830 Z-1.000000 I1.816502 J-0.934963
G02 X171.120085 Y198.813352 Z-1.000000 I7.270511 J-19.448904
G02 X195.819656 Y203.135844 Z-1.000000 I313.873440 J-1720.803986
G02 X220.963666 Y207.158978 Z-1.000000 I283.599298 J-1691.870224
G02 X222.501190 Y207.042761 Z-1.000000 I0.516822 J-3.391208
G02 X224.343621 Y206.441598 Z-1.000000 I-21.712952 J-69.669225
G02 X230.520050 Y204.307972 Z-1.000000 I-267.902613 J-785.532296
G01 X230.520050 Y204.307972 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X518.582420 Y202.486634
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X543.721440 Y197.822921 Z-1.000000 I-174.113161 J-1008.615871 F400.000000
G02 X544.723350 Y197.298490 Z-1.000000 I-0.389777 J-1.963931
G02 X545.981129 Y195.575954 Z-1.000000 I-4.341380 J-4.490511
G02 X556.352795 Y174.161595 Z-1.000000 I-1318.822006 J-651.966937
G02 X566.450735 Y152.408391 Z-1.000000 I-1298.329945 J-615.911218
G02 X566.328220 Y151.755800 Z-1.000000 I-0.526295 J-0.238991
G02 X565.656619 Y151.642989 Z-1.000000 I-0.417029 J0.427177
G02 X516.558520 Y175.219250 Z-1.000000 I1628.192937 J3453.660876
G02 X469.757549 Y199.164597 Z-1.000000 I637.484991 J1303.667565
G02 X469.820710 Y200.210637 Z-1.000000 I0.271907 J0.508509
G02 X489.243737 Y207.138398 Z-1.000000 I134.257144 J-345.718667
G02 X491.441640 Y207.251174 Z-1.000000 I1.307939 J-4.016575
G03 X493.299085 Y206.888129 Z-1.000000 I10.361171 J48.077618
G03 X518.582420 Y202.486634 Z-1.000000 I1563.372608 J8905.602746
G01 X518.582420 Y202.486634 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X153.015990 Y173.645900
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X141.434343 Y150.605305 Z-1.000000 I-737.889087 J356.478168 F400.000000
G02 X140.744010 Y150.160340 Z-1.000000 I-0.736694 J0.384946
G02 X139.172225 Y150.453013 Z-1.000000 I-0.187646 J3.359190
G02 X106.445253 Y165.438425 Z-1.000000 I1370.321699 J3035.905003
G02 X73.928889 Y180.760230 Z-1.000000 I1365.294964 J2939.631668
G02 X73.851571 Y181.228160 Z-1.000000 I0.124740 J0.260964
G02 X74.621279 Y181.627268 Z-1.000000 I1.041515 J-1.066861
G02 X116.465045 Y189.251191 Z-1.000000 I1248.267459 J-6732.452491
G02 X157.726720 Y196.507952 Z-1.000000 I1219.837239 J-6815.015525
G02 X161.651690 Y196.943720 Z-1.000000 I5.559848 J-32.183592
G02 X162.463726 Y195.836885 Z-1.000000 I0.041159 J-0.821098
G02 X153.015990 Y173.645900 Z-1.000000 I-237.949677 J88.199595
G01 X153.015990 Y173.645900 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X245.933610 Y184.733310
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G01 X245.925610 Y172.388340 Z-1.000000 F400.000000
G01 X201.212350 Y162.425110 Z-1.000000
G02 X156.489397 Y152.509428 Z-1.000000 I-4700.411628 J21094.570351
G02 X153.414970 Y151.880920 Z-1.000000 I-21.846754 J99.032506
G02 X153.280170 Y152.276729 Z-1.000000 I-0.039095 J0.207544
G02 X197.210330 Y174.144310 Z-1.000000 I1318.088608 J-2592.871081
G02 X244.135341 Y196.811193 Z-1.000000 I3314.494821 J-6801.765908
G02 X245.015110 Y197.033470 Z-1.000000 I0.990767 J-2.069252
G02 X245.494018 Y196.610115 Z-1.000000 I0.022203 J-0.457437
G02 X245.933010 Y184.733310 Z-1.000000 I-157.887344 J-11.782365
G01 X245.933610 Y184.733310 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X514.286420 Y174.174170
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X558.193395 Y152.261457 Z-1.000000 I-1305.171503 J-2670.145296 F400.000000
G02 X558.059920 Y151.874790 Z-1.000000 I-0.093741 J-0.184012
G02 X554.983406 Y152.516163 Z-1.000000 I25.393394 J129.505472
G02 X510.265730 Y162.433940 Z-1.000000 I6048.959959 J27379.566731
G01 X465.555730 Y172.388340 Z-1.000000
G01 X465.544430 Y184.733310 Z-1.000000
G02 X465.989641 Y196.622850 Z-1.000000 I155.314914 J0.137250
G02 X466.480940 Y197.078280 Z-1.000000 I0.491299 J-0.037282
G02 X467.380742 Y196.873851 Z-1.000000 I0.000000 J-2.082471
G02 X514.285710 Y174.174170 Z-1.000000 I-2869.234321 J-5988.587217
G01 X514.286420 Y174.174170 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X591.939890 Y189.792760
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X633.869133 Y182.285526 Z-1.000000 I-677.315147 J-3903.770920 F400.000000
G02 X636.220620 Y181.618600 Z-1.000000 I-2.368190 J-12.828867
G02 X636.273200 Y180.663189 Z-1.000000 I-0.189997 J-0.489609
G02 X605.112980 Y165.418570 Z-1.000000 I-318.818197 J612.200708
G02 X571.673167 Y150.330182 Z-1.000000 I-1272.584653 J2775.780392
G02 X570.747640 Y150.160340 Z-1.000000 I-0.815328 J1.836325
G02 X569.628700 Y150.890815 Z-1.000000 I0.073337 J1.334569
G02 X558.479316 Y173.651785 Z-1.000000 I838.651464 J424.921818
G02 X548.101690 Y196.400642 Z-1.000000 I805.526757 J381.208520
G02 X548.544130 Y197.078280 Z-1.000000 I0.442440 J0.194382
G02 X549.939045 Y196.962001 Z-1.000000 I-0.000000 J-8.425018
G02 X591.939890 Y189.792760 Z-1.000000 I-1292.402844 J-7698.129773
G01 X591.939890 Y189.792760 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X104.105660 Y164.520690
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X137.802249 Y148.727523 Z-1.000000 I-2512.524665 J-5404.612880 F400.000000
G02 X138.379670 Y148.326190 Z-1.000000 I-0.909111 J-1.924042
G02 X138.318371 Y147.920047 Z-1.000000 I-0.176724 J-0.181025
G02 X108.586600 Y133.050430 Z-1.000000 I-756.007598 J1474.476626
G02 X78.359954 Y118.749703 Z-1.000000 I-620.724483 J1272.895997
G02 X77.672805 Y118.868820 Z-1.000000 I-0.256844 J0.559880
G02 X77.312118 Y119.587048 Z-1.000000 I0.861158 J0.882145
G02 X73.881262 Y147.130870 Z-1.000000 I1457.446337 J195.524989
G03 X70.566668 Y174.774494 Z-1.000000 I-2850.758348 J-327.798129
G03 X70.089459 Y177.506990 Z-1.000000 I-28.263615 J-3.528101
G02 X69.725696 Y180.019466 Z-1.000000 I15.926455 J3.588448
G02 X69.910482 Y180.216860 Z-1.000000 I0.184786 J0.012205
G02 X70.322195 Y180.126882 Z-1.000000 I-0.000000 J-0.986922
G02 X104.105660 Y164.520690 Z-1.000000 I-3210.675094 J-6994.670545
G01 X104.105660 Y164.520690 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X637.998690 Y149.540200
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X634.165477 Y119.588340 Z-1.000000 I-1731.645172 J206.393256 F400.000000
G02 X633.803260 Y118.869380 Z-1.000000 I-1.228071 J0.167986
G02 X633.115465 Y118.750142 Z-1.000000 I-0.430718 J0.441184
G02 X602.888890 Y133.050430 Z-1.000000 I589.931434 J1286.035556
G02 X573.236724 Y147.932883 Z-1.000000 I635.005919 J1302.182511
G02 X573.167470 Y148.396150 Z-1.000000 I0.132496 J0.256617
G02 X574.023635 Y148.988141 Z-1.000000 I2.168996 J-2.221794
G02 X607.507784 Y164.512899 Z-1.000000 I3243.127167 J-6950.978139
G02 X641.163179 Y179.919742 Z-1.000000 I3186.595405 J-6916.478248
G02 X641.427430 Y179.845230 Z-1.000000 I0.088047 J-0.193576
G02 X641.489628 Y179.591970 Z-1.000000 I-0.312087 J-0.210913
G02 X637.998690 Y149.540200 Z-1.000000 I-5038.028469 J570.009577
G01 X637.998690 Y149.540200 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X228.126100 Y151.972450
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X212.312373 Y134.841416 Z-1.000000 I-518.177347 J462.467238 F400.000000
G02 X211.397850 Y134.450130 Z-1.000000 I-0.914523 J0.873080
G02 X209.658346 Y134.630398 Z-1.000000 I-0.000000 J8.482859
G02 X178.977444 Y141.154280 Z-1.000000 I1075.603151 J5133.816470
G02 X148.116988 Y147.910641 Z-1.000000 I1100.989369 J5102.753324
G02 X147.523090 Y148.231990 Z-1.000000 I0.267301 J1.203487
G02 X147.597385 Y148.532033 Z-1.000000 I0.125178 J0.128224
G02 X153.626160 Y150.039370 Z-1.000000 I20.565939 J-69.445885
G03 X160.188094 Y151.445162 Z-1.000000 I-75.147794 J366.791428
G03 X201.528770 Y160.751640 Z-1.000000 I-3257.913919 J14568.582821
G02 X242.815861 Y169.902902 Z-1.000000 I989.796452 J-4367.881893
G02 X243.310680 Y169.805840 Z-1.000000 I0.134857 J-0.622318
G02 X243.358635 Y169.445560 Z-1.000000 I-0.138249 J-0.201733
G02 X228.126100 Y151.972450 Z-1.000000 I-757.390571 J644.894282
G01 X228.126100 Y151.972450 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X563.508650 Y148.927060
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X564.090235 Y148.685953 Z-1.000000 I-0.018680 J-0.867049 F400.000000
G02 X564.091650 Y148.367930 Z-1.000000 I-0.153145 J-0.159696
G02 X563.341088 Y147.959508 Z-1.000000 I-1.098937 J1.125662
G02 X532.905056 Y141.171869 Z-1.000000 I-889.920369 J3918.804422
G02 X502.678974 Y134.674977 Z-1.000000 I-865.304400 J3952.174500
G02 X500.523750 Y134.450130 Z-1.000000 I-2.155224 J10.216793
G02 X499.448319 Y134.907439 Z-1.000000 I0.000000 J1.493175
G02 X483.773279 Y151.526405 Z-1.000000 I784.767776 J755.896272
G02 X468.457431 Y168.461217 Z-1.000000 I802.349885 J741.038647
G02 X468.001880 Y169.662760 Z-1.000000 I1.356795 J1.201543
G02 X468.241908 Y169.856815 Z-1.000000 I0.198458 J0.000000
G02 X515.189820 Y159.554040 Z-1.000000 I-1008.040396 J-4705.578758
G03 X562.391079 Y149.060662 Z-1.000000 I2339.009474 J10409.913385
G03 X563.508650 Y148.927060 Z-1.000000 I1.241710 J5.645814
G01 X563.508650 Y148.927060 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X182.715280 Y138.519400
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X197.531527 Y135.131681 Z-1.000000 I-90.617668 J-430.411319 F400.000000
G02 X202.130548 Y133.731000 Z-1.000000 I-10.760666 J-43.582355
G02 X203.974720 Y132.798393 Z-1.000000 I-2.978153 J-8.178779
G02 X204.018705 Y132.473132 Z-1.000000 I-0.121109 J-0.181982
G02 X203.558241 Y132.223723 Z-1.000000 I-0.561925 J0.487671
G02 X200.603520 Y132.005720 Z-1.000000 I-3.201281 J23.256303
G03 X197.226119 Y131.765814 Z-1.000000 I0.296301 J-28.064809
G03 X137.994680 Y123.739320 Z-1.000000 I1143.767871 J-8662.986983
G02 X78.841681 Y115.730310 Z-1.000000 I-1162.818133 J8365.898672
G02 X78.336094 Y115.876440 Z-1.000000 I-0.085786 J0.650886
G02 X78.371100 Y116.186923 Z-1.000000 I0.119551 J0.143736
G02 X109.721540 Y131.789670 Z-1.000000 I1061.999701 J-2094.567755
G01 X141.549320 Y147.335010 Z-1.000000
G01 X151.491880 Y145.167410 Z-1.000000
G03 X161.439619 Y143.024020 Z-1.000000 I445.107234 J2041.641277
G03 X182.715280 Y138.519400 Z-1.000000 I1985.270101 J9324.087108
G01 X182.715280 Y138.519400 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Start cutting path id: path826)
(Change tool to Plasma cutter)
 
G00 Z5.000000
G00 X633.139480 Y115.870640
G31 Z-100 F500 (find metal)
G92 Z0 (zero z)
G00 Z10 F500 (going up)
M03 (turn on plasma)
G04 P0.2 (pause)
G01 Z1 (going to cutting z)
G01 Z-1.000000 F100.0(Penetrate)
G02 X632.631974 Y115.721870 Z-1.000000 I-0.421914 J0.499271 F400.000000
G02 X573.480810 Y123.750260 Z-1.000000 I1090.443728 J8256.035744
G03 X514.249215 Y131.790977 Z-1.000000 I-1158.803299 J-8314.090740
G03 X510.871980 Y132.022540 Z-1.000000 I-3.526721 J-26.692065
G02 X507.835923 Y132.241810 Z-1.000000 I0.127140 J22.888937
G02 X507.569745 Y132.501732 Z-1.000000 I0.043345 J0.310642
G02 X507.737487 Y132.819799 Z-1.000000 I0.294057 J0.048186
G02 X510.738765 Y133.978696 Z-1.000000 I9.853972 J-21.053775
G02 X517.295299 Y135.761526 Z-1.000000 I29.686150 J-96.226198
G02 X538.630410 Y140.586180 Z-1.000000 I268.436122 J-1137.466537
G01 X569.779800 Y147.324420 Z-1.000000
G01 X601.680850 Y131.784460 Z-1.000000
G02 X633.102683 Y116.184692 Z-1.000000 I-1022.276033 J-2098.565190
G02 X633.139480 Y115.870640 Z-1.000000 I-0.085598 J-0.169211
G01 X633.139480 Y115.870640 Z-1.000000
G00 Z5.000000
M05 (turn off plasma)
(End cutting path id: path826)


(Footer)
M5
G00 X0.0000 Y0.0000
M2
(Using default footer. To add your own footer create file "footer" in the output dir.)
(end)
%
"""

